'********************************************************************
' GRD grid control for FreeBasic
'********************************************************************

'     Compilable for Freebasic by James Klutho
'     LICENSE and DISCLAIMER
'     This code is free for all to use without acknowledging the author.
'     Use this code as you see fit. By using or compiling this code or derivative
'     thereof, you are consenting to the hold the author, James Klutho, harmless
'     for all effects or side-effects its use.  This code is release into the 
'     public domain.
'--------------------------------------------------------------------
' Messages
'--------------------------------------------------------------------
'  GRD_REFRESH      - to refresh control with new array data
' wParam = Not used, should be 0.
' lParam = Not used, should be 0.
' Returns: Nothing, 0
' Example: SendMessage hWndCtrl,  GRD_REFRESH, 0, 0
'--------------------------------------------------------------------

'--------------------------------------------------------------------
' Notifications
'--------------------------------------------------------------------
'  GRDN_RETURN
' Notification sent to parent's  WM_COMMAND when Enter key has been pressed
'--------------------------------------------------------------------
'  GRDN_SPACE
' Notification sent to parent's  WM_COMMAND when Space bar has been pressed
'--------------------------------------------------------------------
'  GRDN_DELETE
' Notification sent to parent's  WM_COMMAND when Delete key has been pressed

'--------------------------------------------------------------------
' Useful Functions and Sub
'--------------------------------------------------------------------


'********************************************************************
' Declares
'********************************************************************

#INCLUDE ONCE "string.bi"    'need for FORMAT
#INCLUDE ONCE "vbcompat.bi"  'need for DATE FORMATTING

#define DEBUGMODE  1
#define VERSION Alpha

'Messages

#define GRD_REFRESH         WM_USER + 2
#define GRD_GETSELECTED     WM_USER + 3
#define GRD_SETSELECTED     WM_USER + 4
#define GRD_GETTOPROW       WM_USER + 5
#define GRD_SETTOPROW       WM_USER + 6
#define GRD_SIZEHANDLER     WM_USER + 7
#define GRD_GETROWCOLDIM    WM_USER + 10
#define GRD_GETROWCOLVIEW   WM_USER + 11
#define GRD_SETROWCOLVIEW   WM_USER + 12
#define GRD_GETTOPLEFT      WM_USER + 13
#define GRD_GETTOTALSHEETS  WM_USER + 14
#define GRD_GETCURRENTSHEET WM_USER + 15
#define GRD_SETCURRENTSHEET WM_USER + 16
#define GRD_GETSETCELL      WM_USER + 17
#define GRD_SETSTATUSBAR    WM_USER + 18
#define GRD_GETLISTHANDLE   WM_USER + 19   
#define GRD_GETEDITHANDLE   WM_USER + 20
#define GRD_SETTOTROWCOL    WM_USER + 21
#define GRD_GETTOTROWCOL    WM_USER + 22
#define GRD_SETHSCROLLWIDTH WM_USER + 23          
#define GRD_SETVSCROLLWIDTH WM_USER + 24
#define GRD_SETSTATUSBARWIDTH WM_USER + 25
#define GRD_SETRETURNKEYMODE WM_USER + 26
#define GRD_SETTABKEYMODE WM_USER + 27
#define GRD_SETNUMBERDATEFORMATSTR WM_USER + 28
#define GRD_SETDATEINPUTFORMAT WM_USER + 29
#define GRD_BITMAPTODRAWRECT  WM_USER + 30
#define GRD_LOADBITMAP      WM_USER + 31    'need to work on
#define GRD_SETTABPOSITION  WM_USER + 32
#define GRD_SETSWITCH       WM_USER + 33

'Notifications
#define GRDN_RCLICKMENU       WM_USER + 203
#define GRDN_COLWIDTHCHANGED  WM_USER + 204
#define GRDN_ROWHEIGHTCHANGED WM_USER + 205
#define GRDN_DIRTY            WM_USER + 208
#define GRDN_CELLCHANGED      WM_USER + 209
#define GRDN_ROWSSELECTED     WM_USER + 210
#define GRDN_COLSSELECTED     WM_USER + 211
#define GRDN_CELLSSELECTED    WM_USER + 212
#define GRDN_SHEETCHANGED     WM_USER + 213
#define GRDN_KILLFOCUS        WM_USER + 218
#define GRDN_SETFOCUS         WM_USER + 219
#define GRDN_CHECKBOXTOGGLE   WM_USER + 220
#define GRDN_USERBUTTON       WM_USER + 221
#define GRDN_COMBOBOX         WM_USER + 222


'Equates
#define GRD_ROWCOLMIN    10
#define GRD_COLMAXWIDTH  600
#define GRD_ROWMAXHEIGHT 255
#define GRD_NUMMENUS     6
#define GRD_NUMLISTS     255
#define SHEETTABBUTTONWIDTH 6
#define GRD_COLRESIZE       5
#define HEADERBYTES        16
#define SBSIZE             16   'Used for SBars, SheetControl and Corner size box
#define MAXGS              10   'Grid Sheets 
#define NUMCELLCOLORS      15   'Fill the Cell Color Array 
#define NUMGRDFONTS        15   'Number of available Fonts
#define NUMDATEINPUTS      16      
#define GRIDSTYLES         WS_VISIBLE OR WS_CHILD OR WS_TABSTOP OR WS_DLGFRAME
#define GS_                v->SH(v->CS)
#define GRD_NUMDATEFORMATSTR  32
#define GRD_DATEINPUTFORMATS  15
#define GRD_NUMBITMAPS  32
#define GRD_DEFAULTFONT "Arial"
#define GRD_MAXCELLLEN     1024
#define DEF_BASE26     "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
#define GRD_CRLF Chr(13,10)

'Child Control IDs
const GRD_HSBAR      =101
const GRD_VSBAR      =102
const GRD_LRCOR      =103
const GRD_STATB      =104
const GRD_SCNTL      =105
const GRD_ECNTL      =106
const GRD_LCNTL      =107

'CellColors
CONST CELLDEFAULT = 0
CONST CELLCOLORBLACK = 1
CONST CELLCOLORWHITE   = 2
CONST CELLCOLORSALMON = 3
CONST CELLCOLORGREEN = 4
CONST CELLCOLORLIGHTGREEN = 5
CONST CELLCOLORBLUE = 6
CONST CELLCOLORLIGHTBLUE = 7
CONST CELLCOLORMAGENTA = 8
CONST CELLCOLORDARKMAGENTA = 9
CONST CELLCOLORCYAN = 10
CONST CELLCOLORAQUAMARINE = 11
CONST CELLCOLORRED = 12
CONST CELLCOLORBROWN = 13
CONST CELLCOLORYELLOW = 14
CONST CELLCOLORORANGE = 15

'System Fonts
CONST GRDSYSFONTDEFAULT = 0
CONST GRDSYSFONTNORM = 1
CONST GRDSYSFONTBOLD = 2
CONST GRDSYSFONTITAL = 3
CONST GRDSYSFONTTABNORM = 4
CONST GRDSYSFONTTABBOLD = 5
CONST GRDSYSFONTNORMVERT = 6
CONST GRDSYSFONTBOLDVERT = 7
CONST GRDSYSFONTSTATUSBAR = 8
'8
'9
'10
'11
'12
'13
'14
'15

'MouseMove Results
CONST GRDRM_CELLS = 0  ' On the Cell Grid
CONST GRDRM_ROWH  = 1  ' On the Row Header
CONST GRDRM_COLH  = 2  ' On the Column Header 
CONST GRDRM_OGTOP = 3  ' Above the Grid if space present - OFF GRID
CONST GRDRM_OGLEFT = 4 ' Left of the Grid if space present - OFF GRID
CONST GRDRM_CORN  = 5  'On the Corner Cell
CONST GRDRM_TAB   = 6  'In the Tab Control

'Format Attributes Actions SET
CONST GRDFORMAT_SETTYPE         = 0  'Set the cell type in column or cell 
CONST GRDFORMAT_SETNDFORMAT     = 1  'Set a Number or Date Format
CONST GRDFORMAT_SETFORECLR      = 2  'Set the cell foreground color in column or cell
CONST GRDFORMAT_SETBACKCLR      = 3  'Set the cell backgroung color in column or cell
CONST GRDFORMAT_SETCASE         = 4  'Set the cell font case in column or cell
CONST GRDFORMAT_SETJUST         = 5  'Set the cell justification in column or cell
CONST GRDFORMAT_SETFONT         = 6  'Set the cell font in column or cell
CONST GRDFORMAT0_SETCASEC       = 7  'Set the corner font case for the column
CONST GRDFORMAT0_SETJUSTC       = 8  'Set the corner justification for the column
CONST GRDFORMAT0_SETCASER       = 9  'Set the corner font case for the row
CONST GRDFORMAT0_SETJUSTR       = 10  'Set the corner justification for the row
CONST GRDFORMAT0_SETROWFORMAT   = 11  'Set the Row Format             
CONST GRDFORMAT0_SETCOLFORMAT   = 12  'Set the Col Format

CONST GRDFORMAT_SETROWRECNO   = 15  'Set the row RecNo from a database

CONST GRDFORMAT_SETROWHEIGHT  = 20  'Set the row Height in pixels
CONST GRDFORMAT_SETCOLWIDTH   = 21  'Set the col Width in pixels
CONST GRDFORMAT_SETROWHIDE    = 22  'Hide the Row
CONST GRDFORMAT_SETCOLHIDE    = 23  'Hide the Column
CONST GRDFORMAT_SETWRITELOCK  = 24  'Set a col cell row writelocked
CONST GRDFORMAT_SETCELLBORDER = 25  'Set a cell border
CONST GRDFORMAT_SETROWUNHIDE  = 26  'Unhide the Row
CONST GRDFORMAT_SETCOLUNHIDE  = 27  'Unhide the Col

'Miscellaneous
CONST GRDFORMAT_CLEARFORMATS   = 45 'EveryThing but RowHeight ColWidth RowHide and ColHide

'Format Attributes Actions GET
CONST GRDFORMAT_GETTYPE      = 50  'Get the cell type in column or cell 
CONST GRDFORMAT_GETNDFORMAT  = 51  'Get a Number or Date Format
CONST GRDFORMAT_GETFORECLR   = 52  'Get the cell foreground color in column or cell
CONST GRDFORMAT_GETBACKCLR   = 53  'Get the cell backgroung color in column or cell
CONST GRDFORMAT_GETCASE      = 54  'Get the cell font case in column or cell
CONST GRDFORMAT_GETJUST      = 55  'Get the cell justification in column or cell
CONST GRDFORMAT_GETFONT      = 56  'Get the dell font in column or cell
CONST GRDFORMAT0_GETCASEC       = 57  'Get the corner font case for the column
CONST GRDFORMAT0_GETJUSTC       = 58  'Get the corner justification for the column
CONST GRDFORMAT0_GETCASER       = 59  'Get the corner font case for the row
CONST GRDFORMAT0_GETJUSTR       = 60  'Get the corner justification for the row

CONST GRDFORMAT_GETROWRECNO  = 65  ' Get the row RecNo from a database

CONST GRDFORMAT_GETROWHEIGHT  = 70  'Get the row Height in pixels
CONST GRDFORMAT_GETCOLWIDTH   = 71  'Get the col Width in pixels
CONST GRDFORMAT_GETROWHIDE    = 72  'Is a row hidden or not
CONST GRDFORMAT_GETCOLHIDE    = 73  'Is a column hidden or not
CONST GRDFORMAT_GETWRITELOCK  = 74  'Is a col cell row writelocked
CONST GRDFORMAT_GETCELLBORDER = 75  'Get a cell border


'Cell Types 0 to 15
CONST GRDFCT_AUTO    = 0 'Text Automatic  Horizontal No Word Wrap
CONST GRDFCT_HORZ    = 1 'Text Horizontal No word wrap
CONST GRDFCT_HORZWW  = 2 'Text Horizontal with word wrap
CONST GRDFCT_VERT    = 3 'Text Vertical No word wrap
CONST GRDFCT_VERTWW  = 4 'Text Vertical with word wrap  
CONST GRDFCT_CKBOX   = 5 'Checkbox
CONST GRDFCT_DDCBS   = 6 'Dropdown Comboxbox Static
CONST GRDFCT_PICTURE   = 7 'Display icon or bitmap in cell
CONST GRDFCT_NUMBER  = 8 'Number
CONST GRDFCT_DATE    = 9 'Date
CONST GRDFCT_USERBUT = 10 'User Button
'11
'12
'13
'14
'15

'Cell Justify
CONST GRDFCJ_AUTO = 0 'Text Left Bottom
CONST GRDFCJ_LEFT = 1 'Text Left Bottom
CONST GRDFCJ_RGHT = 2 'Text Right Bottom
CONST GRDFCJ_CNTR = 3 'Text Center Bottom
CONST GRDFCJ_LEFT_T = 4 'Text Left Top
CONST GRDFCJ_RGHT_T = 5 'Text Right Top
CONST GRDFCJ_CNTR_T = 6 'Text Center Top
CONST GRDFCJ_LEFT_C = 7'Text Left VCenter
CONST GRDFCJ_RGHT_C = 8 'Text Right VCenter
CONST GRDFCJ_CNTR_C = 9 'Text Center VCenter

'Borders -  OR them together for borders on a cell
CONST GRD_TOPB     = 128 'Bold
CONST GRD_LEFTB    = 64  'Bold
CONST GRD_BOTTOMB  = 32  'Bold
CONST GRD_RIGHTB   = 16  'Bold
CONST GRD_TOP      = 8
CONST GRD_LEFT     = 4
CONST GRD_BOTTOM   = 2
CONST GRD_RIGHT    = 1 

'Cell Write Lock
CONST GRD_NOLOCK     = 0
CONST GRD_LOCK       = 1             


'Cell Case
CONST GRDFCC_AUTO = 0 'As is
CONST GRDFCC_LOWR = 1 'Lower Case
CONST GRDFCC_UPPR = 2 'Upper Case
CONST GRDFCC_TITL = 3 'Title Case

'GetSetCell
const GRD_SET = 1
const GRD_GET = 0
const GRD_CURRENTSHEET = -1
const GRD_WIDE = 0
const GRD_ANSI = 1

'Date formats
const YYYYMMDD              = 0   'Default will use a dot
const MMDDYYYYS             = 1   'Slash 
const MMDDYYYYD             = 2   'Dot instead of a slash
const MMDDYYYYL             = 3   'Line instead of a slash
const DDMMYYYYS             = 4   'Slash 
const DDMMYYYYD             = 5   'Dot instead of a slash
const DDMMYYYYL             = 6   'Line instead of a slash
const YYYYMMDDS             = 7   'Slash 
const YYYYMMDDD             = 8   'Dot instead of a slash
const YYYYMMDDL             = 9   'Line instead of a slash
const YYYYMMS               = 10  'Slash 
const YYYYMMD               = 11  'Dot instead of a slash
const YYYYMML               = 12  'Line instead of a slash
const MMYYYYS               = 13  'Slash 
const MMYYYYD               = 14  'Dot instead of a slash
const MMYYYYL               = 15  'Line instead of a slash

'Toggle Switches
const GRD_COMBOSUPRESSEDIT = 1
const GRD_ROWCOLRESIZE = 2
const GRD_RETURNKEYMODE = 3   'Not used
const GRD_TABKEYMODE = 4      'Not used
const GRD_HIDEGRIDLINES = 5
const GRD_CYTWEAK = 6
const GRD_SHOWSTATUSBAR = 7

'********************************************************************
' Types and Unions
'********************************************************************

TYPE GRDArray
  ws as Wstring Ptr
END TYPE


TYPE GETSETCELL
  Ansi    as long 'zero is WSTRING non-zero is STRING 
  Action  as LONG 'zero is get non-zero is set(put)  GRD_SET    GRD_GET
  Row     as LONG
  Col     AS LONG
  Sheet   as LONG  '-1 is current sheet  GRD_CURRENTSHEET or sheet number
END TYPE

TYPE MyNotify
   NMHeader AS NMHDR
   Param1 AS LONG
   Param2 AS LONG
   Param3 AS LONG
   Param4 AS LONG
END TYPE

TYPE CornerFormat0 'Used for Columns and Row Headers
    MyWidth           as uBYTE   'Border in a cell Width for Column and Height on Row
    CHide             as uBYTE   'Zero if column not hidden or stores MyWidth if hidden
    MyHeight          as ubyte   'Row height in pixelds
    RHide             as uBYTE   'Zero if row now hidden
    MyForeColor       as uBYTE   '255
    MyBackColor       as uBYTE   '255
    CellFormatR       as uBYTE   '255
    CellFormatC       as uBYTE   '255
    MyFont            as uBYTE   '255
    MyJustR           as uBYTE   '255
    MyJustC           as uBYTE   '255
    MyCaseR           as uBYTE   '255
    MyCaseC           as uBYTE   '255
    Unused10          as uBYTE   '255
    Unused11          as uBYTE   '255
    Unused12          as uBYTE   '255  
END TYPE  ' 16 Bytes

TYPE RowFormat1   'Used for Row
    CellFormat        as uByte
    RowCombo          as ubyte
    MyHeight          as uBYTE   'Row height in pixels
    RHide             as uBYTE   'Zero if visible - store MyHeight if hidden
    MyForeColor       as uByte
    MyBackColor       as uByte
    CellType          as uByte
    WriteLock         as uByte
    RecNo             as ULongInt   '8 bytes
END TYPE  ' 16 Bytes  includes pad

TYPE ColFormat2 'Used for Columns and Cells
    MyWidth           as uBYTE   'Border in a cell Width for Column and Height on Row
    CHide             as UBYTE   'Zero if not hidden or stores MyWidth if hidden
    CellFormat        as ubyte   '255 possible number or date formats
    ColCombo          as uBYTE   '255 Col Comboboxes
    MyForeColor       as uBYTE   '16 Foreground
    MyBackColor       as uBYTE   '16 Background
    CellType          as uBYTE   '255  For Column and Cell
    WriteLock         as uBYTE   '1 is locked  0 is unlocked  
    MyFont            as uBYTE   '16 Fonts for now
    MyJust            as uBYTE   '4 justification
    MyCase            as uBYTE   '4 cases
    Unused4           as uBYTE   'Allow for Word Size
    Unused5           as uBYTE   'Allow for Word Size
    Unused6           as uBYTE   '255
    Unused7           as uBYTE   '255
    Unused8           as uBYTE   '255
END TYPE  ' 16 Bytes

TYPE CellFormat3 'Used for Cells
    MyBorder          as uBYTE   'Border in a cell Width for Column and Height on Row
    Unused9           as uBYTE   '255
    CellFormat        as ubyte   '255 possible number or date formats
    CellCombo         as uBYTE   '255 Cell Comboboxes
    MyForeColor       as uBYTE   '16 Foreground
    MyBackColor       as uBYTE   '16 Background
    CellType          as uBYTE   '255  For Column and Cell
    WriteLock         as uBYTE   '1 is locked  0 is unlocked  
    MyFont            as uBYTE   '16 Fonts for now
    MyJust            as uBYTE   '4 justification 
    MyCase            as uBYTE   '4 cases
    Unused4           as uBYTE   'Allow for Word Size
    Unused5           as uBYTE   'Allow for Word Size
    Unused6           as uBYTE   '255
    Unused7           as uBYTE   '255
    Unused8           as uBYTE   '255
END TYPE  ' 16 Bytes

TYPE SheetData
  SheetName           as WString * 32
  SheetColor          as UINTeger
  SheetOffset1        as Long
  SheetOffset2        as Long
  

  cyChar              AS LONG
  charWidth           AS LONG
  
  wMaxHeight          AS LONG
  wMaxWidth           AS LONG
  GridTop             AS LONG
  GridLeft            AS LONG
  
  'SBSIZE             16
  
  HScrollWidth        as LONG
  VScrollWidth        as LONG
  StatusBarWidth      as LONG
  
  MyRowTemp           as Long
  OldMyRowTemp        as Long
  MyColTemp           as Long
  OldMyColTemp        as long
  TempMMResult        as long
    
  ColSelectionOn      as LONG
  ColSelection        as LONG
  RowSelectionOn      as LONG
  RowSelection        as LONG
  RowColSelectionOn   as long
  SelRow              AS LONG
  SelCol              AS LONG
  SelRow2             AS LONG
  SelCol2             AS LONG
  OldSelRow           AS LONG
  OldSelCol           AS LONG
  OldSelRow2          AS LONG
  OldSelCol2          AS LONG
  SelectionProcess    as long
  
  BottomRow           AS LONG  
  RowCount            AS LONG 
  RightCol            AS LONG
  TopRow              AS LONG
  TopRow2             AS LONG
  LeftCol             AS LONG
  LeftCol2            AS LONG
  OldRightCol         AS LONG
  OldBottomRow        AS LONG
          
  'MultiCell
  StartRow            AS LONG
  EndRow              AS LONG
  StartCol            AS LONG
  EndCol              AS LONG 
                       
  
  DimCols             AS LONG
  DimRows             AS LONG
  TotRows             AS LONG
  TotCols             AS LONG
  AllRowColtemp       as long
  
  TempCUBRect         as RECT
  TempSelectRect      as RECT
  MultiCellRect       as RECT
  EditRect            as RECT

        
  iHscrollMax         AS LONG
  iHscrollPos         AS LONG
  xPos                AS LONG
  Dirty               AS LONG
 
  hFont               AS hFont

  ws                  AS INTEGER PTR
END TYPE

TYPE GridData                   'Type variable to hold private ListBox data
  hParent             AS HWND     'Can be freely customized to meet extended needs
  hInst               AS HINSTANCE
  id                  AS integer
  '
  hBit                AS hGDIOBJ
  memDC               AS hDC
  PrintDC             as hDC
  
  'Subclassed Procedures
  OriginalEditProc    as WndProc
  OriginalStaticProc  as WndProc
  OriginalComboProc   as WndProc
  
  TempButtonRECT      as RECT
  TabSheetPos         as LONG
  TabSheetPercent     as LONG
  
  'Child Control Handles
  hHScrollBar          as HWND
  hVScrollBar          as HWND
  hLRCorner            AS HWND 'static control
  hStatusBar           as HWND
  hSheetControl        as HWND
  hEditControl         AS HWND 
  hListControl         as HWND 
  
  'Focus
  FocusFlag            as Long
  
  
  'Status Bar
  StatusBarText1       as WSTRING * 128
  StatusBarText2       as WSTRING * 128
  StatusBarText3       as WSTRING * 128
  
  'Dropdown list hListControl
  ListStrings          AS INTEGER PTR   ' GRD_NUMLISTS     255
  
  'Colors
  CellColors(NUMCELLCOLORS) AS UINTeger
  RCHiLiteRGB               AS UINTeger    'High Light Row or Column
  BtnFaceBrushRGB           AS UINTeger
  HeaderSelectRGB           AS UINTeger
 
  'Flags and Toggles
  SetWindowPosFlag          as Long
  FullEditFlag              as Long
  ReturnKeyMode             as Long
  TabKeyMode                as Long
  UserButtonEffectFlag      as Long
  cyTweak                   as LONG
  HideGridLines             as LONG
  ComboEdit                 as long
  ShowStatusBar             as long
  SuppressResize            as LONG
  
  'Cursors
  SSCUR               AS hCursor
  DNCUR               AS hCursor
  LTCUR               AS hCursor
  TABCUR              AS hCursor
  
  'MouseMove Stuff
  CurrentCursor       as Long
  ColResizeFlag       as LONG
  ColResizeTemp       as LONG
  RowResizeFlag       as LONG
  RowResizeTemp       as LONG
  ColResizeChanging   as LONG
  ColResizeStart      as LONG
  ColResizeCol        as LONG
  RowResizeRow        as LONG
  RowResizeStart      as LONG
  RowResizeChanging   as LONG
  ShowResizeLine      as LONG
  Mousemovex          as long
  Mousemovey          as long
  
  'Brushes and Pens
  HeaderBrush         AS hGDIOBJ
  HeaderSelectBrush   AS hGDIOBJ
  CellBkColorBrush    AS hGDIOBJ
  WindowBKColorBrush  AS hGDIOBJ
  RCHiLiteBrush       AS hGDIOBJ
  BtnFaceBrush        AS hGDIOBJ
  CellLineColorPen    AS hGDIOBJ
  BlackLineColorPen   AS hGDIOBJ 
  SolidLine           AS hGDIOBJ
  SolidLine2          AS hGDIOBJ
  SolidLine3          AS hGDIOBJ
  
  'Formats
  HeaderHiliteStyle   as LONG  'emboss flat something else thru out worksheet
  NumberDateFormatString(GRD_NUMDATEFORMATSTR) AS WSTRING * 32
  DefaultDateInput    as long
  Bitmaps(GRD_NUMBITMAPS) as hBitmap
  
  'Fonts
  CellFonts(NUMGRDFONTS) as hGDIOBJ
  cyChar              as Long
  cyCharEx            as Long
  charWidth           as Long
  undo                as WString * 1000
  'Sheets
  TotalSheets         as LONG
  ActiveSheetUbound   as long 
  TabSheetTransport   as LONG
  TabSheetResizeFlag  as LONG
  TabSheetLeft        as LONG
  TabMoveMode         as LONG
  TempSheetRClickMenu as long
  
  'Menus
  RClickMenu(GRD_NUMMENUS) as hMenu
  ProcessRClickMenuFlag as long
  RightClickRow         as Long
  RightClickCol         as Long
  RightClickWhich       as Long
  'Sheet Stuff
  CS                  AS LONG  'Current Sheet Defaults to zero on startup
  SH(MAXGS)           as SheetData
END TYPE

'********************************************************************
' Forward Declares
'********************************************************************
'declare SUB GRD_SetSBar( hwnd AS HWND, BYVAL vhPage AS LONG, BYVAL vhMax AS LONG,BYVAL vhPos AS LONG, BYVAL vhBar AS LONG)
Declare Sub GRD_PutCell(hWnd As HWND,row As Long,col As Long, MyWstr As WString Ptr, MySheet As Long = -1)
Declare Sub GRD_GetCell(hWnd As HWND,row As Long,col As Long, MyWstr As WString Ptr, MySheet As Long = -1)
Declare Sub GRD_PutCellA(hWnd As HWND,row As Long,col As Long, Mystr As ZString Ptr, MySheet As Long = -1)
Declare Sub GRD_GetCellA(hWnd As HWND,row As Long,col As Long, Mystr As ZString Ptr, MySheet As Long = -1)
Declare Sub GRD_PutCellS(hWnd As HWND,row As Long,col As Long, Mystr As String, MySheet As Long = -1)
Declare Function GRD_GetCellS(hWnd As HWND,row As Long,col As Long, MySheet As Long = -1) as string
declare SUB GRD_CalcCellFormat(Hwnd as hwnd,myrow as long,mycol as long,Mycell as CELLFORMAT3 ptr)
declare Function GRD_FormatAttribute(hWnd as hWnd,myrow as long,mycol as long,Action as long,newvalue as LongInt=0,WhichSheet as long = -1) as ULongInt
declare FUNCTION GRD_MakeStandardFont(hWnd AS hWnd,TypeFace AS WSTRING Ptr,TheFontSize AS LONG,TheWeight AS LONG= FW_NORMAL,CharSet AS LONG=DEFAULT_CHARSET) AS hGDIOBJ
declare FUNCTION GRD_MakeStandardFontA(hWnd AS hWnd,TypeFace AS ZSTRING Ptr,TheFontSize AS LONG,TheWeight AS LONG= FW_NORMAL,CharSet AS LONG=DEFAULT_CHARSET) AS hGDIOBJ
declare FUNCTION GRD_MakeStandardFontS(hWnd AS hWnd,TypeFace AS STRING,TheFontSize AS LONG,TheWeight AS LONG= FW_NORMAL,CharSet AS LONG=DEFAULT_CHARSET) AS hGDIOBJ
declare FUNCTION GRD_SendNotifyMessage(BYVAL hWnd AS hWnd, BYVAL hCtrl AS hWnd, BYVAL CtrlID AS LONG, BYVAL NCode AS LONG,BYVAL MyParam1 AS LONG, BYVAL MyParam2 AS LONG, BYVAL MyParam3 AS LONG, BYVAL MyParam4 AS LONG) AS LONG
declare Sub GRD_GetListing(hWnd As HWND,WhichStr as long, MyWstr As WString Ptr)
declare FUNCTION ParseCountW(s as Wstring,del as wstring = ",") as long
declare SUB ParseW(s as Wstring, MyItem as long, MyWStr as WString Ptr,del as wstring = ",")
declare Function GRD_ActivateEditControl(Hwnd as Hwnd,Action as long = 0)as long

FUNCTION GRD_SendNotifyMessage(BYVAL hWnd AS hWnd, BYVAL hCtrl AS hWnd, BYVAL CtrlID AS LONG, BYVAL NCode AS LONG,BYVAL MyParam1 AS LONG, BYVAL MyParam2 AS LONG, BYVAL MyParam3 AS LONG, BYVAL MyParam4 AS LONG) AS LONG
  DIM NMG AS MyNotify

  NMG.NMHeader.hwndFrom=hCtrl
  NMG.NMHeader.idFrom=CtrlID
  NMG.NMHeader.code=NCode
  NMG.Param1=MyParam1
  NMG.Param2=MyParam2
  NMG.Param3=MyParam3
  NMG.Param4=MyParam4

  RETURN SendMessage(hWnd, WM_NOTIFY,cast(wParam,CtrlID), cast(lParam,VARPTR(NMG)))
  
END FUNCTION

          
 FUNCTION GRD_TextToSerialDate(HWND as HWND,MyRow as long, MyCol as long, istr AS WSTRING,OutString as WString Ptr) AS Long
  
 'Declare Function DateSerial ( ByVal year As Long, ByVal month As Long, ByVal day As Long ) As Long
 '#include "vbcompat.bi"
 'Dim a As Double = DateSerial(2005, 11, 28)
 'Print Format(a, "yyyy/mm/dd hh:mm:ss") 
   
    DIM x AS LONG
    DIM m AS LONG,d AS LONG, y AS LONG
    DIM del as WSTRING * 2
    Dim MySerial as LONG
    dim Mycount as LONG
    Dim MyWStr as wstring * 40
    dim MyWStrPtr as wstring ptr
    Dim MyType as long, MyFormat as Long,Mask as long 
    Dim v AS GridData PTR
    Dim MyWFormatStr as wstring * 40
   
   
    
   v = GetProp(hWnd,"GridData")
    
   if v <> 0 THEN
      
     
     mask = v->DefaultDateInput 
     if mask = 0 THEN  mask = 8
        
     MyType = GRD_FormatAttribute(HWND,MyRow,MyCol,GRDFORMAT_GETTYPE)
     MyFormat = GRD_FormatAttribute(HWND,MyRow,MyCol,GRDFORMAT_GETNDFORMAT)
    
     if MyType <> GRDFCT_DATE THEN exit function
     MyWFormatStr = v->NumberDateFormatString(MyFormat)
       
     MyWStrPtr = varPtr(MyWStr)
    
     SELECT CASE mask
           CASE MMDDYYYYS,MMDDYYYYD,MMDDYYYYL
               IF mask=MMDDYYYYS THEN del="/"
               IF mask=MMDDYYYYD THEN del="."
               IF mask=MMDDYYYYL THEN del="-"
               MyCount = ParsecountW(istr,del)
               if MyCount <> 3  THEN MySerial = 0 : Exit Function
               ParseW(istr,1,MyWStrPtr,del)  : m = val(MyWStr) 
               ParseW(istr,2,MyWStrPtr,del)  : d = val(MyWStr)
               ParseW(istr,3,MyWStrPtr,del)  : y = val(MyWStr)
             
           CASE DDMMYYYYS,DDMMYYYYD,DDMMYYYYL
              IF mask=DDMMYYYYS THEN del="/"
              IF mask=DDMMYYYYD THEN del="."
              IF mask=DDMMYYYYL THEN del="-"
              MyCount = ParsecountW(istr,del)
              if MyCount <> 3  THEN MySerial = 0 : Exit Function
              ParseW(istr,1,MyWStrPtr,del)  : d = val(MyWStr) 
              ParseW(istr,2,MyWStrPtr,del)  : m = val(MyWStr)
              ParseW(istr,3,MyWStrPtr,del)  : y = val(MyWStr)
                
           CASE MMYYYYS,MMYYYYD,MMYYYYL
              IF mask=MMYYYYS THEN del="/"
              IF mask=MMYYYYD THEN del="."
              IF mask=MMYYYYL THEN del="-"
              MyCount = ParsecountW(istr,del)
              if MyCount <> 2  THEN MySerial = 0 : Exit Function
              ParseW(istr,1,MyWStrPtr,del)  : m = val(MyWStr) 
              ParseW(istr,2,MyWStrPtr,del)  : y = val(MyWStr)
              d = 1
             
           CASE YYYYMMDDS,YYYYMMDDD,YYYYMMDDL
              IF mask=YYYYMMDDS THEN del="/"
              IF mask=YYYYMMDDD THEN del="."
              IF mask=YYYYMMDDL THEN del="-" 
              MyCount = ParsecountW(istr,del)
              if MyCount <> 3  THEN MySerial = 0 : Exit Function
              ParseW(istr,1,MyWStrPtr,del)  : y = val(MyWStr) 
              ParseW(istr,2,MyWStrPtr,del)  : m = val(MyWStr)
              ParseW(istr,3,MyWStrPtr,del)  : d = val(MyWStr)
              
           CASE YYYYMMS,YYYYMMD,YYYYMML
              IF mask=YYYYMMS THEN del="/"
              IF mask=YYYYMMD THEN del="."
              IF mask=YYYYMML THEN del="-"
              MyCount = ParsecountW(istr,del)
              if MyCount <> 2  THEN MySerial = 0 : Exit Function
              ParseW(istr,1,MyWStrPtr,del)  : y = val(MyWStr) 
              ParseW(istr,2,MyWStrPtr,del)  : m = val(MyWStr)
              d = 1
              
           case else 'YYYY.MM.DD
              del="."
              if MyCount <> 3  THEN MySerial = 0 : Exit Function
              ParseW(istr,1,MyWStrPtr,del)  : y = val(MyWStr) 
              ParseW(istr,2,MyWStrPtr,del)  : m = val(MyWStr)
              ParseW(istr,3,MyWStrPtr,del)  : d = val(MyWStr)  
             
        END SELECT 
        
        if m < 1 THEN m = 1 : if m > 12 THEN m = 12
        if d < 1 THEN d = 1 : if d > 31 THEN m = 31
          
        MySerial = DateSerial(y,m,d)
        if len(MyWFormatStr) > 0 THEN
             *OutString = Format(MySerial,MyWFormatStr)
          else  ' no formatting action
             *OutString = istr  
        END IF
       
   END IF 'v
      Function = MySerial
 end Function

'********************************************************************
' Set Scroll bars
'********************************************************************
SUB GRD_SetSBar( hwnd AS HWND, BYVAL vhPage AS LONG, BYVAL vhMax AS LONG,BYVAL vhPos AS LONG, BYVAL vhBar AS LONG)
  
  DIM si AS SCROLLINFO
  dim si_start as SCROLLINFO
  dim temp5 as Long
  DIM v AS GridData PTR

   v = GetProp(hWnd,"GridData")
   IF v <> 0 THEN
  
       'Get the current position
        si_start.cbSize = SIZEOF(si_start)
        si_start.fMask = SIF_POS
        GetScrollInfo(GetDlgItem(hwnd,vhBar), SB_CTL, @si_start)   

        'New position data
        si.cbSize = SIZEOF(si)
        si.fMask  = SIF_ALL OR SIF_DISABLENOSCROLL

        si.nMin = 1
      

        si.nMax = MAX(1, vhMax+1) 'the +1 gives room to clear the last row
        si.nPage  = vhPage
        si.nPos   = vhPos

        IF si_start.nPos <> vhPos THEN
          ' SetWindowPos v->hList, 0, 0, 0, 0, 0,SWP_HIDEWINDOW
        END IF

        SetScrollInfo(GetDlgItem(hwnd,vhBar), SB_CTL, @si,1)   
  end if 'v                                                        
END SUB

Function GRD_CheckCellChange(hwnd as hwnd) as Long
Function = 0   
END FUNCTION

FUNCTION GRD_SetSBarLimits(hWnd AS HWND) AS LONG
    Dim wRect AS RECT
    DIM temp as long,temp1 as long,temp2 as long
    DIM maxbottom as long,maxright as long
    dim oldbkbrush AS hGDIOBJ
    DIM v AS GridData PTR

   v = GetProp(hWnd,"GridData")
   IF v <> 0 THEN

           GetClientRect(hWnd, @wRect)
           
           GS_.TopRow = MAX(1, GS_.TopRow)
'           GS_.TopRow2 = MAX&(1, GS_.TopRow2)
 
           GS_.LeftCol = MAX&(1, GS_.LeftCol)
'           GS_.LeftCol2 = MAX&(1, GS_.LeftCol2)
 
           GS_.OldBottomRow = GS_.BottomRow
           GS_.OldRightCol = GS_.RightCol
'
'
           'Figure out the rightmost visible col
           temp=GRD_FormatAttribute(hwnd,0,0,GRDFORMAT_GETCOLWIDTH) + GS_.GridLeft
           GS_.RightCol=GS_.LeftCol
           FOR temp1=GS_.LeftCol TO GS_.TotCols
               IF temp1>GS_.TotCols THEN EXIT FOR
               temp=temp+GRD_FormatAttribute(hwnd,0,temp1,GRDFORMAT_GETCOLWIDTH)
               IF temp > wRect.Right THEN EXIT FOR
               GS_.RightCol = GS_.RightCol + 1
               GS_.RightCol=MIN(GS_.RightCol,GS_.TotCols)
           NEXT temp1
'
'          'Figure out the bottommost visible row
           temp=GRD_FormatAttribute(hwnd,0,0,GRDFORMAT_GETROWHEIGHT) + GS_.GridTop
    
           GS_.BottomRow=GS_.TopRow
           FOR temp1=GS_.TopRow TO GS_.TotRows
               IF temp1>GS_.TotRows THEN EXIT FOR
               temp=temp+GRD_FormatAttribute(hwnd,Temp1,0,GRDFORMAT_GETROWHEIGHT)
               IF temp > wRect.Bottom THEN EXIT FOR
               GS_.BottomRow = GS_.BottomRow + 1
               GS_.BottomRow=MIN(GS_.BottomRow,GS_.TotRows)
           NEXT temp1
'
'
           GS_.wMaxHeight = GS_.BottomRow - GS_.TopRow + 1
'           GS_.wMaxHeight2 = GS_.BottomRow2 - GS_.TopRow2 + 1
           GS_.wMaxWidth = GS_.LeftCol - GS_.RightCol + 1
'           GS_.wMaxWidth2 = GS_.LeftCol2 - GS_.RightCol2 + 1
'
           GRD_SetSBAR hwnd, 3, MAX(1,GS_.TotCols), GS_.LeftCol, GRD_HSBAR
           GRD_SetSBAR hwnd, (GS_.BottomRow - GS_.TopRow), MAX(1, GS_.TotRows), GS_.TopRow, GRD_VSBAR
'           GRD_SetSBAR hwnd, 3, MAX&(1,@v.TotCols), @v.LeftCol2, %MLG_HSBAR2
'           GRD_SetSBAR hwnd, @v.wMaxHeight2, MAX&(1, @v.TotRows), @v.TopRow2, %MLG_VSBAR2
'
'           'This will be changed to the desired background brush
            oldbkbrush= SelectObject(v->memDC, v->WindowBKColorBrush)
            PatBlt v->memDC, 0, 0, wRect.Right, wRect.Bottom, PATCOPY  'Paint with selected brush
'
      END IF 'v
      
      function = 0
END FUNCTION        

Function GRD_MySubClassEditProc(ByVal hWnd As HWND, ByVal uMsg As UINT, ByVal wParam As WPARAM, ByVal lParam As LPARAM) As Integer
 'Subclassed edit control
 Dim MyParent as HWND
 DIM v as GridData PTR
 dim myedittext as WString * GRD_MAXCELLLEN
 dim temp as Long, skey as LONG,MyType as LONG , MyChar as UBYTE
 dim del as WString * 30 , F as Long
 
 MyParent=GetParent(hWnd)
 v = GetProp(MyParent,"GridData")
 if v = 0 THEN
    exit function
 END IF
 
 
   
 Select Case uMsg
    
    
    
   case WM_SYSCOMMAND
     ' uMsg = NULL : wParam = 0
      
   case WM_CLOSE
     ' Function = 0
     ' Exit Function    

   Case WM_SETFOCUS 
    
   Case WM_KILLFOCUS
   
   Case EM_SETSEL
      
   case WM_CHAR
      
      if wParam = VK_TAB or wParam = VK_RETURN or wParam = VK_ESCAPE THEN
         Function = 0
         Exit Function
      END IF
       
      MyType = GRD_FormatAttribute(MyParent,GS_.SelRow,GS_.SelCol,GRDFORMAT_GETTYPE)
      F = v->DefaultDateInput
       
      if MyType = GRDFCT_NUMBER THEN
         IF instr( chr(wParam) ,any "+-.0123456789") OR WPARAM = 8  OR WPARAM = 13 THEN
               'Nothing for now
             else
              uMsg=Null : wParam=49 : Exit Function 
         end if   
      END IF
      
      if MyType = GRDFCT_DATE THEN
         if F = 0 or F = 2  or F = 5 or F = 8 or F = 11 or F = 14 then del = ".0123456789"
         if F = 1 or F = 4  or F = 7 or F = 10 or F = 13 then del = "/0123456789"
         if F = 3 or F = 6  or F = 9 or F = 12 or F = 15 then del = "-0123456789"
         IF instr( chr(wParam) ,any del) OR WPARAM = 8  OR WPARAM = 13 THEN
               'Nothing for now
             else
              uMsg=Null : wParam=49 : Exit Function 
         end if   
      END IF 
         
      
   Case WM_KEYDOWN
       SELECT CASE (wParam)
           case VK_TAB
              
              v->FullEditFlag = 0
              SetWindowPos v->hEditControl, 0, 0, 0, 0, 0,SWP_HIDEWINDOW OR SWP_NOSENDCHANGING
             
              GetWindowTextW v->hEditControl,varptr(myedittext),GRD_MAXCELLLEN    
              GRD_PutCell(MyParent,GS_.SelRow,GS_.SelCol,varptr(MyEditText))
              GRD_SendNotifyMessage v->hParent,MyParent,v->id,GRDN_DIRTY,GS_.SelRow,GS_.SelCol,0,0
              
              SendMessage MyParent, WM_KEYDOWN, VK_TAB, 0
              SetFocus MyParent

          case VK_RETURN
            
              SetWindowPos v->hEditControl, 0, 0, 0, 0, 0,SWP_HIDEWINDOW OR SWP_NOSENDCHANGING
             
              GetWindowTextW v->hEditControl,varptr(myedittext),GRD_MAXCELLLEN    
              GRD_PutCell(MyParent,GS_.SelRow,GS_.SelCol,varptr(MyEditText))
              GRD_SendNotifyMessage v->hParent,MyParent,v->id,GRDN_DIRTY,GS_.SelRow,GS_.SelCol,0,0
             
              SendMessage MyParent, WM_KEYDOWN, VK_RETURN, 0 
              SetFocus MyParent
       
           case VK_ESCAPE
              
              SetWindowPos v->hEditControl, 0, 0, 0, 0, 0,SWP_HIDEWINDOW OR SWP_NOSENDCHANGING
              
              SendMessage MyParent, WM_KEYDOWN, VK_ESCAPE, 0 
              SetFocus MyParent
                        
       end select
        
   Case WM_KEYUP, WM_LBUTTONUP
    
    
   CASE WM_GETDLGCODE
           FUNCTION = DLGC_WANTALLKEYS
           EXIT FUNCTION
          
 End Select

 Function = CallWindowProc(v->OriginalEditProc, hWnd, uMsg, wParam, lParam)

End Function

Function GRD_MySubClassComboProc(ByVal hWnd As HWND, ByVal uMsg As UINT, ByVal wParam As WPARAM, ByVal lParam As LPARAM) As Integer
 'Subclassed Combo control
 Dim MyParent as HWND
 DIM v as GridData PTR
 dim MyWStr as WString * GRD_MAXCELLLEN
 dim temp as Long, skey as LONG
 dim I as LONG
 
 MyParent=GetParent(hWnd)
 v = GetProp(MyParent,"GridData")
 
    Select Case uMsg
       
      case WM_SYSCOMMAND
        
         
      case WM_CLOSE
           

      Case WM_SETFOCUS 

       
      Case WM_KILLFOCUS

      
      Case WM_KEYDOWN
         
         print "List KeyDown"
         
         if wParam = VK_ESCAPE THEN
            SendMessage v->hListControl, LB_RESETCONTENT, 0, 0
            SetWindowPos v->hListControl, 0, 0, 0, 0, 0,SWP_HIDEWINDOW
            SetFocus MyParent
            SendMessage (MyParent,GRD_REFRESH, 0,0)
         END IF
         
         if wParam = VK_RETURN THEN
            I = SendMessageW (v->hListControl, LB_GETCURSEL, 0, 0)
            IF I = LB_ERR THEN 
               Function = 0
               EXIT Function
            end if
            SendMessageW (v->hListControl, LB_GETTEXT, I, cast(lParam,VARPTR(MyWStr)))
            GRD_PutCell(MyParent,GS_.SelRow,GS_.SelCol,varptr(MyWStr))
            SetWindowPos v->hListControl, 0, 0, 0, 0, 0,SWP_HIDEWINDOW
            SetFocus MyParent
            SendMessage (MyParent,GRD_REFRESH, 0,0)
         END IF
         
           
      Case WM_CHAR
         
     'if wParam = VK_RETURN or wParam = VK_ESCAPE THEN
         Function = 0
         Exit Function
     ' END IF
         

      Case WM_LBUTTONUP
         
         I = SendMessageW (v->hListControl, LB_GETCURSEL, 0, 0)
         IF I = LB_ERR THEN 
            Function = 0
            EXIT Function
         end if
         SendMessageW (v->hListControl, LB_GETTEXT, I, cast(lParam,VARPTR(MyWStr)))
         GRD_PutCell(MyParent,GS_.SelRow,GS_.SelCol,varptr(MyWStr))
         SetWindowPos v->hListControl, 0, 0, 0, 0, 0,SWP_HIDEWINDOW
         SetFocus MyParent
         SendMessage (MyParent,GRD_REFRESH, 0,0)
       
      CASE WM_GETDLGCODE
         
          FUNCTION = DLGC_WANTALLKEYS
          EXIT FUNCTION
                
    End Select

 Function = CallWindowProc(v->OriginalComboProc, hWnd, uMsg, wParam, lParam)

End Function



FUNCTION GRD_MySubClassStaticProc (BYVAL HWND AS HWND, BYVAL uMsg AS UInt,BYVAL wParam AS DWORD, BYVAL lParam AS LONG) AS INTEGER

 Dim MyParent as HWND
 DIM v as GridData PTR
 dim myedittext as WString * GRD_MAXCELLLEN
 dim temp as Long, skey as LONG
 DIM buffer AS wstring * 32
 DIM rc AS RECT
 DIM rd AS RECT
 DIM re AS RECT
 DIM MyTabPoint AS POINT
 DIM MyTabPointOld AS POINT
 dim counter as LONG
 dim CurrentSheet as LONG
 
 MyParent=GetParent(hWnd)
 v = GetProp(MyParent,"GridData")
 if v=0 THEN goto staticbail
 GetClientRect(hWnd,@rc)
 GetClientRect(MyParent,@rd)

    SELECT CASE uMsg

        CASE WM_RBUTTONDOWN
           
           IF LOWORD(lParam) > rc.Left + 64 AND LOWORD(lParam) < rc.Right - SHEETTABBUTTONWIDTH THEN
              temp = v->SH(cast(long,v->TabSheetLeft)).SheetOffset1 - 73
              FOR counter = v->TabSheetLeft TO MAXGS
                      IF LOWORD(lParam) < v->SH(counter).SheetOffset2 - temp AND LOWORD(lParam) > v->SH(counter).SheetOffset1 - temp  THEN
                          v->TempSheetRClickMenu = counter
                          EXIT FOR
                      END IF
              NEXT counter
            end if
           
            
        CASE WM_KILLFOCUS
            IF wParam = MyParent OR wParam=v->hListControl OR wParam=v->hEditControl THEN
             'The grid is still active 'do nothing
               v->FocusFlag=1
              ELSE
              'The grid is no longer active
               GRD_SendNotifyMessage v->hParent,MyParent,v->id,GRDN_KILLFOCUS,cast(integer,MyParent),0,0,0
            END IF

        CASE WM_SETFOCUS
            IF wParam = MyParent OR wParam=v->hListControl OR wParam=v->hEditControl THEN
              'The grid is still active - do nothing
               v->FocusFlag=1
              ELSE
               GRD_SendNotifyMessage v->hParent,MyParent,v->id,GRDN_SETFOCUS,cast(long,wParam),0,0,0
            END IF
'
        CASE WM_LBUTTONDOWN, WM_LBUTTONDBLCLK

'            'Resize Area
            IF LOWORD(lParam) > rc.Right - SHEETTABBUTTONWIDTH THEN
                SetCapture(hWnd)
                v->TabSheetResizeFlag =1
            END IF
'
'            'Transport area
            IF LOWORD(lParam) < rc.Left + 63 THEN
              v->TabSheetTransport=(LOWORD(lParam) \ 16) + 1

              IF v->TabSheetTransport = 1 THEN v->TabSheetLeft = 0
              IF v->TabSheetTransport = 4 THEN v->TabSheetLeft = v->ActiveSheetUBound
              IF v->TabSheetTransport = 2 THEN v->TabSheetLeft = v->TabSheetLeft - 1
              IF v->TabSheetTransport = 3 THEN v->TabSheetLeft = v->TabSheetLeft + 1
              IF v->TabSheetLeft < 0 THEN v->TabSheetLeft = 0
              IF v->TabSheetLeft > v->ActiveSheetUBound THEN v->TabSheetLeft = v->ActiveSheetUBound
 
              InvalidateRect hWnd, BYVAL NULL, 0
              UpdateWindow hwnd
              
            END IF
            
            'Tab Area
            IF LOWORD(lParam) > rc.Left + 64 AND LOWORD(lParam) < rc.Right - SHEETTABBUTTONWIDTH THEN
              CurrentSheet  = sendmessage(MyParent,GRD_GETCURRENTSHEET,0,0)
              
              temp = v->SH(cast(long,v->TabSheetLeft)).SheetOffset1 - 73
              FOR counter = v->TabSheetLeft TO MAXGS
                      IF LOWORD(lParam) < v->SH(counter).SheetOffset2 - temp AND LOWORD(lParam) > v->SH(counter).SheetOffset1 - temp  THEN
                          if counter <> CurrentSheet then SendMessage(MyParent,GRD_SETCURRENTSHEET,counter,0)
                          'Print counter
                          EXIT FOR
                      END IF
              NEXT counter
              

              InvalidateRect HWND, BYVAL NULL, 0
              UpdateWindow hwnd
              
              InvalidateRect MyParent, BYVAL NULL, 0
              UpdateWindow MyParent
'
            END IF 'If in Tab Area
'
'            'case %WM_Keydown
'
'
            CASE WM_CHAR

'
         CASE WM_LBUTTONUP
               if v->TabSheetResizeFlag =1 THEN
                ReleaseCapture
                v->TabSheetResizeFlag =0
              END IF
              're highlight the transport button
              if v->TabSheetTransport <> 0 then
                 v->TabSheetTransport = 0
                InvalidateRect hWnd, BYVAL NULL, 0
                UpdateWindow hwnd 
              end if   
'
'
        CASE WM_MOUSEMOVE
          GetClientRect(hWnd,@rc)
          GetClientRect(MyParent,@rd)
          MyTabPointOld.x = MyTabPoint.x
          MyTabPointOld.y = MyTabPoint.y
          MyTabPoint.x = Loword(lParam)
          MyTabPoint.y = HIword(lParam)
          
 '    
'         '
       IF v->TabMoveMode = 0 THEN

 '        Mypoint.x=LOWORD(lParam)
         IF MyTabpoint.x > rc.Right - SHEETTABBUTTONWIDTH THEN
           SetClassLongPtr HWND, GCLP_HCURSOR, cast(LONG_PTR,LoadCursor (NULL, BYVAL IDC_SIZEWE))
          ELSE
           SetClassLongPtr HWND, GCLP_HCURSOR, cast(LONG_PTR,LoadCursor (NULL, BYVAL IDC_ARROW))
         END IF
'
         IF MyTabpoint.x <> MyTabPointOld.x THEN
             IF v->TabSheetResizeFlag > 0 AND (wParam AND MK_LBUTTON)<> 0 THEN
                 v->TabSheetPos = MyTabPoint.x
                 v->TabSheetPos=MAX(v->TabSheetPos,70)
                 v->TabSheetPos=MIN(v->TabSheetPos,rd.right-40)
                 IF v->TabSheetPos < 0 THEN v->TabSheetPos = 0 '<<<<<------------------------ADDED
                 GetClientRect(MyParent,@re)
                 'Make TabSheetPercent
                 IF re.Right > 0 THEN v->TabSheetPercent=INT((v->TabSheetPos/(re.Right-0))* 1000)
                  'Move the Windows
                  MoveWindow( v->hSheetControl,0, rd.bottom-GS_.HScrollWidth-GS_.StatusBarWidth, v->TabSheetPos, GS_.HScrollWidth, FALSE )
                  MoveWindow( v->hHScrollBar,v->TabSheetPos , rd.bottom-GS_.HScrollWidth-GS_.StatusBarWidth, rd.right-v->TabSheetPos-GS_.VScrollWidth , GS_.HScrollWidth, TRUE ) 
                 'redraw the hSheetControl
                 InvalidateRect hWnd, BYVAL NULL, 0
                 UpdateWindow hwnd
             END IF
         END IF      
      END IF
    END SELECT

  StaticBail:
    FUNCTION = CallWindowProc(v->OriginalStaticProc, HWND, uMSG, WPARAM, LPARAM)
'
End function

'********************************************************************
' Intialize the a Grid Control when created
'********************************************************************
SUB GRD_Create(BYVAL hwnd AS HWND, BYVAL v as GridData PTR )
   DIM myfontsize as long
   DIM s as string * 75
   DIM lb AS logbrush
   Dim i as LONG
   Dim MyFont as WSTRING  * 40
   Dim MyFontPtr as WSTRING PTR
   
   MyFontPtr = varptr(MyFont)
   
   '
    v->TabSheetPos = 0
   v->TotalSheets=MAXGS
   for i= 0 to v->TotalSheets
     'v->SH(i).TabSheetPos = 0
     v->SH(i).HScrollWidth = SBSIZE
     v->SH(i).VScrollWidth = SBSIZE
     v->SH(i).StatusBarWidth = 0  'SBSIZE     'Hide the status bar to begin with
     v->SH(i).GridTop = 0
     v->SH(i).GridLeft = 0
   NEXT
   
   'Date and Number Format Strings
     v->NumberDateFormatString(0)=""
     v->NumberDateFormatString(1)="#.###"
     v->NumberDateFormatString(2)="#"
     v->NumberDateFormatString(3)="#,"
     v->NumberDateFormatString(4)="$* #,.00"
     v->NumberDateFormatString(5)="0.0%"
     v->NumberDateFormatString(6)=""
     v->NumberDateFormatString(7)=""
     v->NumberDateFormatString(8)=""
     v->NumberDateFormatString(9)=""
     v->NumberDateFormatString(10)=""
     v->NumberDateFormatString(11)=""
     v->NumberDateFormatString(12)=""
     v->NumberDateFormatString(13)=""
     v->NumberDateFormatString(14)=""
     v->NumberDateFormatString(15)=""     
     v->NumberDateFormatString(16)="ddd dd/MM/yyyy"
     v->NumberDateFormatString(17)="yyyy/mm/dd"
   
   
   
   'Colors  the FreeBasic macro has been renamed BGR instead of RGB to resolve a Windows conflict
   v->CellColors(0)  = BGR(255,255,255) ' Default White
   v->CellColors(1)  = BGR(0,0,0) 'Black
   v->CellColors(2)  = BGR(255,255,255) 'White
   v->CellColors(3)  = BGR(250,128,114) 'Light Red Salmon
   v->CellColors(4)  = BGR(0,128,0) 'Green
   v->CellColors(5)  = BGR(0,255,127) 'Light Green
   v->CellColors(6)  = BGR(0,0,128) 'Blue
   v->CellColors(7)  = BGR(175,238,238) 'Light Blue
   v->CellColors(8)  = BGR(255,0,255) 'Magenta
   v->CellColors(9)  = BGR(139,0,139) 'Dark white
   v->CellColors(10) = BGR(0,255,255) 'Cyan
   v->CellColors(11) = BGR(127,255,212) 'Light Cyan Aquamarine
   v->CellColors(12) = BGR(255,0,0) 'Red
   v->CellColors(13) = BGR(244,164,96) 'Brown
   v->CellColors(14) = BGR(255,255,0) 'Yellow
   v->CellColors(15) = BGR(255,127,80) 'Orange
   
   v->BtnFaceBrushRGB = GetSysColor(COLOR_BTNFACE) 
   v->RCHiLiteRGB=RGB(168,224,255)                   'High Light Row or Column even rows or columns                 
   v->HeaderSelectRGB=RGB(244,164,96)  'sort of brown           
   
   'Pens
   v->CellLineColorPen = CreatePen(PS_SOLID,1,GetSysColor(COLOR_SCROLLBAR))
   v->BlackLineColorPen = CreatePen(PS_SOLID,1,v->CellColors(1))
   v->SolidLine=CreatePen(PS_SOLID,1,RGB(0,0,0))
   v->SolidLine2=CreatePen(PS_SOLID,2,RGB(0,0,0))
   v->SolidLine3=CreatePen(PS_SOLID,3,RGB(0,0,0))
   
   
   'Brushes
   v->HeaderBrush = GetStockObject(LTGRAY_BRUSH)      
   v->CellBkColorBrush = GetStockObject(WHITE_BRUSH)   
   v->WindowBKColorBrush = GetStockObject(LTGRAY_BRUSH)  
    
   
   lb.lbstyle=BS_SOLID
   lb.lbcolor=v->HeaderSelectRGB 
   lb.lbhatch=0
   v->HeaderSelectBrush=CreateBrushIndirect(@lb)
   
   lb.lbcolor=v->RCHiLiteRGB
   v->RCHiLiteBrush=CreateBrushIndirect(@lb)    
  
   lb.lbcolor=v->BtnFaceBrushRGB
   v->BtnFaceBrush=CreateBrushIndirect(@lb) 
 
   
   'Fonts
   v->cyCharEx = 2   'start out with a couple of extra pixels of height
   GRD_MakeStandardFontS(HWND,"Arial",10)
  

'ScrollBars and other controls

       v->hHScrollBar   = CreateWindowEx (0,"scrollbar",_      'window class name
                       "",_                'window caption
                       WS_CHILD OR WS_VISIBLE OR SBS_HORZ OR WS_CLIPSIBLINGS,_  'window style
                       0, 0, 0, 0,_              'initial y
                       hWnd,_                 'parent window handle
                       cast(hMENU,GRD_HSBAR),_               'window menu handle
                       v->hInst,_                  'program instance handle
                       BYVAL NULL)
 
        v->hVScrollBar   = CreateWindowEx (0,"scrollbar",_      'window class name
                       "",_                'window caption
                       WS_CHILD OR WS_VISIBLE OR SBS_VERT OR WS_CLIPSIBLINGS,_  'window style
                       0, 0,0, 0,_              'initial y
                       hWnd,_                 'parent window handle
                       cast(hMENU,GRD_VSBAR),_                          'window menu handle
                       v->hInst,_                  'program instance handle
                       BYVAL NULL)
 
       v->hSheetControl = CreateWindowEx(WS_EX_LEFT OR WS_EX_LTRREADING, _
                                 "static", _  ' window class name
                                 "", _    ' window caption
                                 WS_CHILD OR WS_VISIBLE OR SS_OWNERDRAW _
                                 OR SS_LEFT OR SS_NOTIFY, _
                                 0, 0,0, 0, _
                                 HWND, cast(hMENU,GRD_SCNTL), v->hInst, BYVAL Null)
                                 
       v->hStatusBar = CreateWindowEx(WS_EX_LEFT OR WS_EX_LTRREADING or WS_EX_STATICEDGE, _
                                 "static", _  ' window class name
                                 "", _    ' window caption
                                 WS_CHILD OR WS_VISIBLE OR SS_OWNERDRAW _
                                 OR SS_LEFT OR SS_NOTIFY or WS_BORDER, _
                                 0, 0,0, 0, _
                                 HWND, cast(hMENU,GRD_STATB), v->hInst, BYVAL Null)                          
  
       v->hLRCorner= CreateWindowEx(WS_EX_LEFT OR WS_EX_LTRREADING, _
                                 "static", _  ' window class name
                                 "", _    ' window caption
                                 WS_CHILD OR WS_VISIBLE OR WS_CLIPSIBLINGS OR WS_CLIPCHILDREN _
                                 OR SS_LEFT, _
                                 0, 0,0, 0, _
                                 HWND, cast(hMENU,GRD_LRCOR), v->hInst, BYVAL Null)
 
 
       v->hListControl = CreateWindowW ("listbox",_  'window class name
                               BYVAL NULL,_   'window caption
                               LBS_NOTIFY OR WS_CHILDWINDOW OR WS_BORDER OR WS_VSCROLL, _
                               100,_ 'initial x position
                               100,_ 'initial y position
                               100,_  'initial x size
                               50,_   'initial y
                               HWND,_ 'parent window handle
                               cast(hMENU,GRD_LCNTL),_
                               v->hInst,_  'program instance handle
                               BYVAL NULL)
 
      v->hEditControl = CreateWindowW("EDIT", _
                         BYVAL NULL, _
                        WS_CHILD OR ES_AUTOHSCROLL OR ES_OEMCONVERT OR WS_CLIPSIBLINGS OR WS_CLIPCHILDREN , _
                        0, 0, 0, 0, _
                        HWND, cast(hMENU,GRD_ECNTL), v->hInst, BYVAL Null)
  
     'Subclass the edit part of the ComboBox     
     v->OriginalEditProc = Cast(WndProc,SetWindowLongPtr(v->hEditControl, GWLP_WNDPROC, Cast(LONG_PTR, ProcPtr(GRD_MySubClassEditProc))))
     v->OriginalComboProc = Cast(WndProc,SetWindowLongPtr(v->hListControl, GWLP_WNDPROC, Cast(LONG_PTR, ProcPtr(GRD_MySubClassComboProc)))) 
     'Subclass for the Tab Sheet Control which is an owner drawn static control
     v->OriginalStaticProc = Cast(WndProc,SetWindowLongPtr(v->hSheetControl, GWLP_WNDPROC, Cast(LONG_PTR, ProcPtr(GRD_MySubClassStaticProc))))

END SUB

'Painting

SUB GRD_DrawTextVert(hWnd AS HWND,byval MyWstrPtr AS WSTRING PTR,whichfont AS long,byref temprc AS RECT PTR,just AS LONG)
DIM Xpos AS LONG
DIM Ypos AS LONG
DIM slen AS LONG
DIM sdc as integer
dim hdc AS HDC
DIM myfont AS hGDIOBJ
DIM v AS GridData PTR
DIM mysize AS SIZEL
DIM result AS LONG
dim rc as RECT
      
      CopyMemory(varptr(rc),temprc,sizeof(RECT))
     
      slen = LEN(*MyWstrPtr)
    
      v = GetProp(hWnd,"GridData")
      IF v <> 0 THEN
        hDC = v->memDC
        SDC = SaveDC(HDC)
        IntersectClipRect(HDC, RC.LEFT, RC.TOP, RC.RIGHT, RC.BOTTOM)
        
        IF whichfont = GRDSYSFONTBOLDVERT THEN
            myfont = v->CellFonts(GRDSYSFONTBOLDVERT)
          ELSE
            myfont = v->CellFonts(GRDSYSFONTNORMVERT)
        END IF
        
        result=GetTextExtentPoint32W(hDC,BYVAL MyWstrPtr,slen,@mysize) 
        SelectObject hDC,myfont
  
        IF just = GRDFCJ_LEFT THEN 'Left
              Xpos = (Rc.Left) + 0  '(mysize.cy) + 2
              Ypos = Rc.Bottom - 2
            ELSEIF just = GRDFCJ_RGHT THEN 'Right
              Xpos = (Rc.Right)-(mysize.cy) - 2
              Ypos = Rc.Bottom - 2
            ELSEIF GRDFCJ_CNTR THEN 'Center
              Xpos = ((Rc.Left + Rc.Right)\2) - ((mysize.cy)\2) + 2
              Ypos = Rc.Bottom - 2
            ELSE 'revert to left
              Xpos = (Rc.Left) + 0  '(mysize.cy) + 2
              Ypos = Rc.Bottom - 2
        END IF
       
        TextOutW(HDC, Xpos, Ypos, MyWstrPtr, slen)
        RestoreDC(HDC, SDC)
        
    END IF 'v
END SUB          

SUB GRD_PaintCells(BYVAL hwnd AS HWND, BYVAL MyQuad as long = 0 )
   DIM i as long,MyRow as long , MyCol as long , z as long,tempwidth as LONG,tempheight as LONG
   DIM y as long,TheElement as LONG, MaxRightCol as long,MaxBottomRow as long
   dim RowHeaderWidth as LONG, ColHeaderHeight as LONG
   
   dim MywStr as WString * 1000
   dim DrawRect as RECT
   dim GridRect as RECT
   dim CBRect as RECT ' Checkbox RECT
   DIM s as STRING
   Dim MywStrPtr as WSTRING ptr
   Dim MyCell as CELLFORMAT3
   Dim MyRowCell as ROWFORMAT1
   Dim MyColCell as COLFORMAT2
   Dim Corn as CORNERFORMAT0
   Dim str2 as wstring ptr
   Dim MyColor as ULongint
   Dim LeftColFlag as LONG, SelectionColor as UINTEGER, HeaderColor as UInteger, CellColor as UINTEGER
   dim TopRowFlag as LONG
   Dim lb as LogBrush
   Dim TempBrush as hGDIOBJ , SelectionBrush as hGDIOBJ
   dim tempx as long, tempy as LONG , tempdt as LONG, temptype as LONG,tempdtsaved as LONG
   dim MyPoint as point
   Dim v As GridData Ptr
   dim temp1 as long, temp2 as Long, temp3 as Long , temp4 as Long, temp5 as Long
   dim temp6 as long, x as long, whichfont as Long
   dim MyBorder as UByte
   dim GridWidth as long, GridHeight as Long,bbbRow as long,  bbbCol as long
   dim temp1GDI as hGDIOBJ, temp2GDI as hGDIOBJ
   Dim tempDouble as DOUBLE
   dim myRgnRect as hRgn
   dim MyWOutStr as wstring * 50
   
   
                      '  Print "Draw Cells"
               
   
 
   MywStrPtr=varptr(MywStr)
   v = GetProp(hWnd,"GridData")
   
   GetClientRect(HWND,@GridRect)
   
   if v <> 0 then
      
        'ExcludeClipRect hDC,L,T,R,B
      ExcludeClipRect(v->memDC,0,GridRect.Bottom - GS_.HScrollWidth-GS_.StatusBarWidth,GridRect.Right,GridRect.Bottom)'Bottom
      ExcludeClipRect(v->memDC,GridRect.Right - GS_.VScrollWidth,0,GridRect.Right,GridRect.Bottom) 'Right
      ExcludeClipRect(v->memDC,0,0,GS_.GridLeft,GridRect.Bottom)'LeftHeader
      ExcludeClipRect(v->memDC,0,0,GridRect.Right,GS_.GridTop)'TopHeader
      
      
      
       i=0 : RowHeaderWidth=GRD_FormatAttribute(hwnd,0,0,GRDFORMAT_GETCOLWIDTH)
       y=RowHeaderWidth
       for i = GS_.LeftCol to GS_.DimCols
         y= y + GRD_FormatAttribute(hwnd,0,i,GRDFORMAT_GETCOLWIDTH)
         if y > gridrect.right THEN MaxRightCol = i : exit for
         if i=GS_.DimCols then MaxRightCol = i
       next i
       
       i=0 : ColHeaderHeight=GRD_FormatAttribute(hwnd,0,0,GRDFORMAT_GETROWHEIGHT)
       y = ColHeaderHeight
       for i = GS_.TopRow to  GS_.DimRows
         y= y+ GRD_FormatAttribute(hwnd,0,i,GRDFORMAT_GETROWHEIGHT) 
         if y > gridrect.bottom THEN MaxBottomRow = i : exit for
         if i=GS_.DimRows then MaxBottomRow = i
       next i
     
      
      Dim ttt as GRDArray Pointer = cast(any Ptr,GS_.ws)
      if ttt = 0 THEN
         exit sub  
      END IF
      
      'Create the write ahead border array
      GridWidth = GS_.BottomRow - GS_.TopRow + 10  : GridHeight = GS_.RightCol - GS_.LeftCol + 5
    
      Dim bbb(GridWidth,GridHeight) as UByte
      
      'Get the CornerCell Format Header
      str2=ttt[0].ws  
      CopyMemory(@corn,str2,HEADERBYTES)
      
      HeaderColor = &HD3D3D3  'LTGRAY
      SelectionColor = &HE6D8AD  'Light Blue
      lb.lbstyle=BS_SOLID
      lb.lbcolor=SelectionColor 
      SelectionBrush=CreateBrushIndirect(@lb)
      
      SelectObject v->memDC,v->CellFonts(GRDSYSFONTBOLD)
 
      'Draw the corner
        DrawRect.left = GS_.GridLeft  :  DrawRect.right = DrawRect.left + corn.MyWidth
        DrawRect.top = GS_.GridTop  :  DrawRect.bottom = DrawRect.top + corn.MyHeight
        IF corn.MyBackColor = 0 THEN
                 lb.lbstyle=BS_SOLID
                 lb.lbcolor=HeaderColor
                 TempBrush=CreateBrushIndirect(@lb)
                 SetBkColor v->memDC, HeaderColor 
               else
                 'make a brush
                 lb.lbstyle=BS_SOLID
                 lb.lbcolor=v->CellColors(corn.MyBackColor)
                 TempBrush=CreateBrushIndirect(@lb)
                 SetBkColor v->memDC, v->CellColors(corn.MyBackColor)
                 HeaderColor = v->CellColors(corn.MyBackColor)
      END IF
      SelectObject v->memDC, GetStockObject(BLACK_PEN)
      SelectObject v->memDC, TempBrush
      FillRect v->memDC,@DrawRect,TempBrush 
      FrameRect(v->memDC,@DrawRect,GetStockObject(BLACK_BRUSH))
      tempx = DrawRect.right : tempy=DrawRect.bottom
      
      'Draw the Column Header
      for i = GS_.LeftCol to GS_.RightCol
         'print i
         DrawRect.left = DrawRect.right-1  :  DrawRect.right = DrawRect.left + GRD_FormatAttribute(hwnd,0,i,GRDFORMAT_GETCOLWIDTH)
         GRD_GetCell(HWND,0,i,varptr(MyWStr))
         FillRect v->memDC,@DrawRect,TempBrush
         if i = GS_.SelCol THEN
            FillRect v->memDC,@DrawRect,SelectionBrush
            SetBkColor v->memDC, SelectionColor
          else  
            FillRect v->memDC,@DrawRect,TempBrush
            SetBkColor v->memDC, HeaderColor
         end if
         if instr(MyWStr,GRD_CRLF) > 0 THEN 
              DrawTextW v->memDC,BYVAL MywStrPtr, len(MyWStr),@DrawRect,DT_CENTER OR DT_NOPREFIX OR DT_BOTTOM or DT_WORDBREAK or DT_NOFULLWIDTHCHARBREAK
            else    
              DrawTextW v->memDC,BYVAL MywStrPtr, len(MyWStr),@DrawRect,DT_CENTER OR DT_NOPREFIX OR DT_BOTTOM or DT_SINGLELINE 'DT_WORDBREAK
         end if
         FrameRect(v->memDC,@DrawRect,GetStockObject(BLACK_BRUSH))
         DrawRect.left = DrawRect.left + 1
      NEXT i
     'Draw the row header
      DrawRect.left = GS_.GridLeft  :  DrawRect.right = tempx
      DrawRect.top = GS_.GridTop    :  DrawRect.bottom = tempy
      for i = GS_.TopRow to GS_.BottomRow
         DrawRect.top = DrawRect.bottom -1 :  DrawRect.bottom = DrawRect.bottom + GRD_FormatAttribute(hwnd,i,0,GRDFORMAT_GETROWHEIGHT)
         GRD_GetCell(HWND,i,0,varptr(MyWStr))
         if i = GS_.SelRow THEN
            FillRect v->memDC,@DrawRect,SelectionBrush
            SetBkColor v->memDC, SelectionColor
          else  
            FillRect v->memDC,@DrawRect,TempBrush
            SetBkColor v->memDC, HeaderColor
         end if 
         DrawRect.Right = DrawRect.Right - 3 
         if instr(MyWStr,GRD_CRLF) > 0 THEN
              DrawTextW v->memDC,BYVAL MywStrPtr, len(MyWStr),@DrawRect,DT_RIGHT OR DT_NOPREFIX OR DT_BOTTOM or DT_WORDBREAK or DT_NOFULLWIDTHCHARBREAK
            Else  
              DrawTextW v->memDC,BYVAL MywStrPtr, len(MyWStr),@DrawRect,DT_RIGHT OR DT_NOPREFIX OR DT_BOTTOM or DT_SINGLELINE
         END IF
         DrawRect.Right = DrawRect.Right + 3 
         FrameRect(v->memDC,@DrawRect,GetStockObject(BLACK_BRUSH))
         DrawRect.top = DrawRect.top + 1
      NEXT i 
      DeleteObject Tempbrush
      DeleteObject SelectionBrush
  
     'Draw the cells
     'Start with the corner cell
     DrawRect.left = GS_.GridLeft  :  DrawRect.right = DrawRect.left + corn.MyWidth
     DrawRect.top = GS_.GridTop  :  DrawRect.bottom = DrawRect.top + corn.MyHeight
   
     For MyRow = GS_.TopRow  to Min(GS_.BottomRow + 5,GS_.TotRows)
       bbbRow = MyRow - GS_.TopRow 'border write ahead array
       
       DrawRect.top = DrawRect.bottom -1  : DrawRect.bottom = DrawRect.bottom + GRD_FormatAttribute(HWnd,MyRow,0,GRDFORMAT_GETROWHEIGHT)
       DrawRect.left = GS_.GridLeft  :  DrawRect.right = DrawRect.left + corn.MyWidth
       For MyCol = GS_.LeftCol to Min(GS_.RightCol + 3,GS_.TotCols)
            bbbCol = MyCol-GS_.LeftCol  'border write ahead array
            
            DrawRect.left = DrawRect.right-1 : DrawRect.right= DrawRect.left + GRD_FormatAttribute(HWnd,0,MyCol,GRDFORMAT_GETCOLWIDTH)
               
            TheElement = (Myrow * (GS_.DimCols+1) ) + (Mycol)
            MywStr=""
            GRD_GetCell(hwnd,Myrow,MyCol,MywStrPtr)
            GRD_CalcCellFormat(Hwnd,Myrow,Mycol,@Mycell)

             'CellType
               temptype=GRDFCT_HORZ
               if MyCell.CellType > GRDFCT_AUTO then
                  if MyCell.CellType < GRDFCT_USERBUT + 1 THEN
                     temptype=MyCell.CellType
                  END IF
               End if
               
             'CellFormat
               if MyCell.CellFormat = GRDFCT_NUMBER then
               
                  ELSEIF MyCell.MyJust = GRDFCT_DATE THEN
         
               End if   
            
             'Justify
            Select CASE MyCell.MyJust
               case GRDFCJ_AUTO
                  tempdt=DT_LEFT + DT_BOTTOM
               case GRDFCJ_LEFT
                  tempdt=DT_LEFT + DT_BOTTOM
               case GRDFCJ_RGHT
                  tempdt=DT_RIGHT + DT_BOTTOM
               case GRDFCJ_CNTR
                  tempdt=DT_CENTER + DT_BOTTOM
               case GRDFCJ_LEFT_T
                  tempdt=DT_LEFT + DT_TOP
               case GRDFCJ_RGHT_T
                  tempdt=DT_RIGHT + DT_TOP
               case GRDFCJ_CNTR_T
                  tempdt=DT_CENTER + DT_TOP
               case GRDFCJ_LEFT_C
                  tempdt=DT_LEFT + DT_VCENTER
               case GRDFCJ_RGHT_C
                  tempdt=DT_RIGHT + DT_VCENTER
               case GRDFCJ_CNTR_C                           
                  tempdt=DT_CENTER + DT_VCENTER 
               CASE ELSE
                  tempdt=DT_LEFT + DT_BOTTOM
            END SELECT

              'Case
              IF MyCell.MyCase > GRDFCC_AUTO THEN
                   IF MyCell.MyCase=GRDFCC_UPPR THEN
                       MyWStr=UCASE(MyWStr)
                     ELSEIF MyCell.MyCase=GRDFCC_LOWR THEN
                       MyWStr=LCASE(MyWStr)
                     ELSE  'must be GRDFCC_TITL
                      ' MyWStr=MCASE(MyWStr) 'not programmed yet
                   END IF
               END IF

             'Font
            
            if MyCell.MyFont = 0 THEN
                SelectObject v->memDC,v->CellFonts(GRDSYSFONTDEFAULT)
              else
                SelectObject v->memDC,v->CellFonts(MyCell.MyFont)
            END IF
            
          
            'Fore Color
            if MyCell.MyForeColor = 0 THEN
                SetTextColor v->memDC, v->CellColors(CELLCOLORBLACK)
              else
                SetTextColor v->memDC, v->CellColors(MyCell.MyForeColor)
            END IF
            
            'Background Color
            IF MyCell.MyBackColor = 0 THEN
                 FillRect v->memDC,@DrawRect,GetStockObject(WHITE_BRUSH)
                 SetBkColor v->memDC, v->CellColors(CELLCOLORWHITE)
                 CellColor = v->CellColors(CELLCOLORWHITE)
               else
                 'make a brush
                 lb.lbstyle=BS_SOLID
                 lb.lbcolor=v->CellColors(MyCell.MyBackColor)
                 TempBrush=CreateBrushIndirect(@lb)
                 FillRect v->memDC,@DrawRect,TempBrush
                 SetBkColor v->memDC, v->CellColors(MyCell.MyBackColor)
                 CellColor = v->CellColors(MyCell.MyBackColor)
                 DeleteObject TempBrush
            END IF
            
            SELECT CASE MyCell.CellType
               case GRDFCT_PICTURE                     
                  if MyCell.CellFormat = 0 THEN
                  
                     'GRD_SendNotifyMessage v->hParent,hWnd,v->id,GRDN_DRAWYOUROWN,cast(long,v->PrintDC),cast(RECT PTR,@DrawRect),MyRow,MyCol 
                   else
                     SelectObject(v->PrintDC, v->Bitmaps(MyCell.CellFormat)) 
                     BitBlt v->memDC, DrawRect.Left, DrawRect.Top, DrawRect.Right-DrawRect.Left,DrawRect.Bottom-DrawRect.Top,v->PrintDC, 0, 0, SRCCOPY '
                  END IF
                  if v->HideGridLines = 0 then FrameRect(v->memDC,@DrawRect,GetStockObject(LTGRAY_BRUSH))
                
        
               case GRDFCT_HORZ,GRDFCT_AUTO,GRDFCT_HORZWW,GRDFCT_NUMBER,GRDFCT_DATE
                  
                  if MyCell.CellType=GRDFCT_HORZWW  THEN
                     
                  END IF
                  
                  if MyCell.CellType=GRDFCT_NUMBER  THEN
                      if len(MyWStr) > 0 THEN
                         tempDouble = val(MyWStr)
                         MyWStr = Format(tempDouble,v->NumberDateFormatString(MyCell.CellFormat))
                      END IF
                  END IF
                  
                  if MyCell.CellType=GRDFCT_DATE  THEN
                      
                      if len(MyWStr) > 0 THEN
                         GRD_TextToSerialDate(HWND,MyRow,MyCol,MyWStr,VarPtr(MyWOutStr))
                         MyWStr = MyWOutStr
                      END IF
                  END IF
                  
                  if MyCell.CellType=GRDFCT_HORZWW  THEN
                     if instr(MyWStr,GRD_CRLF) > 0 THEN 
                        tempdt = tempdt + DT_WORDBREAK
                     else
                        tempdt = tempdt + DT_SINGLELINE
                      end if  
                  END IF
                  
                  if MyCell.CellType=GRDFCT_HORZ or MyCell.CellType=GRDFCT_AUTO  THEN
                    tempdt = tempdt + DT_SINGLELINE 
                  END IF
                  
                  tempdtsaved = tempdt
                  
                 FillRect v->memDC,@DrawRect,TempBrush 
                 DrawRect.Left = DrawRect.Left + 3   
                 DrawTextW v->memDC,BYVAL MywStrPtr, len(MyWStr),@DrawRect, tempdt OR DT_NOPREFIX
                 DrawRect.Left = DrawRect.Left - 3 
                 if v->HideGridLines = 0 then FrameRect(v->memDC,@DrawRect,GetStockObject(LTGRAY_BRUSH)) 
                 
                
                  
               case GRDFCT_VERT,GRDFCT_VERTWW    'Text Vertical No word wrap
                
                 if MyCell.MyFont = 0 THEN
                      whichfont = 6
                    else
                      whichfont = 7
                  END IF 
                  FillRect v->memDC,@DrawRect,TempBrush
                
                  tempdt = tempdt + DT_SINGLELINE 
                  GRD_DrawTextVert(Hwnd, Byval MyWstrPtr,GRDSYSFONTNORMVERT,@DrawRect,tempdt)
                 
                  if v->HideGridLines = 0 then FrameRect(v->memDC,@DrawRect,GetStockObject(LTGRAY_BRUSH))  
               case GRDFCT_CKBOX   'Checkbox
                 temp4 = GRD_FormatAttribute(hwnd,MyRow,0,GRDFORMAT_GETROWHEIGHT) 
                 temp1 = (DrawRect.Bottom - DrawRect.Top)\2 + Drawrect.Top     'get the middle Vertical
                 temp2 = (DrawRect.Right - DrawRect.Left)\2 + DrawRect.Left    'get the middle Horizontal
                 temp3 = 7
                
                 IF temp4 < 17 THEN temp3 = 6
                 IF temp4 < 14 THEN temp3 = 5
                 IF temp4 < 12 THEN temp3 = 4
                 IF temp4 < 9  THEN temp3 = 3
                           
                 CBRect.Top = temp1 - temp3  : CBRect.Bottom = temp1 + temp3 
                 CBRect.Left = temp2 - temp3 : CBRect.Right = temp2 + temp3
                  
                 FillRect v->memDC,@DrawRect,TempBrush 
                 if len(MyWStr) = 0 THEN
                     FrameRect(v->memDC,@CBRect,GetStockObject(BLACK_BRUSH))
                   else 'check the box
                     SelectObject v->memDC, v->SolidLine2
                     FrameRect(v->memDC,@CBRect,GetStockObject(BLACK_BRUSH))
                     MoveToEx(v->memDC,CBRect.Left,CBRect.Top,@mypoint) :  LineTo(v->memDC,CBRect.Right-1,CBRect.Bottom-1)
                     MoveToEx(v->memDC,CBRect.Right-1,CBRect.Top,@mypoint) :  LineTo(v->memDC,CBRect.Left,CBRect.Bottom-1)
                 END IF
                 if v->HideGridLines = 0 then FrameRect(v->memDC,@DrawRect,GetStockObject(LTGRAY_BRUSH))
                 SelectObject v->memDC, GetStockObject(BLACK_PEN)
                 
                  
               case GRDFCT_DDCBS,GRDFCT_USERBUT    'Dropdown Combo or User Button
                  
                 temp4 = GRD_FormatAttribute(hwnd,MyRow,0,GRDFORMAT_GETROWHEIGHT) 
                 temp5 = GRD_FormatAttribute(hwnd,0,MyCol,GRDFORMAT_GETCOLWIDTH) 
                 temp3 = 17 'Button Height
                
                 IF temp4 < 17 THEN temp3 = 17
                 IF temp4 < 15 THEN temp3 = 13
                 IF temp4 < 13 THEN temp3 = 11
                 IF temp4 < 11 THEN temp3 = 9
                 IF temp4 < 9  THEN temp3 = 7
                 
                 FillRect v->memDC,@DrawRect,TempBrush 
               
                 v->TempButtonRect.Top = DrawRect.Top + temp4 - temp3 : v->TempButtonRect.Bottom = v->TempButtonRect.Top + temp3 - 1
                 v->TempButtonRect.Left =DrawRect.Left + temp5 - temp3 - 1 : v->TempButtonRect.Right = v->TempButtonRect.Left + temp3 - 1
                 if MyCell.CellType = GRDFCT_USERBUT then  
                    v->TempButtonRect.Left =DrawRect.Left + 2
                    v->TempButtonRect.Top =DrawRect.Top + 2
                 end if   
                 RoundRect(v->memDC, v->TempButtonRect.Left,v->TempButtonRect.Top,v->TempButtonRect.Right,v->TempButtonRect.Bottom,2,2 )
                 if  v->UserButtonEffectFlag <> 0 THEN
                    if LOWORD(v->UserButtonEffectFlag) = MyRow and HIWORD(v->UserButtonEffectFlag)= MyCol THEN
                        FillRect v->memDC,@v->TempButtonRect,GetStockObject(LTGRAY_BRUSH)'Pushed down effect
                        SetBkColor v->memDC,&HD3D3D3
                      else
                        FillRect v->memDC,@v->TempButtonRect,v->BtnFaceBrush
                        SetBkColor v->memDC,v->BtnFaceBrushRGB   
                    END IF
                  else
                     FillRect v->memDC,@v->TempButtonRect,v->BtnFaceBrush
                     SetBkColor v->memDC,v->BtnFaceBrushRGB
                 END IF
                 DrawEdge (v->memDC, @v->TempButtonRect, EDGE_RAISED, BF_RECT)
                
                 if MyCell.CellType = GRDFCT_DDCBS then   'Dropdown Comboxbox Edit
                        x = ((v->TempButtonRect.Right - v->TempButtonRect.Left)\2)-2
                        y = ((v->TempButtonRect.Bottom - v->TempButtonRect.Top)\2)-1
                       
                        SelectObject v->memDC, GetStockObject(BLACK_PEN)
                        
                        FOR temp6 = 0 TO 2
                          MoveToEx(v->memDC,v->TempButtonRect.Left + x + temp6,v->TempButtonRect.Top + y + temp6,@mypoint)
                          LineTo(v->memDC,v->TempButtonRect.Left + x + 5 - temp6,v->TempButtonRect.Top + y + temp6)
                        NEXT temp6
                        SetBkColor v->memDC, CellColor
                        DrawRect.Left = DrawRect.Left + 3 : DrawRect.Right = DrawRect.Right - x  
                        DrawTextW v->memDC,BYVAL MywStrPtr, len(MyWStr),@DrawRect, tempdtsaved OR DT_NOPREFIX
                        DrawRect.Left = DrawRect.Left - 3  : DrawRect.Right = DrawRect.Right + x 
                        
                     else 'User Button
                          tempdt=DT_CENTER
                          'SetBkColor v->memDC,v->BtnFaceBrushRGB 
                          DrawRect.Left = DrawRect.Left + 3 
                          DrawTextW v->memDC,BYVAL MywStrPtr, len(MyWStr),@DrawRect, tempdt OR DT_NOPREFIX OR DT_VCENTER or DT_SINGLELINE
                          DrawRect.Left = DrawRect.Left - 3
                 end if
                 
                if v->HideGridLines = 0 then FrameRect(v->memDC,@DrawRect,GetStockObject(LTGRAY_BRUSH))     
               CASE ELSE
                  'Nothing for now   
            END SELECT
               'Because of overdrawing - the logic is simpler to restrict to top and left for now
               if MyCell.MyBorder > 0 THEN 'or bbb(bbbRow,bbbCol) > 0
                  SelectObject v->memDC, GetStockObject(BLACK_PEN)
                  MyBorder = MyCell.MyBorder
                  if (MyBorder and GRD_TOP)  THEN  MoveToEx(v->memDC,DrawRect.Left,DrawRect.Top,@mypoint) :  LineTo(v->memDC,DrawRect.Right-1,DrawRect.Top)
                  if (MyBorder and GRD_LEFT)  THEN  MoveToEx(v->memDC,DrawRect.Left,DrawRect.Top,@mypoint) :  LineTo(v->memDC,DrawRect.Left,DrawRect.bottom)
                  if (MyBorder and GRD_BOTTOM) then bbb(bbbRow + 1,bbbCol)= bbb(bbbRow + 1,bbbCol) or GRD_TOP 
                  if (MyBorder and GRD_RIGHT) then  bbb(bbbRow ,bbbCol + 1)= bbb(bbbRow,bbbCol + 1) or GRD_LEFT
                  SelectObject v->memDC, v->SolidLine2 'Switch to Bolder Pen
                  if (MyBorder and GRD_TOPB)  THEN  MoveToEx(v->memDC,DrawRect.Left-1,DrawRect.Top,@mypoint) :  LineTo(v->memDC,DrawRect.Right-1,DrawRect.Top) 
                  if (MyBorder and GRD_LEFTB)  THEN  MoveToEx(v->memDC,DrawRect.Left,DrawRect.Top,@mypoint) :  LineTo(v->memDC,DrawRect.Left,DrawRect.bottom)
                  if (MyBorder and GRD_BOTTOMB) then bbb(bbbRow +1,bbbCol)= bbb(bbbRow,bbbCol) or GRD_TOPB
                  if (MyBorder and GRD_RIGHTB) then  bbb(bbbRow ,bbbCol+1)= bbb(bbbRow,bbbCol) or GRD_LEFTB
                  SelectObject v->memDC, GetStockObject(BLACK_PEN) 'Leave with a SolidPen1
                  'print "Original  " & Str(MyRow)  & " : "  & Str(MyCol)   & " : "  & Str(MyBorder)
               END IF
               
                if bbb(bbbRow,bbbCol) > 0 THEN 
                  if (bbb(bbbRow,bbbCol) and GRD_TOP)  THEN  MoveToEx(v->memDC,DrawRect.Left,DrawRect.Top,@mypoint) :  LineTo(v->memDC,DrawRect.Right-1,DrawRect.Top) : bbb(bbbRow,bbbCol)=0
                  if (bbb(bbbRow,bbbCol) and GRD_LEFT)  THEN  MoveToEx(v->memDC,DrawRect.Left,DrawRect.Top,@mypoint) :  LineTo(v->memDC,DrawRect.Left,DrawRect.bottom)  : bbb(bbbRow,bbbCol)=0
                  SelectObject v->memDC, v->SolidLine2 'Switch to Bolder Pen
                  if (bbb(bbbRow,bbbCol) and GRD_TOPB)  THEN  MoveToEx(v->memDC,DrawRect.Left - 1,DrawRect.Top,@mypoint) :  LineTo(v->memDC,DrawRect.Right-1,DrawRect.Top) : bbb(bbbRow,bbbCol)=0 
                  if (bbb(bbbRow,bbbCol) and GRD_LEFTB)  THEN  MoveToEx(v->memDC,DrawRect.Left,DrawRect.Top,@mypoint) :  LineTo(v->memDC,DrawRect.Left,DrawRect.bottom)  : bbb(bbbRow,bbbCol)=0
                  SelectObject v->memDC, GetStockObject(BLACK_PEN) 'Leave with a SolidPen1 
               END IF
       NEXT MyCol
     NEXT MyRow
     'Now draw a black line over the ltgray cell line on the header
     MoveToEx(v->memDC,tempx-1,tempy,@mypoint) :  LineTo(v->memDC,tempx-1,DrawRect.bottom)
     MoveToEx(v->memDC,tempx-1,tempy-1,@mypoint) :  LineTo(v->memDC,DrawRect.Right,tempy-1) 
     
     
   GetClientRect(HWND,@GridRect)
   myRgnRect=CreateRectRgnIndirect(@GridRect)
   SelectClipRgn v->memDC,myRgnRect
   DeleteObject(myRgnRect)
   end if
END SUB


SUB GRD_PaintSelection(hwnd AS HWND)
  
   DIM rc AS RECT, wRect AS RECT
   Dim hPen AS hGDIOBJ, hBrush AS hGDIOBJ
   Dim bkBrush AS hGDIOBJ, hBrushSel AS hGDIOBJ
   Dim v As GridData Ptr
   Dim temp as long, temp1 as long, temp2 as long, temp3 as long, temp4 as Long
   dim temp5 as hGDIOBJ, temp6 as hGDIOBJ
   dim hDC as hDC
   dim RowHeaderWidth as Long , ColHeaderHeight as Long, SelColWidth as Long, SelRowHeight as Long
   dim myRgnRect as hRgn
   
   v = GetProp(hWnd,"GridData")
   
   GetClientRect(HWND,@wRect)
      
   if v <> 0 then
      
      RowHeaderWidth=GRD_FormatAttribute(hwnd,0,0,GRDFORMAT_GETCOLWIDTH)
      ColHeaderHeight=GRD_FormatAttribute(hwnd,0,0,GRDFORMAT_GETROWHEIGHT)
      SelColWidth=GRD_FormatAttribute(hwnd,0,GS_.SelCol,GRDFORMAT_GETCOLWIDTH)
      SelRowHeight=GRD_FormatAttribute(hwnd,GS_.SelRow,0,GRDFORMAT_GETROWHEIGHT)
      
      'GetClipRgn(v->memDC,myRgnRect)
      
       'ExcludeClipRect hDC,L,T,R,B
      ExcludeClipRect(v->memDC,0,wRect.Bottom - GS_.HScrollWidth-GS_.StatusBarWidth,wRect.Right,wRect.Bottom)'Bottom
      ExcludeClipRect(v->memDC,wRect.Right - GS_.VScrollWidth,0,wRect.Right,wRect.Bottom) 'Right
      ExcludeClipRect(v->memDC,0,0,RowHeaderWidth + GS_.GridLeft,wRect.Bottom)'LeftHeader
      ExcludeClipRect(v->memDC,0,0,wRect.Right,ColHeaderHeight+GS_.GridTop)'TopHeader
        
     IF GS_.SelRow2 > 0 AND GS_.SelCol2 > 0 THEN
     
        
          'we have multicell
          'check for wholerow or whole column
         temp=GRD_FormatAttribute(hwnd,0,0,GRDFORMAT_GETCOLWIDTH) + GS_.GridLeft    
         temp3 = temp
         FOR temp1 = GS_.LeftCol TO GS_.SelCol
             IF temp1 < GS_.SelCol AND temp < wRect.Right THEN temp=temp + GRD_FormatAttribute(hwnd,0,temp1,GRDFORMAT_GETCOLWIDTH) - 1
             IF temp > wRect.Right THEN EXIT FOR
         NEXT temp1
 
         FOR temp1 = GS_.LeftCol TO GS_.SelCol2
             IF temp1 < GS_.SelCol2 AND temp3 < wRect.Right THEN temp3=temp3 + GRD_FormatAttribute(hwnd,0,temp1,GRDFORMAT_GETCOLWIDTH) - 1
             IF temp3 > wRect.Right THEN EXIT FOR
         NEXT temp1
 
         temp2=GRD_FormatAttribute(hwnd,0,0,GRDFORMAT_GETROWHEIGHT) + GS_.GridTop
         temp4=temp2
         FOR temp1=GS_.TopRow TO GS_.SelRow
             IF temp1 < GS_.SelRow AND temp2 < wRect.Bottom THEN temp2=temp2 + GRD_FormatAttribute(hwnd,Temp1,0,GRDFORMAT_GETROWHEIGHT)
             IF temp2 > wRect.Bottom THEN EXIT FOR
         NEXT temp1
 
         FOR temp1=GS_.TopRow TO GS_.SelRow2
             IF temp1 < GS_.SelRow2 AND temp4 < wRect.Bottom THEN temp4=temp4 + GRD_FormatAttribute(hwnd,Temp1,0,GRDFORMAT_GETROWHEIGHT)
             IF temp4 > wRect.Bottom THEN EXIT FOR
         next temp1
  
         GS_.StartRow = GS_.SelRow : GS_.EndRow= GS_.SelRow2  : IF GS_.EndRow < GS_.StartRow THEN SWAP GS_.EndRow,GS_.StartRow
         GS_.StartCol = GS_.SelCol : GS_.EndCol= GS_.SelCol2  : IF GS_.EndCol < GS_.StartCol THEN SWAP GS_.EndCol,GS_.StartCol
         
         IF GS_.SelRow2 >= GS_.SelRow AND GS_.SelCol2 >= GS_.SelCol THEN   '1     
                 GS_.TempSelectRect.Left=temp - 2 : GS_.TempSelectRect.Right = temp3 + GRD_FormatAttribute(hwnd,0,GS_.SelCol2,GRDFORMAT_GETCOLWIDTH)
                 GS_.TempSelectRect.Top=temp2 - 2 : GS_.TempSelectRect.Bottom = temp4 + 1 + GRD_FormatAttribute(hwnd,GS_.SelRow2,0,GRDFORMAT_GETROWHEIGHT)
                 'Print "Here1"
           elseif  GS_.SelRow2 >= GS_.SelRow AND GS_.SelCol2 < GS_.SelCol THEN
                 GS_.TempSelectRect.Left=temp3 - 2 : GS_.TempSelectRect.Right = temp + GRD_FormatAttribute(hwnd,0,GS_.SelCol,GRDFORMAT_GETCOLWIDTH)
                 GS_.TempSelectRect.Top=temp2 - 2 : GS_.TempSelectRect.Bottom = temp4 + 1 + GRD_FormatAttribute(hwnd,GS_.SelRow2,0,GRDFORMAT_GETROWHEIGHT)   
                 'Print "Here2"
           elseif  GS_.SelRow2 < GS_.SelRow AND GS_.SelCol2 >= GS_.SelCol THEN
                GS_.TempSelectRect.Left=temp - 2 : GS_.TempSelectRect.Right = temp3 + GRD_FormatAttribute(hwnd,0,GS_.SelCol2,GRDFORMAT_GETCOLWIDTH)
                GS_.TempSelectRect.Top=temp4 - 2 : GS_.TempSelectRect.Bottom = temp2 + 1 + GRD_FormatAttribute(hwnd,GS_.SelRow,0,GRDFORMAT_GETROWHEIGHT) 
                 'Print "Here3"
           else 'GS_.SelRow2 < GS_.SelRow AND GS_.SelCol2 < GS_.SelCol    
                GS_.TempSelectRect.Left=temp3 - 2 : GS_.TempSelectRect.Right = temp + GRD_FormatAttribute(hwnd,0,GS_.SelCol,GRDFORMAT_GETCOLWIDTH)
                GS_.TempSelectRect.Top=temp4 - 2 : GS_.TempSelectRect.Bottom = temp2 + 1 + GRD_FormatAttribute(hwnd,GS_.SelRow,0,GRDFORMAT_GETROWHEIGHT)
                 'Print "Here4"
         end if 
     
      if GS_.EndCol < GS_.LeftCol or GS_.StartCol > GS_.RightCol THEN
          goto SelectionIsDone
       END IF
   
       if GS_.EndRow < GS_.TopRow or GS_.StartRow > GS_.BottomRow THEN
           goto SelectionIsDone
       END IF  
      
          SelectObject v->memDC, v->SolidLine3
          SelectObject v->memDC, GetStockObject(HOLLOW_BRUSH)

          Rectangle v->memDC,GS_.TempSelectRect.Left,GS_.TempSelectRect.Top,GS_.TempSelectRect.Right,GS_.TempSelectRect.Bottom
          
          'Highlight Rect
          hBrushSel = GetSysColorBrush(COLOR_HIGHLIGHT)
          temp6 = SelectObject(v->memDC, hBrushSel)
          temp5 = SelectObject(v->memDC,GetStockObject(NULL_PEN))
          SetROP2(v->memDC, R2_MERGEPENNOT)
          Rectangle v->memDC,GS_.TempSelectRect.Left +3,GS_.TempSelectRect.Top+3,GS_.TempSelectRect.Right-2,GS_.TempSelectRect.Bottom-2
          
          'now reverse highlight the selection
          IF GS_.SelRow >= GS_.TopRow AND GS_.SelRow < GS_.BottomRow AND GS_.SelCol >= GS_.LeftCol AND GS_.SelCol <= GS_.RightCol THEN
                 Rectangle v->memDC,temp + 1, temp2+1, temp + SelColWidth-2,temp2 + SelRowHeight -1
          END IF
          
          GS_.EditRect.Left=temp : GS_.EditRect.Top=temp2
          GS_.EditRect.Right=temp + SelColWidth-1 : GS_.EditRect.Bottom=temp2 + SelRowHeight
          
          
          SetROP2(v->memDC, R2_COPYPEN)
                                             
          DeleteObject(hBrushSel)
                                                                              
    ELSE  
               
       if GS_.SelCol < GS_.LeftCol or GS_.SelCol > GS_.RightCol THEN
          goto SelectionIsDone
       END IF
   
       if GS_.SelRow < GS_.TopRow or GS_.SelRow > GS_.BottomRow THEN
           goto SelectionIsDone
       END IF             
               
      'Figure out the SelCol location
      temp=GRD_FormatAttribute(hwnd,0,0,GRDFORMAT_GETCOLWIDTH) + GS_.GridLeft      
      FOR temp1=GS_.LeftCol TO GS_.SelCol - 1
               IF temp1 > GS_.TotCols THEN EXIT FOR
               temp=temp+GRD_FormatAttribute(hwnd,0,temp1,GRDFORMAT_GETCOLWIDTH)-1
      NEXT temp1
      GS_.TempSelectRect.Left=temp-2 : GS_.TempSelectRect.Right = temp + GRD_FormatAttribute(hwnd,0,temp1,GRDFORMAT_GETCOLWIDTH)
     
'     'Figure out the SelRow location
       temp=GRD_FormatAttribute(hwnd,0,0,GRDFORMAT_GETROWHEIGHT)+GS_.GridTop
       FOR temp1=GS_.TopRow TO GS_.SelRow - 1
         IF temp1>GS_.TotRows THEN EXIT FOR
         temp=temp+GRD_FormatAttribute(hwnd,Temp1,0,GRDFORMAT_GETROWHEIGHT)
       NEXT temp1        
       GS_.TempSelectRect.Top=temp - 2 : GS_.TempSelectRect.Bottom = temp + 1 + GRD_FormatAttribute(hwnd,temp1,0,GRDFORMAT_GETROWHEIGHT)
' 
    
      SelectObject v->memDC, v->SolidLine3
      SelectObject v->memDC, GetStockObject(HOLLOW_BRUSH)

      Rectangle v->memDC,GS_.TempSelectRect.Left,GS_.TempSelectRect.Top,GS_.TempSelectRect.Right,GS_.TempSelectRect.Bottom
      GS_.EditRect.Left=GS_.TempSelectRect.Left : GS_.EditRect.Top=GS_.TempSelectRect.Top
      GS_.EditRect.Right=GS_.TempSelectRect.Right : GS_.EditRect.Bottom=GS_.TempSelectRect.Bottom
   end if 'if multicell selection
   
   SelectionIsDone:
 
   GetClientRect(HWND,@wRect)
   myRgnRect=CreateRectRgnIndirect(@wRect)
   SelectClipRgn v->memDC,myRgnRect
   DeleteObject(myRgnRect) 
      
 end if 'v
 
END SUB


FUNCTION GRD_ButtonPoly(op AS LONG) AS HRGN
   ' Returns a handle to a polygon
    DIM pt(2) AS POINT

   IF op = 1 THEN
    pt(0).x = 11    : pt(0).y = 3
    pt(1).x = 11    : pt(1).y = 13
    pt(2).x = 6     : pt(2).y = 8
   END IF

   IF op = 2 THEN
    pt(0).x = 27    : pt(0).y = 3
    pt(1).x = 27    : pt(1).y = 13
    pt(2).x = 22    : pt(2).y = 8
   END IF

   IF op = 3 THEN
    pt(0).x = 44    : pt(0).y = 8
    pt(1).x = 39    : pt(1).y = 13
    pt(2).x = 39    : pt(2).y = 3
   END IF

   IF op = 4 THEN
    pt(0).x = 59    : pt(0).y = 8
    pt(1).x = 54    : pt(1).y = 13
    pt(2).x = 54    : pt(2).y = 3
   END IF

    RETURN CreatePolygonRgn( varptr(pt(0)), 3, WINDING)
END FUNCTION

Function GRD_LoadFont(hWnd AS hWnd,Slot as long, MyFont as hGDIOBJ) as long
    Dim v AS GridData PTR
    
   if Slot < 9 or Slot > 15 THEN 
     Function = 0
     exit Function 
   END IF
   
    v = GetProp(hWnd,"GridData")
    if v->CellFonts(Slot) <> 0 then DeleteObject v->CellFonts(Slot)
    v->CellFonts(Slot)=MyFont
    Function = 1
   
END FUNCTION 

Function GRD_ActivateListControl(Hwnd as Hwnd,Action as long = 0)as long
    'Print "List Control Activate"
   
    Dim v AS GridData PTR
    dim temp as Long
    Dim I as LONG
    dim MyWStr as WSTRING * GRD_MAXCELLLEN
    dim MyWStr2 as WSTRING * GRD_MAXCELLLEN
    dim MyCount as LONG
    dim temp1 as long, temp2 as long, temp3 as long, temp4 as LONG
    dim tempHWND as HWND
   
    v = GetProp(hWnd,"GridData")
      
    if Action = 0 THEN 'Deactivate     
         SetWindowPos v->hListControl, 0, 0, 0, 0, 0,SWP_HIDEWINDOW
         
       '  tempHWND = GetFocus
       '  if TempHWND = v->hListControl THEN
       '     SetFocus HWND
       '  END IF
        SetFocus HWND
        'Print "DeActivate List"
'     
     else 'Activate
             
        SendMessage v->hListControl,WM_SETFONT,cast(wParam,v->CellFonts(GRDSYSFONTDEFAULT)),cast(lParam,MAKELONG(TRUE,0))
        'tempHWND = GetFocus
                 
        if Action = 1 THEN 'CHAR or LButtonClick
                  MyWStr = ""
                  SendMessage v->hListControl, LB_RESETCONTENT, 0, 0
                  GRD_GetListing(HWND,0,Varptr(MyWStr))
                  if len(MyWStr) = 0 THEN  Exit function
                  temp1 = ParseCountW(MyWStr)
                  for temp2 = 1 to temp1
                     ParseW(MyWStr,temp2,varptr(MyWStr2))
                     SendMessageW v->hListControl, LB_ADDSTRING, 0, cast(lParam,VARPTR(MyWStr2))
                  next temp2                                                                    
                  SetWindowPos v->hListControl, 0, GS_.TempSelectRect.Left - 2, GS_.TempSelectRect.Bottom+2, 100, min(v->cyChar * temp1,v->cyChar * 4) ,SWP_SHOWWINDOW
                  SetFocus v->hListControl 
                  SendMessage v->hListControl, LB_SETCURSEL, temp4, 0
                  
        END IF 
        
              
    END IF
                                                                                                                   
  function = 0  
END FUNCTION      



Function GRD_ActivateEditControl(Hwnd as Hwnd,Action as long = 0)as long
    'Print "Edit Control Activate"
    Dim MyEditText as WString * GRD_MAXCELLLEN
    Dim v AS GridData PTR
    dim temp as Long,MyType as LONG
    dim tempHWND as hwnd
    dim MyCell as CellFormat3
    
    v = GetProp(hWnd,"GridData")
    
    if v = 0 THEN
       exit function
    END IF
      
    if Action = 0 THEN 'Deactivate
      MyEditText = ""
      temp = len(myedittext)
      SetWindowTextW v->hEditControl,varptr(MyEditText)
      SendMessage v->hEditControl,EM_SETSEL,temp,temp
      SetWindowPos v->hEditControl, 0, 0, 0, 0, 0,SWP_HIDEWINDOW OR SWP_NOSENDCHANGING
      tempHWND = GetFocus
      if TempHWND = v->hEditControl THEN
            SetFocus HWND
      END IF
      SetFocus HWND
     
     else 'Activate
        'Don't allow an exit field for the following reasons 
        GRD_CalcCellFormat(Hwnd,GS_.SelRow,GS_.SelCol,@Mycell)
        if MyCell.WriteLock <> GRD_NOLOCK THEN exit function
        if MyCell.CellType = GRDFCT_USERBUT or MyCell.CellType = GRDFCT_CKBOX or MyCell.CellType = GRDFCT_PICTURE THEN
            exit function
        END IF
        
        SendMessage v->hEditControl,WM_SETFONT,cast(wParam,v->CellFonts(GRDSYSFONTDEFAULT)),cast(lParam,MAKELONG(TRUE,0))
        MyType = GRD_FormatAttribute(HWND,GS_.SelRow,GS_.SelCol,GRDFORMAT_GETTYPE)
        if MyType = GRDFCT_DDCBS and v->ComboEdit <> 0 THEN
           Exit Function
        END IF
       
        if Action = 1 THEN 'VK2
          SetWindowPos v->hEditControl, HWND_BOTTOM, GS_.EditRect.Left +2, GS_.EditRect.Top+2, GS_.EditRect.Right-GS_.EditRect.Left-4, _
          GS_.EditRect.Bottom-GS_.EditRect.Top-4 ,SWP_SHOWWINDOW OR SWP_DRAWFRAME OR SWP_NOSENDCHANGING
        
          GRD_GetCell(HWND,GS_.SelRow,GS_.SelCol,varptr(MyEditText)) 
          temp = len(myedittext)
          SetWindowTextW v->hEditControl,varptr(MyEditText)
          SendMessage v->hEditControl,EM_SETSEL,temp,temp
          SetFocus v->hEditControl
        END IF 
        
        if Action = 2 THEN 'CHAR
          MyType = GRD_FormatAttribute(HWND,GS_.SelRow,GS_.SelCol,GRDFORMAT_GETTYPE)  
          if MyType = GRDFCT_NUMBER or MyType = GRDFCT_DATE THEN
               GetWindowTextW v->hEditControl,varptr(MyEditText),2
               IF instr(MyEditText ,any "+-.0123456789") THEN
                      'Nothing for now"
                   else
                     MyEditText = ""
                     SetWindowTextW v->hEditControl,varptr(MyEditText) 
                     SendMessage v->hEditControl,EM_SETSEL,0,0
               end if
             else  
               'Nothing for now
          end if
          SetWindowPos v->hEditControl, HWND_BOTTOM, GS_.EditRect.Left +2, GS_.EditRect.Top+2, GS_.EditRect.Right-GS_.EditRect.Left-4, _
          GS_.EditRect.Bottom-GS_.EditRect.Top-4 ,SWP_SHOWWINDOW OR SWP_DRAWFRAME OR SWP_NOSENDCHANGING
          SetFocus v->hEditControl
        END IF 
        
        if Action = 3 THEN 'LButtonDClick
          SetWindowPos v->hEditControl, HWND_BOTTOM, GS_.EditRect.Left +2, GS_.EditRect.Top+2, GS_.EditRect.Right-GS_.EditRect.Left-4, _
          GS_.EditRect.Bottom-GS_.EditRect.Top-4 ,SWP_SHOWWINDOW OR SWP_DRAWFRAME OR SWP_NOSENDCHANGING
        
          GRD_GetCell(HWND,GS_.SelRow,GS_.SelCol,varptr(MyEditText)) 
          temp = len(myedittext)
          SetWindowTextW v->hEditControl,varptr(MyEditText)
          SendMessage v->hEditControl,EM_SETSEL,0,temp
          SetFocus v->hEditControl
        END IF   
    END IF
                                                                                                                   
  function = 0  
END FUNCTION      



FUNCTION GRD_MakeStandardFont(hWnd AS hWnd,TypeFace AS WSTRING Ptr,TheFontSize AS LONG,TheWeight AS LONG= FW_NORMAL,CharSet AS LONG=DEFAULT_CHARSET) AS hGDIOBJ
  
   DIM  hDC2 as hDC
   DIM MyFontSize as LONG,TabFontSize as long, UsedFontSize as long , VertFlag as LONG
   DIM Family as LONG
   DIM CyPixels2 as LONG
   DIM typefacename AS WString * 50
   DIM temptype as hGDIOBJ
   Dim v AS GridData PTR
   Dim SystemFontFlag as LONG,FontCount as long
   Dim MyWeight as integer
   dim MyUnderline as ulong
   Dim MyItalic as ULONG
   DIM tm AS TEXTMETRIC
   Dim tempdc as HDC
   DIM StatusFontSize as LONG
                
    v = GetProp(hWnd,"GridData")
    
    MyFontSize = TheFontSize
    MyWeight = TheWeight
    TypeFaceName = *TypeFace

     hDC2 = GetDC(HWND_DESKTOP)
     CyPixels2  = GetDeviceCaps(hDC2, LOGPIXELSY)
     ReleaseDC HWND_DESKTOP, hDC2
     MyFontSize = (myfontsize * CyPixels2) \ 72
     TabFontSize = (10 * CyPixels2) \ 72  
     StatusFontSize = (9 * CyPixels2) \ 72 
     
    'System Fonts
'CONST GRDSYSFONTDEFAULT = 0
'CONST GRDSYSFONTNORM = 1
'CONST GRDSYSFONTBOLD = 2
'CONST GRDSYSFONTITAL = 3
'CONST GRDSYSFONTTABNORM = 4
'CONST GRDSYSFONTTABBOLD = 5
'CONST GRDSYSFONTNORMVERT = 6
'CONST GRDSYSFONTBOLDVERT = 7
'CONST GRDSYSFONTSTATUSBAR = 8

    for FontCount = 0 to 8
        if FontCount = 0 then MyWeight = FW_NORMAL : MyUnderline = 0 : MyItalic = 0    : UsedFontSize = MyFontSize  :VertFlag = 0
        if FontCount = 1 then MyWeight = FW_NORMAL : MyUnderline = 0 : MyItalic = 0    : UsedFontSize = MyFontSize  :VertFlag = 0
        if FontCount = 2 then MyWeight = FW_BOLD   : MyUnderline = 0 : MyItalic = 0    : UsedFontSize = MyFontSize  :VertFlag = 0
        if FontCount = 3 then MyWeight = FW_NORMAL : MyUnderline = 0 : MyItalic = TRUE : UsedFontSize = MyFontSize  :VertFlag = 0
        if FontCount = 4 then MyWeight = FW_NORMAL : MyUnderline = 0 : MyItalic = 0    : UsedFontSize = TabFontSize :VertFlag = 0
        if FontCount = 5 then MyWeight = FW_BOLD   : MyUnderline = 0 : MyItalic = 0    : UsedFontSize = TabFontSize :VertFlag = 0
        if FontCount = 6 then MyWeight = FW_NORMAL : MyUnderline = 0 : MyItalic = 0    : UsedFontSize = MyFontSize  :VertFlag = 900
        if FontCount = 7 then MyWeight = FW_BOLD   : MyUnderline = 0 : MyItalic = 0    : UsedFontSize = MyFontSize  :VertFlag = 900
        if FontCount = 8 then MyWeight = FW_NORMAL : MyUnderline = 0 : MyItalic = 0    : UsedFontSize = StatusFontSize  :VertFlag = 0
        
        temptype = CreateFontW(0 - Usedfontsize, 0, VertFlag, VertFlag, MyWeight, MyItalic, MyUnderline, 0, _
                       CharSet, OUT_TT_PRECIS, CLIP_DEFAULT_PRECIS, _
                       DEFAULT_QUALITY, FF_DONTCARE, TypeFaceName)
   
         if v->CellFonts(FontCount) <> 0 then DeleteObject v->CellFonts(FontCount)
         v->CellFonts(FontCount)=temptype
         
    NEXT 
    
    tempdc = GetDC(Hwnd)
    SelectObject tempdc, v->CellFonts(GRDSYSFONTNORM)
    GetTextMetrics (tempdc, @tm)              'Get font parameters
    v->cyChar = tm.tmHeight + tm.tmExternalLeading + v->cyCharEx 'Line spacing
    v->charWidth  = tm.tmAveCharWidth               'Average character width
    ReleaseDC(HWND,tempDC)
    
                                                
  Return temptype                                                
END FUNCTION

FUNCTION GRD_MakeStandardFontA(hWnd AS hWnd,TypeFace AS ZString Ptr,TheFontSize AS LONG,TheWeight AS LONG= FW_NORMAL,CharSet AS LONG=DEFAULT_CHARSET) AS hGDIOBJ
      dim tempstr as Wstring * GRD_MAXCELLLEN
      dim tempstrptr as Wstring ptr
      dim z as string * GRD_MAXCELLLEN
      tempstrptr=varptr(tempstr)
      
      z = *TypeFace
      tempstr = wstr(z)
     
      Function = GRD_MakeStandardFont(hWnd,tempstrptr, TheFontSize, TheWeight,CharSet) 
end Function

FUNCTION GRD_MakeStandardFontS(hWnd AS hWnd,TypeFace AS string,TheFontSize AS LONG,TheWeight AS LONG= FW_NORMAL,CharSet AS LONG=DEFAULT_CHARSET) AS hGDIOBJ
      dim tempstr as Wstring * GRD_MAXCELLLEN
      dim tempstrptr as Wstring ptr
      dim z as string * GRD_MAXCELLLEN
      tempstrptr=varptr(tempstr)
     
      tempstr = wstr(TypeFace)
      
      function =GRD_MakeStandardFont(hWnd,tempstrptr, TheFontSize, TheWeight,CharSet)
       
end function

FUNCTION ParseCountW(s as Wstring,del as Wstring = ",") as long
  dim MyCount as long
  dim x as long
  
 
  MyCount = 0 
  for x = 0 to len(s)-1
     if chr(s[x]) = del then MyCount = MyCount +1
  next x
    
  RETURN MyCount + 1

END FUNCTION

SUB ParseW(s as Wstring, MyItem as long, MyWStr as WString Ptr,del as wstring = ",")
  dim MyTargetStart as long, MyTargetEnd as long, MyCount as long
  dim x as long
  dim MyStart as long, MyEnd as long ' Location of characters
'
MyCount = 0 
MyTargetStart = MyItem -1
MyTargetEnd = MyItem
If MyItem =1 then MyStart = 1 : MyEnd = len(s)  

  for x = 0 to len(s)-1
     if chr(s[x]) = del then 
       MyCount = MyCount + 1
       If MyCount = MyTargetStart then MyStart = x + 2
       If MyCount = MyTargetEnd then MyEnd = x : Goto Done
     end if   
  next x
 
 Done: 
  
*MyWStr = mid(s,MyStart,MyEnd-MyStart+1)
'    
'
END SUB

'********************************************************************
' TabSheet procedures
'********************************************************************

Function GRD_ShowTabSheets(hwnd as hwnd, items as wstring Ptr, TabSheetSize as long = 120) as long

dim x as long, temp as long
dim tempstr2 as wstring * 35 
dim szBuffer as wstring * 32
Dim v as GRIDDATA PTR
dim rc as RECT
DIM MyEditText as Wstring * 1000
DIM mysize AS SIZEL
DIM result AS LONG
dim MyOffset as long

MyOffset = 73  ' From transport buttons
  
 v = GetProp(hWnd,"GridData")
    
 if v <> 0 THEN
    IF LEN(TRIM(*items)) > 0 THEN
         x=PARSECOUNTW(*items)
         IF x > MAXGS THEN GOTO Done  'max of 10 sheets in demo  MAXGS was set to 10 - Total 11 with 0 element
   '     IF x < 2 THEN GOTO Done  'need more than just one sheet to have a tab control    
         FOR temp = 1 TO x
            PARSEW(*items,temp,tempstr2)
            szBuffer=tempstr2
             v->SH(temp-1).SheetName = szBuffer
             result=GetTextExtentPoint32W(v->memDC,szBuffer,LEN(szBuffer),@mysize)
             if temp = 1 then v->SH(temp-1).SheetOffset1 = MyOffset  else v->SH(temp-1).SheetOffset1 = v->SH(temp-2).SheetOffset2
             MyOffset = MyOffset + mysize.cx
             v->SH(temp-1).SheetOffset2 =  v->SH(temp-1). SheetOffset1 + mysize.cx + 6   
             v->SH(temp-1).SheetColor = cast(uinteger,GetStockObject(LTGRAY_BRUSH))  
         NEXT temp
         v->ActiveSheetUbound = temp - 2
        
        'Put the same default Tab Position on all sheets
        ' for temp = 0 to MAXGS
       '   v->SH(temp).TabSheetPos= TabSheetSize
        ' NEXT
        
        v->TabSheetPos = TabSheetSize
       
        'Show the Tab Sheet Control by an WM_SIZE message
        GetClientRect(hWnd,@rc)
        PostMessage(hWnd,WM_SIZE,SIZE_RESTORED,MAKELONG(rc.Right-rc.Left,rc.Bottom-rc.Top))
        InvalidateRect hwnd, BYVAL NULL, 0 : UpdateWindow hwnd
        
        'temp demo
        SetWindowTextW v->hSheetControl,varptr(MyEditText) 
            
      END IF 'len
   end if 'v
   Done:
 function = 0    
END FUNCTION

   
'********************************************************************
' Menu procedures
'********************************************************************

SUB GRD_MakeRClickMenu(BYVAL hwnd AS HWND, BYVAL items as wstring ptr,which as Long )  'items are comma delimited
   DIM x as long
   DIM temp as long
   Dim hMenuPopUp as hMenu
   'Dim hMenuPopUpTemp as hMenu
   Dim szBuffer as zstring * 50
   Dim tempstr2 as wstring * 50
   dim tempstr2ptr as Wstring ptr
   Dim v as GRIDDATA PTR
   
    if which > GRD_NUMMENUS then exit sub
    if which < 0 THEN exit sub
       
   
   tempstr2ptr = varptr(tempstr2)
   v = GetProp(hWnd,"GridData")
    
      if v <> 0 THEN
       'Menu
         IF LEN(TRIM(*items)) > 0 THEN
            x=PARSECOUNTW(*items)
            IF x > 20 THEN GOTO Done  'max of 20 items
 '          
            hMenuPopUp=CreatePopUpMenu
            FOR temp = 1 TO x
               PARSEW(*items,temp,tempstr2)
               szBuffer=tempstr2
              IF tempstr2="-" THEN
                 AppendMenu hMenuPopUp,MF_SEPARATOR,0,""
                ELSE
                 AppendMenu hMenuPopUp,MF_STRING,temp,szBuffer
              END IF
            NEXT temp
            IF v->RClickMenu(which) <> 0 THEN DestroyMenu v->RClickMenu(which)
            v->RClickMenu(which)=hMenuPopUp
         END IF 'len
      END IF 'v 
  Done:
End Sub

'********************************************************************
' Miscellaneous  cell formatting procedures
'********************************************************************
'Substitution of Header Formats
  'Row Col and Cell 
    'MyForeColor      uByte5
    'MyBackColor      uByte6
    'CellType         uByte7
    'WriteLock        uByte8
  'Col and Cell
    'CellFormat       uBYTE3   Row CellFormat is uBYTE1
    'CellCombo        uBYTE4   Col is ColCombo
    
    'MyFont           uBYTE9
    'MyJust           uBYTE10
    'MyCase           uBYTE11



    
sub GRD_CalcCellFormat(Hwnd as hwnd,myrow as long,mycol as long,Mycell as CELLFORMAT3 ptr)
  'This is for the Current Sheet  -- Used in Painting the cells properly formatted
  'Get Column, then Row, then Cell with each superceding the other
 
  Dim v As GridData Ptr
  Dim cellstr as wstring ptr,colstr as WSTRING ptr, rowstr as WSTRING ptr, cornstr as WSTRING ptr
  dim CellElement as long,RowElement as long, ColElement as LONG
  
  DIM dimrows as long,dimcols as long
  Dim cell as CELLFORMAT3, row as ROWFORMAT1, col as COLFORMAT2, corn as CornerFormat0 ,rescell as CELLFORMAT3
  
   
  v = GetProp(hWnd,"GridData")
  if v <> 0 then
  
   dimrows = GS_.DimRows
   dimcols = GS_.DimCols
   if myrow < 0 then Goto Done2
   if mycol < 0 then Goto Done2 
   if myrow > dimrows then Goto Done2
   if mycol > dimcols then Goto Done2
  
   
   if myrow > 0 and mycol > 0 then
       CellElement=(myrow * (GS_.DimCols + 1) ) + mycol
       RowElement=(myrow * (GS_.DimCols + 1) ) + 0
       ColElement=(0 * (GS_.DimCols + 1) ) + mycol
      
      Dim ttt as GRDArray Pointer = cast(GRDArray Ptr,GS_.ws)
      
      If ttt <> 0 Then
            cellstr = ttt[CellElement].ws 'test if zero
            rowstr  = ttt[RowElement].ws  'should not be zero
            colstr  = ttt[ColElement].ws  'should not be zero
            
            CopyMemory(@row,rowstr,HEADERBYTES)
            CopyMemory(@col,colstr,HEADERBYTES)
            if cellstr <> 0 THEN
              CopyMemory(@cell,cellstr,HEADERBYTES)
            END IF 
            
            'Cell Border
            if cellstr <> 0 THEN
              if cell.MyBorder <> 0 THEN rescell.MyBorder = cell.MyBorder
            END IF
            
            'Cell Font
            rescell.MyFont = col.MyFont
            if cellstr <> 0 THEN
              if cell.MyFont <> 0 THEN rescell.MyFont = cell.MyFont
            END IF 
            
            'Cell Case
            rescell.MyCase = col.MyCase
            if cellstr <> 0 THEN
              if cell.MyCase <> 0 THEN rescell.MyCase = cell.MyCase
            END IF
            
            'Cell Just
            rescell.MyJust = col.MyJust
            if cellstr <> 0 THEN
              if cell.MyJust <> 0 THEN rescell.MyJust = cell.MyJust
            END IF
            
            'Cell Format
            rescell.CellFormat = col.CellFormat
            if row.CellFormat <> 0 THEN rescell.CellFormat = row.CellFormat
            if cellstr <> 0 THEN
              if cell.CellFormat <> 0 THEN rescell.cellFormat = cell.CellFormat
            END IF 
            
            'Cell Type
            rescell.CellType = col.CellType
            if row.CellType <> 0 THEN rescell.CellType = row.CellType
            if cellstr <> 0 THEN
              if cell.CellType <> 0 THEN rescell.CellType = cell.CellType
            END IF 
            
            'Cell WriteLock
            rescell.WriteLock = col.WriteLock
            if row.WriteLock <> 0 THEN rescell.WriteLock = row.WriteLock
            if cellstr <> 0 THEN
              if cell.WriteLock <> 0 THEN rescell.WriteLock = cell.WriteLock
            END IF 
            
            'Cell Background Color
            rescell.MyBackColor = col.MyBackColor
            if row.MyBackColor <> 0 THEN rescell.MyBackColor = row.MyBackColor
            if cellstr <> 0 THEN
              if cell.MyBackColor <> 0 THEN rescell.MyBackColor = cell.MyBackColor
            END IF
            
             'Cell Foreground Color
            rescell.MyForeColor = col.MyForeColor
            if row.MyForeColor <> 0 THEN rescell.MyForeColor = row.MyForeColor
            if cellstr <> 0 THEN
              if cell.MyForeColor <> 0 THEN rescell.MyForeColor = cell.MyForeColor
            END IF
            
            'if myrow = 2 and mycol=2 THEN
            '   Print "Get Header Col  "  & str(col.MyForeColor)
            '   Print "Get Header Row  "  & str(row.MyForeColor)
            '   Print "Get Header Cell "  & str(cell.MyForeColor)
            'END IF
            
             'Cell Combo Box
            rescell.CellCombo = col.ColCombo
            if row.RowCombo <> 0 THEN rescell.CellCombo = row.RowCombo
            if cellstr <> 0 THEN
              if cell.CellCombo <> 0 THEN rescell.CellCombo = cell.CellCombo
            END IF
            
            CopyMemory(Mycell,@rescell,HEADERBYTES)
            
       else 'ttt = 0
            'function = 1
            Goto Done2   
      end if 'ttt 
    else  'dealing with a row or column or corner   
      Dim ttt as GRDArray Pointer = cast(GRDArray Ptr,GS_.ws) 
      If ttt <> 0 Then
         cornstr = ttt[0].ws ' corner used for row and header info
         CopyMemory(@corn,cornstr,HEADERBYTES)
         if MyRow = 0 and MyCol=0 THEN goto Done2
         if MyRow = 0 THEN
             rescell.MyFont = corn.MyFont
             rescell.MyCase = corn.MyCaseR
             rescell.MyJust = corn.MyJustR
             rescell.MyForeColor = corn.MyForeColor
             rescell.MyBackColor = corn.MyBackColor
           else 'MyCol = 0
             rescell.MyFont = corn.MyFont
             rescell.MyCase = corn.MyCaseC
             rescell.MyJust = corn.MyJustC
             rescell.MyForeColor = corn.MyForeColor
             rescell.MyBackColor = corn.MyBackColor
         END IF
         
         CopyMemory(Mycell,@rescell,HEADERBYTES)
         
       else
           Goto Done2
      end if 'ttt
   end if ' myrow mycol
  end if 'v
   
   Done2:
   
END sub    

    
'Note GRD_FormatAttribute will exit if a cell does not have its header defined by a string input    
Function GRD_FormatAttribute(hWnd as hWnd,myrow as long,mycol as long,Action as long,newvalue as LongInt=0,WhichSheet as long = -1) as ULongInt
    Dim CellType as long
    Dim v As GridData Ptr
    Dim ttt As Integer Ptr
    Dim str2 as wstring ptr
    Dim MyWstrA as wstring * 9                'Attributes 8 bytes
    Dim MyWstrC as wstring * GRD_MAXCELLLEN   'visible in the cell
    Dim MyWstrS as wstring * GRD_MAXCELLLEN   'full string

    Dim j As Long, i As Long, k as long,dimrows as long,dimcols as long
    Dim MyCS as long,TheElement as long
    
    dim row as RowFormat1
    dim col as ColFormat2
    dim cell as CellFormat3
    Dim corn as CornerFormat0
   
   v = GetProp(hWnd,"GridData")
   if v <> 0 then
       if WhichSheet = -1 then MyCS = v->CS  else MyCS = WhichSheet
       if MyCS > v->TotalSheets then exit function
       dimrows = v->SH(MyCS).DimRows
       dimcols = v->SH(MyCS).DimCols
       if myrow < 0 then exit function
       if mycol < 0 then exit function 
       if myrow > dimrows then exit function
       if mycol > dimcols then exit function
       
       if myrow = 0 and mycol = 0 then CellType = 0 ' the upper left corner cell
       if myrow = 0 and mycol > 0 then CellType = 1 ' a col header
       if myrow > 0 and mycol = 0 then CellType = 2 ' a row header
       if myrow > 0 and mycol > 0 then CellType = 3 ' a cell
       
       TheElement=(myrow * (v->SH(MyCS).DimCols +1) ) + (mycol)
       
       'if the string is missing, make a header
       Dim ttt as GRDArray Pointer = cast(GRDArray Ptr,v->SH(MyCS).ws)
       If ttt <> 0 Then
          str2=ttt[TheElement].ws
          if str2 = 0 THEN
              ttt[TheElement].ws = cast(wstring ptr,CAllocate((9) * Len(WString)))  ' 8 for header 1 null
             ' *ttt[TheElement].ws = space(8)     causing some problems but don't understand
          end if
       end if   
       
       'If a cell does not have a header then exit (the above routine failed
       'some fields are common across formats so format types are interchanged
       If ttt <> 0 Then
          str2=ttt[TheElement].ws
          if str2 <> 0 THEN   'copy first 8 wstr chars to 4 different header type for flexibility
             CopyMemory(@row,str2,HEADERBYTES)
             CopyMemory(@col,str2,HEADERBYTES)
             CopyMemory(@cell,str2,HEADERBYTES)
             CopyMemory(@corn,str2,HEADERBYTES)
            else   'for now a cell has to have a string put before a format can be adjusted
             exit function
           END IF
         else  
          exit function   
       end if 'ttt 
       
      
         
      Select Case as const Action
         
         CASE GRDFORMAT_SETROWHEIGHT
            if CellType = 2 THEN    
               if newvalue > GRD_ROWCOLMIN and newvalue < 256 then 
                  row.MyHeight = cUbyte(newvalue)
                  row.RHide = 0
                  CopyMemory(str2,@row,HEADERBYTES)
               end if 
               exit function 
            END IF
            if CellType = 0 THEN 
               If newvalue > 255 THEN newvalue = 255 
               if newvalue > GRD_ROWCOLMIN and newvalue < 256 then 
                  corn.MyHeight = cUbyte(newvalue)
                  CopyMemory(str2,@corn,HEADERBYTES)
               end if
               exit function   
            END IF
             
         CASE GRDFORMAT_GETROWHEIGHT
               if CellType = 2 THEN      
                  Return cast(UlongInt,row.MyHeight)
               END IF
                if CellType = 0 THEN      
                  Return cast(UlongInt,corn.MyHeight)
               END IF
               
              ' print  "Jimmy" & str(row.MyHeight)
                
         CASE GRDFORMAT_SETCOLWIDTH 
            if CellType = 1 THEN
               If newvalue > 255 THEN newvalue = 255    
               if newvalue > GRD_ROWCOLMIN and newvalue < 256 then 
                  col.MyWidth = cUbyte(newvalue)
                  col.CHide = 0
                  CopyMemory(str2,@col,HEADERBYTES)
               end if 
               exit function  
            END IF
            if CellType = 0 THEN
               If newvalue > 255 THEN newvalue = 255    
               if newvalue > GRD_ROWCOLMIN and newvalue < 256 then 
                  corn.MyWidth = cUbyte(newvalue)
                  CopyMemory(str2,@corn,HEADERBYTES)
                  'print "Row Header Width"
               end if
               exit function   
            END IF
             
         CASE GRDFORMAT_GETCOLWIDTH
            if CellType = 1 THEN      
               Return cast(UlongInt,col.MyWidth)
            END IF
            if CellType = 0 THEN      
               Return cast(UlongInt,corn.MyWidth)
            END IF
         
         case GRDFORMAT_SETCELLBORDER
            if CellType = 3 THEN    
               if newvalue > -1 and newvalue < 256 then 
                  cell.MyBorder = cUbyte(newvalue)
                  CopyMemory(str2,@cell,HEADERBYTES)
               end if   
            END IF
             
         case GRDFORMAT_GETCELLBORDER
            if CellType = 3 THEN      
               Return cast(UlongInt,cell.MyBorder)
            END IF
               
         CASE GRDFORMAT_SETWRITELOCK   'WriteLock is the same for row col cell so use cell
               if CellType = 0 THEN goto done
               if newvalue > -1 and newvalue < 256 then 
                  cell.WriteLock = cUbyte(newvalue)
                  CopyMemory(str2,@cell,HEADERBYTES)
               end if 
                
         CASE GRDFORMAT_GETWRITELOCK    'WriteLock is the same for row col cell so use cell
               if CellType = 0 THEN goto done
                Return cast(UlongInt,cell.WriteLock)
                
         CASE GRDFORMAT_SETROWHIDE  'added code to store previous row height
                if CellType = 2 THEN
                   if row.Rhide <> 0 then goto done
                     Row.RHide = row.MyHeight 
                     Row.MyHeight = 0
                     CopyMemory(str2,@row,HEADERBYTES) 
                end if
                
         CASE GRDFORMAT_SETROWUNHIDE  'added code to store previous row height
                if CellType = 2 THEN
                   if row.Rhide = 0 then goto done
                     row.MyHeight = Row.RHide  
                     Row.RHide = 0
                     CopyMemory(str2,@row,HEADERBYTES) 
                end if
                
         CASE GRDFORMAT_GETROWHIDE  'added code to store previous row height
               if CellType = 2 THEN
                     Return cast(UlongInt,row.RHide)
               end if 
               
         CASE GRDFORMAT_SETCOLHIDE  'added code to store previous col width
                if CellType = 1 THEN
                   if col.Chide <> 0 then goto done
                     col.CHide = col.MyWidth 
                     col.MyWidth = 0
                     CopyMemory(str2,@col,HEADERBYTES) 
                end if 
                
         CASE GRDFORMAT_SETCOLUNHIDE  'added code to store previous col width
                if CellType = 1 THEN
                   if col.Chide = 0 then goto done
                     col.MyWidth = col.CHide 
                     col.CHide = 0
                     CopyMemory(str2,@col,HEADERBYTES) 
                end if    
                
         CASE GRDFORMAT_GETCOLHIDE  'added code to store previous col width 
               if CellType = 1 THEN
                     Return cast(UlongInt,col.CHide)
               end if  
               
         CASE GRDFORMAT_SETNDFORMAT  'cell number date format is in a different place on row
               if CellType = 0 THEN goto done
               IF CellType = 2 THEN
                  if newvalue > -1 and newvalue < 256 then 
                    row.CellFormat = cUbyte(newvalue)
                    CopyMemory(str2,@row,HEADERBYTES)
                  end if
                 else 'cell or col
                  if newvalue > -1 and newvalue < 256 then 
                    Cell.CellFormat = cUbyte(newvalue)
                    CopyMemory(str2,@cell,HEADERBYTES)
                  end if
               END IF
                 
         CASE GRDFORMAT_GETNDFORMAT  'cell number date format is in a different place on row
               if CellType = 0 THEN goto done
               IF CellType = 2 THEN
                  Return cast(UlongInt,Row.CellFormat)
                 else
                  Return cast(UlongInt,cell.CellFormat)
               END IF
                
         CASE GRDFORMAT_SETTYPE   'Cell Type is the same for row col cell so use cell
               if CellType = 0 THEN goto done
               if newvalue > -1 and newvalue < 256 then 
                  cell.CellType = cUbyte(newvalue)
                  CopyMemory(str2,@cell,HEADERBYTES)
               end if
                
         CASE GRDFORMAT_GETTYPE   'Cell Type is the same for row col cell so use cell
                if CellType = 0 THEN goto done
                Return cast(UlongInt,cell.CellType)
               
         CASE GRDFORMAT_SETFONT   'cell font case and just are for cell and column only use cell   
               if CellType = 2 THEN goto done
               'Corner OK on this
               if newvalue > -1 and newvalue < 256 then 
                  cell.MyFont = cUbyte(newvalue)
                  CopyMemory(str2,@cell,HEADERBYTES)
               end if 
                
         CASE GRDFORMAT_GETFONT   'cell font case and just are for cell and column only use cell 
               if CellType = 2 THEN goto done
               'Corner OK on this
               Return cast(UlongInt,cell.MyFont)
                
         CASE GRDFORMAT_SETCASE   'cell font case and just are for cell and column only  use cell    
               if CellType = 2 THEN goto done
               if CellType = 0 THEN goto done
               if newvalue > -1 and newvalue < 256 then 
                  cell.MyCase = cUbyte(newvalue)
                  CopyMemory(str2,@cell,HEADERBYTES)
               end if 
                goto Done
         CASE GRDFORMAT_GETCASE   'cell font case and just are for cell and column only use cell       
               if CellType = 2 THEN goto done
               if CellType = 0 THEN goto done
               Return cast(UlongInt,cell.MyCase)
                
         CASE GRDFORMAT_SETJUST   'cell font case and just are for cell and column only use cell 
               if CellType = 2 THEN goto done
               if CellType = 0 THEN goto done
               if newvalue > -1 and newvalue < 256 then 
                  cell.MyJust = cUbyte(newvalue)
                  CopyMemory(str2,@cell,HEADERBYTES)
               end if 
                
         CASE GRDFORMAT_GETJUST   'cell font case and just are for cell and column only  use cell 
               if CellType = 2 THEN goto done
               if CellType = 0 THEN goto done
               Return cast(UlongInt,cell.MyJust)
               
         CASE GRDFORMAT_SETFORECLR  'Foreground color is the same for row col cell so use cell
               if cast(uBYTE,newvalue) > -1 and cast(ubyte,newvalue) < 256 then 
                  cell.MyForeColor = cUbyte(newvalue)
                  CopyMemory(str2,@cell,HEADERBYTES)
               end if
                
         CASE GRDFORMAT_GETFORECLR  ' Foreground color is the same for row col cell so use cell
                Return cast(UlongInt,cell.MyForeColor)
                
         CASE GRDFORMAT_SETBACKCLR  'Background color is the same for row col cell so use cell
               if newvalue > -1 and newvalue < 256 then 
                  cell.MyBackColor = cUbyte(newvalue)
                  CopyMemory(str2,@cell,HEADERBYTES)
               end if 
                
         CASE GRDFORMAT_GETBACKCLR  'Background color is the same for row col cell so use cell
                Return cast(UlongInt,cell.MyBackColor)
                  
         CASE GRDFORMAT_SETROWRECNO
               if CellType = 2 THEN    
                  row.RecNo = newvalue
                  CopyMemory(str2,@row,HEADERBYTES)
               END IF
                
         CASE GRDFORMAT_GETROWRECNO
               if CellType = 2 THEN      
                  Return cast(UlongInt,row.RecNo)
               END IF
            
         case GRDFORMAT0_SETCASEC
            if CellType = 0 THEN      
                if newvalue > -1 and newvalue < 256 then 
                corn.MyCaseC = cUbyte(newvalue)
                CopyMemory(str2,@corn,HEADERBYTES)
               end if
            END IF
         case GRDFORMAT0_SETJUSTC
            if CellType = 0 THEN      
               if newvalue > -1 and newvalue < 256 then 
                 corn.MyJustC = cUbyte(newvalue)
                 CopyMemory(str2,@corn,HEADERBYTES)
               END IF
            end if
         case GRDFORMAT0_SETCASER
            if CellType = 0 THEN      
               if newvalue > -1 and newvalue < 256 then 
                 corn.MyCaseR = cUbyte(newvalue)
                 CopyMemory(str2,@corn,HEADERBYTES)
               END IF
            end if   
         case GRDFORMAT0_SETJUSTR
            if CellType = 0 THEN      
               if newvalue > -1 and newvalue < 256 then 
                 corn.MyJustR = cUbyte(newvalue)
                 CopyMemory(str2,@corn,HEADERBYTES)
               END IF
            end if
         case GRDFORMAT0_GETCASEC
            if CellType = 0 THEN      
               Return cast(UlongInt,corn.MyCaseC)
            END IF
         case GRDFORMAT0_GETJUSTC
            if CellType = 0 THEN      
               Return cast(UlongInt,corn.MyJustC)
            END IF
         case GRDFORMAT0_GETCASER
            if CellType = 0 THEN      
                Return cast(UlongInt,corn.MyCaseR)
            END IF
         case GRDFORMAT0_GETJUSTR          
            if CellType = 0 THEN      
              Return cast(UlongInt,corn.MyJustR) 
            END IF
            
         Case  GRDFORMAT0_SETROWFORMAT   'CellFormatC
            if CellType = 0 THEN      
               if newvalue > -1 and newvalue < 256 then 
                 corn.CellFormatR = cUbyte(newvalue)
                 CopyMemory(str2,@corn,HEADERBYTES)
               END IF
            end if
                 
         case  GRDFORMAT0_SETCOLFORMAT   'CellFormatC
            if CellType = 0 THEN      
               if newvalue > -1 and newvalue < 256 then 
                 corn.CellFormatC = cUbyte(newvalue)
                 CopyMemory(str2,@corn,HEADERBYTES)
               END IF
            end if
            
     '        if myrow = 0 and mycol = 0 then CellType = 0 ' the upper left corner cell
     '  if myrow = 0 and mycol > 0 then CellType = 1 ' a col header
     '  if myrow > 0 and mycol = 0 then CellType = 2 ' a row header
     '  if myrow > 0 and mycol > 0 then CellType = 3 ' a cell
            
         Case GRDFORMAT_CLEARFORMATS        'EveryThing but RowHeight ColWidth RowHide and ColHide   
             if CellType = 0 THEN 'Corner      
                 corn.MyForeColor = 0 : corn.MyBackColor = 0 : corn.MyFont = 0   : corn.MyJustR = 0  : corn.MyJustC = 0
                 corn.MyCaseR = 0 : corn.MyCaseC = 0 : corn.CellFormatC = 0   : corn.CellFormatR = 0 
                 CopyMemory(str2,@corn,HEADERBYTES)
            end if
            if CellType = 1 THEN  'Col    
                 col.MyForeColor = 0 : col.MyBackColor = 0 : col.MyFont = 0   : col.MyJust = 0 
                 col.MyCase = 0  : col.CellFormat = 0  : col.WriteLock = 0  : col.CellType = 0  : col.ColCombo  = 0
                 CopyMemory(str2,@col,HEADERBYTES)
            end if
            if CellType = 2 THEN  'Row      
                 row.MyForeColor = 0 : row.MyBackColor = 0 : row.CellType = 0  : row.RecNo = 0
                 row.CellFormat = 0   : row.WriteLock = 0  : row.RowCombo = 0
                 CopyMemory(str2,@row,HEADERBYTES)
            end if 
            if CellType = 3 THEN 'Cell     
                 cell.MyForeColor = 0 : cell.MyBackColor = 0 : cell.MyFont = 0   : cell.MyJust = 0  : cell.MyBorder = 0
                 cell.MyCase = 0 : cell.WriteLock = 0 : cell.CellFormat = 0   : cell.CellTYpe = 0 
                 CopyMemory(str2,@cell,HEADERBYTES)
            end if 
         CASE ELSE
            
      End Select  
   end if 'v
 
 Done:
END Function

FUNCTION GRD_DecToBase26(BYVAL lVal AS UINT) AS STRING 'Need to convert this to Wstr for use on grid
    DIM sVal AS STRING
    DIM flag AS LONG

    sVal = ""

    DO WHILE lVal > 26
        flag= lVal MOD 26
        IF flag > 0 THEN
            sVal = sVal & GRD_DecToBase26(lVal \ 26)
            lVal = lVal - 26 * ((lVal) \ 26)
           ELSE
            sVal = sVal & GRD_DecToBase26((lVal-1) \ 26) & "Z"
            lVal = lVal - 26 * ((lVal) \ 26)
        END IF
    LOOP

      sVal = sVal & MID(DEF_BASE26, lVal, 1)
    RETURN sVal
END FUNCTION 

SUB GRD_InitializeRowHeaders(hWnd as HWND,TheHeight as long = -1,MySheet as long = -1)
   'Input A B C ... and 50 width in an initialized format block ( 5 wide chars)
   'If MySheet <> -1 then just initialize MySheet
   Dim v As GridData Ptr
    'Dim MyWstr as wstring * 16
    Dim r as RowFormat1
    Dim MyRow As Long,rows As Long,cols As Long,j As Long, TotalSheets As Long
    Dim TheElement As Long
    'DIM MyWPtr as WSTRING PTR
    DIM MyCol as LONG
    Dim MyWidth as uBYTE
    dim StartSheet as long
    dim tempwstr as wstring * 8
    Dim MyLen as LONG
    dim MyHeight as Long 
    
    'MyWPtr = VarPtr(MyWstr) + 8  'get past the header
    
    v = GetProp(hWnd,"GridData")
    StartSheet = 0
    If TheHeight < 0 THEN
        MyHeight = v->cyChar
      else
        MyHeight = TheHeight   
    END IF
    
    If v <> 0 Then
      TotalSheets = v->TotalSheets
      
      If MySheet <> -1 THEN  'Only do one sheet - not all sheets
        If MySheet > -1 and MySheet < TotalSheets + 1 THEN
            StartSheet = MySheet
            TotalSheets = MySheet
           else
            exit sub
        END IF 
      END IF
      
      
      for j = StartSheet to TotalSheets 
         rows = v->SH(j).DimRows
         for MyRow = 1 to rows  'Initialize Columns will take care of the Corner Cell
            
            r.MyHeight = MyHeight
            r.CellFormat = 0
            r.RowCombo = 0
            r.RHide = 0
            r.MyForeColor = 0
            r.MyBackColor = 0
            r.CellType = 0
            r.WriteLock = 0
            r.RecNo = 0
           
            MyCol = 0
            TheElement=(MyRow * (v->SH(j).DimCols +1) ) + (MyCol)
            
            Dim ttt as GRDArray Pointer = cast(GRDArray Ptr,v->SH(j).ws)
            If ttt <> 0 Then
               if ttt[TheElement].ws <> 0 then  'need deallocate previous column header string
                  Deallocate ttt[TheElement].ws
               end if
               tempwstr=trim(wstr(Myrow))
               MyLen=len(tempwstr)   
               ttt[TheElement].ws = cast(wstring ptr,CAllocate((MyLen+10) * Len(WString)))  '10 is 1 extra - 8 for header 1 null
               *ttt[TheElement].ws = space(8) & tempwstr
               CopyMemory(ttt[TheElement].ws,@r,HEADERBYTES) 'if there was a header then restore it  if not c should be all zeroes 
            end if
         next MyRow   
      next j 
    end if

END SUB

SUB GRD_InitializeColHeaders(hWnd as HWND,TheWidth as long=-1,MySheet as long = -1)
  'Col 0 will always be initialized to 25
  'Input A B C ... and 40 width in an initialized format block ( 8 wide chars)
  'If MySheet <> -1 then just initialize MySheet
    Dim v As GridData Ptr
    Dim MyWstr as wstring * 12
    Dim c as ColFormat2
    Dim MyCol As Long,rows As Long,cols As Long,j As Long, TotalSheets As Long
    Dim TheElement As Long, MyHeight as UBYTE
    DIM MyWPtr as WSTRING PTR
    DIM StartSheet as long
    dim tempwstr as wstring * 5
    Dim MyLen as LONG
    Dim MyWidth as Long
    
    v = GetProp(hWnd,"GridData")
    StartSheet = 0
     if TheWidth < 0 THEN
        MyWidth = 80
      else
        MyWidth = TheWidth   
     END IF
    
    If v <> 0 Then
      TotalSheets = v->TotalSheets
      
      If MySheet <> -1 THEN  'Only do one sheet - not all sheets
        If MySheet > -1 and MySheet < TotalSheets + 1 THEN
            StartSheet = MySheet
            TotalSheets = MySheet
           else
            exit sub
        END IF 
      END IF
            
      for j = StartSheet to TotalSheets 
         cols = v->SH(j).DimCols
         for MyCol = 0 to cols  
            c.MyWidth = MyWidth
            c.CHide = 0 
            c.CellFormat = 0
            c.ColCombo = 0
            c.MyForeColor = 0
            c.MyBackColor = 0
            c.WriteLock = 0
            c.MyFont = 0
            c.MyJust = 0
            c.MyCase = 0
            c.Unused4 = 0
            c.Unused5 = 0
            c.Unused6 = 0
            c.Unused7 = 0
            c.Unused8 = 0
            
            if MyCol = 0 then 'This is the corner - need to alias corner 
                c.MyWidth = 50     'Same on CornerFormat0
                c.MyFont = 0       'Same on CornerFormat0
                c.MyJust = 0       'MyJustR on CornerFormat0
                c.MyCase = 0       'MyJustC on CornerFormat0
                c.CellFormat = v->cyChar  'MyHeight on CornerFormat0
                c.MyForeColor = 0  'Same on CornerFormat0
                c.MyBackColor = 0  'Same on CornerFormat0
                c.Unused4 = 0      'MyCaseR on CornerFormat0
                c.Unused5 = 0      'MyCaseC CornerFormat0
            end if    
            rows = 0
            TheElement=(rows * (v->SH(j).DimCols+1) ) + (MyCol)
            Dim ttt as GRDArray Pointer = cast(GRDArray Ptr,v->SH(j).ws)
            If ttt <> 0 Then
               if ttt[TheElement].ws <> 0 then  'need deallocate previous column header string
                  Deallocate ttt[TheElement].ws
               end if
               tempwstr=GRD_DecToBase26(MyCol)
               if MyCol = 0  THEN tempwstr = "   "  'Corner cell is always blank    
               MyLen=len(tempwstr)   
               ttt[TheElement].ws = cast(wstring ptr,CAllocate((MyLen+10) * Len(WString)))  '10 is 1 extra - 8 for header 1 null
               *ttt[TheElement].ws = space(8) & tempwstr
               CopyMemory(ttt[TheElement].ws,@c,HEADERBYTES) 'if there was a header then restore it  if not c should be all zeroes
            end if 'ttt
         next MyCol   
      next j 'Mysheet 
    end if
    
END SUB


       
'********************************************************************
' Cell Array procedures
'********************************************************************

Function GRD_ClearCells(HWND as HWND,MyViewRows as LONG,MyViewCols as long, WhichSheet as long = -1) as LONG
    Dim v As GridData Ptr
    Dim j as LONG
    
   'WhichSheet = -1 is the current sheet. 
    v = GetProp(hWnd,"GridData")
    if v <> 0 then
       if WhichSheet = -1 then j = v->CS else j = WhichSheet 
       if j < 0 or j > v->TotalSheets THEN exit FUNCTION
       if MyViewRows > v->SH(j).DimRows then Exit FUNCTION 
       if MyViewRows < 0 then Exit FUNCTION
       if MyViewCols > v->SH(j).DimCols then Exit FUNCTION 
       if MyViewCols < 0 then Exit FUNCTION
       
      
    end if    
 function = 0  
END FUNCTION 

function GRD_InsertRowCol(HWND as HWND,MyViewRows as LONG,MyViewCols as long, WhichSheet as long = -1) as LONG
    Dim v As GridData Ptr
    Dim j as LONG
    
   'WhichSheet = -1 is the current sheet. 
    v = GetProp(hWnd,"GridData")
    if v <> 0 then
       if WhichSheet = -1 then j = v->CS else j = WhichSheet 
       if j < 0 or j > v->TotalSheets THEN exit FUNCTION
       if MyViewRows > v->SH(j).DimRows then Exit FUNCTION 
       if MyViewRows < 0 then Exit FUNCTION
       if MyViewCols > v->SH(j).DimCols then Exit FUNCTION 
       if MyViewCols < 0 then Exit FUNCTION
       
       
    end if    
Function = 0   
END FUNCTION

Function GRD_DeleteRowCol(HWND as HWND,MyViewRows as LONG,MyViewCols as long, WhichSheet as long = -1) as LONG
    Dim v As GridData Ptr
    Dim j as LONG
    
   'WhichSheet = -1 is the current sheet. 
    v = GetProp(hWnd,"GridData")
    if v <> 0 then
       if WhichSheet = -1 then j = v->CS else j = WhichSheet 
       if j < 0 or j > v->TotalSheets THEN exit FUNCTION
       if MyViewRows > v->SH(j).DimRows then Exit FUNCTION 
       if MyViewRows < 0 then Exit FUNCTION
       if MyViewCols > v->SH(j).DimCols then Exit FUNCTION 
       if MyViewCols < 0 then Exit FUNCTION
       
    end if    
 Function = 0  
END FUNCTION

function GRD_AppendRows(HWND as HWND,MyViewRows as LONG,MyViewCols as long, WhichSheet as long = -1) as LONG
    Dim v As GridData Ptr
    Dim j as LONG
    
   'WhichSheet = -1 is the current sheet. 
    v = GetProp(hWnd,"GridData")
    if v <> 0 then
       if WhichSheet = -1 then j = v->CS else j = WhichSheet 
       if j < 0 or j > v->TotalSheets THEN exit FUNCTION
       if MyViewRows > v->SH(j).DimRows then Exit FUNCTION 
       if MyViewRows < 0 then Exit FUNCTION
       if MyViewCols > v->SH(j).DimCols then Exit FUNCTION 
       if MyViewCols < 0 then Exit FUNCTION
       
      
    end if    
  Function = 0 
END FUNCTION

Sub GRD_SetView(HWND as HWND,MyViewRows as LONG,MyViewCols as long, WhichSheet as long = -1)
 'TotRows TotCols
    Dim v As GridData Ptr
    Dim j as LONG
    
   'WhichSheet = -1 is the current sheet. 
    v = GetProp(hWnd,"GridData")
    if v <> 0 then
       if WhichSheet = -1 then j = v->CS else j = WhichSheet 
       if j < 0 or j > v->TotalSheets THEN exit sub
       if MyViewRows > v->SH(j).DimRows then Exit sub 
       if MyViewRows < 0 then Exit sub
       if MyViewCols > v->SH(j).DimCols then Exit sub 
       if MyViewCols < 0 then Exit sub
       
       v->SH(j).TotRows = MyViewRows 
       v->SH(j).TotCols = MyViewCols 
    end if     
END SUB

Sub GRD_GetView(HWND as HWND,MyViewRows as LONG PTR,MyViewCols as long PTR, WhichSheet as long = -1)
 'TotRows TotCols
    Dim v As GridData Ptr
    Dim j as LONG
    
    'WhichSheet = -1 is the current sheet. 
    v = GetProp(hWnd,"GridData")
    if v <> 0 then
       if WhichSheet = -1 then j = v->CS else j = WhichSheet 
       if j < 0 or j > v->TotalSheets THEN exit sub
       if MyViewRows > v->SH(j).DimRows then Exit sub 
       if MyViewRows < 0 then Exit sub
       if MyViewCols > v->SH(j).DimCols then Exit sub 
       if MyViewCols < 0 then Exit sub
       
       *MyViewRows = v->SH(j).TotRows 
       *MyViewCols = v->SH(j).TotCols
    end if     
END SUB


Function GRD_InitializeCellsAllSheet(hWnd As HWND) As Integer
    ' Allocate memory for 10 rows and 5 cols for all sheets.
    Dim ttt As Integer Ptr
    Dim As WString Ptr str2
    Dim v As GridData Ptr
    Dim totelements As Long,i As Long,rows As Long,cols As Long,j As Long, TotalSheets As Long 
    
    rows = 10 : cols = 5 
    totelements=(rows + 1) * (cols + 1) 'Have to include the zero row and cols which are the headers
    
    v = GetProp(hWnd,"GridData")
    'Check for a pointer to an array and clear it and zero it
    If v <> 0 Then
      TotalSheets = v->TotalSheets
      For j = 0 To TotalSheets
    
        ttt = Cptr(Integer Ptr, CAllocate((totelements+4) * Sizeof(Integer)))   'add 4 safety 
        v->SH(j).ws = ttt
        
        v->SH(j).DimRows = rows
        v->SH(j).DimCols = cols
        v->SH(j).TotRows = rows
        v->SH(j).TotCols = cols
        v->SH(j).TopRow = 1
        v->SH(j).LeftCol = 1
        v->SH(j).SelRow = 1
        v->SH(j).SelCol = 1
    
      Next j 
    End If
    
    j=j-1
    
    GRD_InitializeColHeaders HWnd
    GRD_InitializeRowHeaders HWND
    
    Return Cast(Integer,j)
End Function

Function GRD_ReDimCellsSheet(hWnd As HWND, rows As Long, cols As Long,WhichSheet as long = -1) As Integer
    ' Allocate memory for current sheet cell strings.
    Dim ttt As integer Ptr
    Dim As WString Ptr str2
    Dim v As GridData Ptr
    Dim totelements As Long,i As Long,totelementscurrent as long
    Dim j as LONG
    
    'WhichSheet = -1 is the current sheet. 
   
    v = GetProp(hWnd,"GridData")
    if v <> 0 then
       if WhichSheet = -1 then j = v->CS else j = WhichSheet 
       if j < 0 or j> v->TotalSheets THEN exit function
       totelementscurrent  = (v->SH(j).DimRows + 1) * (v->SH(j).DimCols + 1)
       totelements=(rows + 1) * (cols + 1)
    
       'Check for a pointer to an array and clear it and zero it   
       ttt=v->SH(j).ws
       If ttt <> 0 Then
           For i = 0 To totelementscurrent -1
             If ttt[i] <> 0 Then
                str2 = Cast(WString Ptr,ttt[i])
                Deallocate str2
                ttt[i] = 0                  
             End If                 
           Next i
           Deallocate v->SH(j).ws
           v->SH(j).ws=0
       End If
    End If
     
    ttt = Cptr(Integer Ptr, CAllocate((totelements+4) * Sizeof(Integer)))   'add 4 safety
    v->SH(j).ws =ttt 
    
    v->SH(j).DimRows = rows
    v->SH(j).DimCols = cols
    v->SH(j).TotRows = rows
    v->SH(j).TotCols = cols
    v->SH(j).TopRow  = 1
    v->SH(j).LeftCol = 1
    
    GRD_InitializeColHeaders HWnd,-1,j
    GRD_InitializeRowHeaders HWND,-1,j
  
    Return cast(integer,GS_.ws)
End Function

Function GRD_EraseCellsAllSheets(hWnd As HWND) As Integer
    ' We are destroying the grid control some release all the cell memory on all sheets
    Dim ttt As Integer Ptr
    Dim As WString Ptr str2
    Dim v As GridData Ptr
    Dim totelements As Long,i As Long,j As Long, TotalSheets As Long 
    
    v = GetProp(hWnd,"GridData")
    If v <> 0 Then
    
      TotalSheets = v->TotalSheets
      For j = 0 To TotalSheets
          'v->CS = j 'cycle thru the sheets
          totelements = v->SH(j).DimRows* v->SH(j).DimCols
          ttt=v->SH(j).ws
          If ttt <> 0 Then 
           For i = 0 To totelements - 1
             If ttt[i] <> 0 Then
                str2 = Cast(WString Ptr,ttt[i])
                Deallocate str2 : ttt[i] = 0
                ttt[i] = 0 
             End If                   
           Next i 'elements
           Deallocate v->SH(j).ws  : v->SH(j).ws = 0
         End If   
       Next j  'Sheets
    End If 'v
    GS_.DimRows = 0
    GS_.DimCols = 0
    GS_.TotRows = 0
    GS_.TotCols = 0
    Return Cast(Integer,v)
End Function

'**********s**********************************************************
' Cell Selection Change & MouseMove Update procedures
'********************************************************************
'Grid data structure gets update

function GRD_MouseMoveUpdate(hwND AS hWND,wParam as wParam,lParam as lParam) as long

    dim skey as long,x as long, y as long
    DIM p AS POINT
    dim result as Long , temp as long, temp1 as Long,temp2 as long ,myflag as Long
    DIM xRightSide as long,xLeftSide as long, yBottomSide as Long,yTopSide as Long
    Dim v As GridData Ptr
    dim MyCol as long, MyLeftCol as long,MyRow as long, MyTopRow as long
    dim ButtonHeight as Long

    'result passes back 2 if in grid but in the row header and 1 if in grid but in the col header
    
    

    skey=GetKeyState(VK_SHIFT) AND &H8000

    x=LOWoRD(lParam)
    y=HIWoRD(lParam)
    p.x = x
    p.y = y
    
    'p.x = px
    'p.y = py
    
   ' Print "MouseMove Update" & str(x) & " : " & str(y) & " : " & str(GetFocus)
    
   v = GetProp(hWnd,"GridData")
   if v <> 0 then 
   'result 0 is we are on cells of the grid
   IF x < GS_.GridLeft THEN  'we are left of the grid
           'result 3 is off the grid to the left
           result = 3
        ELSE 'we are on the grid
            'result 1 is row header
            IF x < GRD_FormatAttribute(hwnd,0,0,GRDFORMAT_GETCOLWIDTH) + GS_.GridLeft THEN result = 1 
            MyCol = GS_.LeftCol
            temp = GRD_FormatAttribute(hwnd,0,0,GRDFORMAT_GETCOLWIDTH) + GS_.GridLeft
            
            FOR temp1=MyCol TO GS_.TotCols
                  temp2=temp
                  temp=temp+GRD_FormatAttribute(hwnd,0,temp1,GRDFORMAT_GETCOLWIDTH)-1
                  IF temp > x THEN exit for
                  MyCol = MyCol + 1
                  MyCol=MIN(MyCol,GS_.TotCols)
            NEXT temp1
        
            xRightSide = temp
            xLeftSide = temp2
   END IF
    
   IF y < GS_.GridTop THEN  'we are in the row header
          'result 4 is off the grid to the top
          result = 4
        ELSE 'we are on the grid
            IF y < GRD_FormatAttribute(hwnd,0,0,GRDFORMAT_GETROWHEIGHT) + GS_.GridTop THEN
               'result 5 is corner  result 2 is column header
                if result = 1 then result = 5 else result = 2
            end if    
            MyRow = GS_.TopRow
            temp = GRD_FormatAttribute(hwnd,0,0,GRDFORMAT_GETROWHEIGHT) + GS_.GridTop
           
            FOR temp1=MyRow TO GS_.TotRows
                  temp2 = temp 
                  temp=temp+GRD_FormatAttribute(hwnd,temp1,0,GRDFORMAT_GETROWHEIGHT)
                  IF temp > y THEN exit for
                  MyRow = MyRow +1
                  MyRow=MIN(MyRow,GS_.TotRows)
            NEXT temp1
            yBottomSide = temp
            yTopSide = temp2
   END IF
    
    'v->AllRowCol will be determined by lbuttondown in mousemove
    IF result = 5 THEN GS_.AllRowColtemp= 1 ELSE GS_.AllRowColtemp= 0
    GS_.OldMyRowTemp = GS_.MyRowTemp
    GS_.OldMyColTemp = GS_.MyColTemp
    GS_.MyRowTemp = MyRow
    GS_.MyColTemp = MyCol                     
    
               
    temp=GRD_FormatAttribute(hwnd,MyRow,0,GRDFORMAT_GETROWHEIGHT)
    ButtonHeight = 17
    IF temp < 17 THEN ButtonHeight = 15
    IF temp < 15 THEN ButtonHeight = 13
    IF temp < 13 THEN ButtonHeight = 11
    IF temp < 11 THEN ButtonHeight = 9
    IF temp < 9 THEN ButtonHeight = 7
  
    GS_.TempCUBRect.Right = xRightSide
    GS_.TempCUBRect.Bottom = yBottomSide
    GS_.TempCUBRect.Top = GS_.TempCUBRect.Bottom - ButtonHeight
    GS_.TempCUBRect.Left = GS_.TempCUBRect.Right - ButtonHeight
    
    GS_.TempSelectRect.Right = xRightSide
    GS_.TempSelectRect.Bottom = yBottomSide
    GS_.TempSelectRect.Left = xLeftSide-1
    GS_.TempSelectRect.Top = yTopSide
   
  end if 'v
  function = result 
END function

SUB GRD_CellSelectionUpdate (Hwnd as hwnd)
       'IF  (@v.SelRow <> @v.OldSelRow) OR (@v.SelCol <> @v.OldSelCol) THEN
                                                                            
       'end if
        
        '@v.SetWindowPosFlag=1   'This will forces a mousemove and autoscrolling window update
       ' IF  (@v.TopRow <> @v.OldTopRow) OR (@v.LeftCol <> @v.OldLeftCol) OR (@v.TopRow2 <> @v.OldTopRow2) OR (@v.LeftCol2 <> @v.OldLeftCol2) THEN
       '     @v.SetWindowPosFlag=1   'This will force a window update
       '     @v.OldTopRow = @v.TopRow : @v.OldLeftCol = @v.LeftCol
       '     @v.OldTopRow2 = @v.TopRow2 : @v.OldLeftCol2 = @v.LeftCol2
       ' END IF  
end Sub

'********************************************************************
' Listbox Contents Transfer procedures
' WStrings are comma delimited for the fake combobox
'********************************************************************
Sub GRD_PutListing(hWnd As HWND,WhichStr as long, MyWstr As WString Ptr)
    ' Put to a new listing string in MyWstr.
    Dim ttt As Integer Ptr,sss As Integer Ptr
    Dim v As GridData Ptr
    Dim MyLen as long
    dim tempwstr as Wstring * GRD_MAXCELLLEN
    
    if WhichStr > GRD_NUMLISTS or WhichStr < 0 THEN  Exit Sub
       
       
    v = GetProp(hWnd,"GridData")
       
       
    'Initialize the Liststrings if this is the first time thru
    if v->ListStrings = 0 and v <> 0 THEN
       sss  = Cptr(Integer Ptr, CAllocate((GRD_NUMLISTS+4) * Sizeof(Integer)))   'add 4 safety 
       v->ListStrings = sss
    END IF
 
    tempwstr = *MyWstr
    MyLen = len(tempwstr)
   
    if v <> 0 then
      Dim ttt as GRDArray Pointer = cast(GRDArray Ptr,v->ListStrings)
      If ttt <> 0 Then
         if ttt[WhichStr].ws <> 0 then  'need to save the header and deallocate
            Deallocate ttt[WhichStr].ws
         end if   
         ttt[WhichStr].ws = cast(wstring ptr,CAllocate((MyLen + 1) * Len(WString))) ' 1 null
         *ttt[WhichStr].ws = tempwstr
      end if 'ttt
    end if 'v
  
End Sub
  

Sub GRD_GetListing(hWnd As HWND,WhichStr as long, MyWstr As WString Ptr)
    ' Get the listing string contents and copy it to MyWstr.
    Dim ttt As Integer Ptr
    Dim v As GridData Ptr
    
    if WhichStr > GRD_NUMLISTS or WhichStr < 0 THEN  Exit Sub
   
    v = GetProp(hWnd,"GridData")
   
    if v <> 0 then 
      Dim ttt as GRDArray Pointer = cast(any Ptr,v->ListStrings)
  
       If ttt <> 0 Then
          if ttt[WhichStr].ws <> 0 then *MyWstr = *(ttt[WhichStr].ws)
       end if 'ttt
    end if 'v
    
End Sub

Sub GRD_PutListingS(hWnd As HWND,WhichStr as Long, Mystr As String)
      dim tempstr as Wstring * GRD_MAXCELLLEN
      dim tempstrptr as Wstring ptr
      dim z as string * GRD_MAXCELLLEN
      tempstrptr=varptr(tempstr)
      
      z = Mystr
      tempstr = wstr(z)
    
      GRD_PutListing(hWnd,WhichStr, tempstrptr) 
end sub

Sub GRD_PutListingA(hWnd As HWND,WhichStr as LONG, Mystr As ZString Ptr)
      dim tempstr as Wstring * GRD_MAXCELLLEN
      dim tempstrptr as Wstring ptr
      dim z as string * GRD_MAXCELLLEN
      tempstrptr=varptr(tempstr)
      
      z = *Mystr
      tempstr = wstr(z)
     
      GRD_PutListing(hWnd,WhichStr, tempstrptr) 
end sub



'********************************************************************
' Cell Contents Transfer procedures
'********************************************************************
 'Need to incorporate the 5 widechar format block
Sub GRD_PutCell(hWnd As HWND,row As Long,col As Long, MyWstr As WString Ptr, MySheet As Long = -1)
    ' Put the string contents of a cell to a new string in MyWstr.
    Dim ttt As Integer Ptr
    Dim As WString Ptr str2
    Dim v As GridData Ptr
    Dim TheElement As Long,i As Long,MyLen as long
    dim tempwstr as Wstring * GRD_MAXCELLLEN
    dim MyCS as long
    Dim MyAttr as WSTRING * 9
    dim c as cellformat3 
   
    MyAttr = space(8)  'should never need to do this
    tempwstr = *MyWstr
    MyLen = len(tempwstr)
     
    v = GetProp(hWnd,"GridData")
    if v <> 0 then
       
      if MySheet = -1 then
        MyCS = v->CS  'Current Sheet
       else 
        If MySheet < 0 or MySheet > v->TotalSheets then exit sub
        MyCS = MySheet
      end if 'MySheet
      
      if row > v->SH(MyCS).DimRows  or row < 0 then exit sub
      if col > v->SH(MyCS).DimCols  or col < 0 then exit sub
      
      TheElement=(row * (v->SH(MyCS).DimCols+1) ) + (col)
      Dim ttt as GRDArray Pointer = cast(GRDArray Ptr,v->SH(MyCS).ws)
      If ttt <> 0 Then
         if ttt[TheElement].ws <> 0 then  'need to save the header and deallocate
            CopyMemory(@c,ttt[TheElement].ws,HEADERBYTES)
            Deallocate ttt[TheElement].ws : ttt[TheElement].ws = 0
         end if   
         ttt[TheElement].ws = cast(wstring ptr,CAllocate((MyLen+10) * Len(WString)))  '10 is 1 extra - 8 for header 1 null
         *ttt[TheElement].ws = space(8) & tempwstr
         CopyMemory(ttt[TheElement].ws,@c,HEADERBYTES) 'if there was a header then restore it  if not c should be all zeroes
      end if 'ttt
        
      
    end if 'v
End Sub

Sub GRD_GetCell(hWnd As HWND,row As Long,col As Long, MyWstr As WString Ptr, MySheet As Long = -1)
    ' Get the string contents of a cell and copy it to MyWstr.
    Dim ttt As Integer Ptr
    Dim v As GridData Ptr
    Dim TheElement As Long,i As Long
    dim MyCS as long 
   
    v = GetProp(hWnd,"GridData")
   
    if v <> 0 then
      if MySheet = -1 then
        MyCS = v->CS  'Current Sheet
       else 
        If MySheet < 0 or MySheet > v->TotalSheets then exit sub
        MyCS = MySheet
      end if 'MySheet
      
      if row > v->SH(MyCS).DimRows  or row < 0 then exit sub
      if col > v->SH(MyCS).DimCols  or col < 0 then exit sub
      TheElement=(row * (v->SH(MyCS).DimCols+1) ) + (col)
      
      Dim ttt as GRDArray Pointer = cast(any Ptr,v->SH(MyCS).ws)
  
       If ttt <> 0 Then
          if ttt[TheElement].ws <> 0 then *MyWstr = *(ttt[TheElement].ws+8)
       end if 'ttt
    end if 'v
    
End Sub

Sub GRD_GetCellA(hWnd As HWND,row As Long,col As Long, Mystr As ZString Ptr, MySheet As Long = -1)
      dim tempstr as Wstring * GRD_MAXCELLLEN
      dim tempstrptr as Wstring ptr
      dim z as string * GRD_MAXCELLLEN
      tempstrptr=varptr(tempstr)
      
      GRD_GetCell(hWnd,row,col, tempstrptr, MySheet)
      z = str(tempstr)
     
      *Mystr = z
end sub

Sub GRD_PutCellA(hWnd As HWND,row As Long,col As Long, Mystr As ZString Ptr, MySheet As Long = -1)
      dim tempstr as Wstring * GRD_MAXCELLLEN
      dim tempstrptr as Wstring ptr
      dim z as string * GRD_MAXCELLLEN
      tempstrptr=varptr(tempstr)
      
      z = *Mystr
      tempstr = wstr(z)
     
      GRD_PutCell(hWnd,row,col, tempstrptr, MySheet) 
end sub

function GRD_GetCellS(hWnd As HWND,row As Long,col As Long, MySheet As Long = -1) as string
      dim tempstr as Wstring * GRD_MAXCELLLEN
      dim tempstrptr as Wstring ptr
      dim z as string * GRD_MAXCELLLEN
      tempstrptr=varptr(tempstr)
      dim s as string
      
      GRD_GetCell(hWnd,row,col, tempstrptr, MySheet)
      z = str(tempstr)
      s = z
      function = s 
     
end function

Sub GRD_PutCellS(hWnd As HWND,row As Long,col As Long, Mystr As String, MySheet As Long = -1)
      dim tempstr as Wstring * GRD_MAXCELLLEN
      dim tempstrptr as Wstring ptr
      dim z as string * GRD_MAXCELLLEN
      tempstrptr=varptr(tempstr)
      
      z = Mystr
      tempstr = wstr(z)
     
      GRD_PutCell(hWnd,row,col, tempstrptr, MySheet) 
end sub




'********************************************************************
' Main control procedure
'********************************************************************
FUNCTION GRD_GridProc (hWnd AS Hwnd, wMsg AS UINT, _
                    wParam AS WPARAM, lParam AS LPARAM) AS LRESULT
                    
       DIM tm AS TEXTMETRIC, rc AS RECT, wRect AS RECT
       DIM ps AS PAINTSTRUCT,_
              si   AS SCROLLINFO, _
              hdc  AS hDC, hPen AS hGDIOBJ, hBrush AS hGDIOBJ, _
              tSel AS LONG, bkBrush AS hGDIOBJ, hBrushSel AS hGDIOBJ, _
              y AS LONG, x as long, i AS LONG,skey as long, iVscrollInc AS LONG, hScrlInc AS LONG
       dim wstemp as WSTRING * 100
       dim zstr as ZSTRING * 70
       dim templp as point
       Dim MyMouseMoveResult as Long
       Dim v AS GridData PTR
       Dim temp as long, temp1 as long, temp2 as long, temp3 as long, temp4 as long
       dim tempgdiobj as hGDIOBJ
       static lp AS POINT
       DIM TempHWND as HWND
       Dim MyEditText as wstring * GRD_MAXCELLLEN
       Dim MyWStr1 as wstring * GRD_MAXCELLLEN,MyWStr2 as wstring * GRD_MAXCELLLEN
       Dim tDrawItem AS DRAWITEMSTRUCT PTR
       DIM MyCell as CELLFORMAT3
       Dim ttt as GRDArray Pointer
       dim temp5 AS hGDIOBJ, temp6 AS hGDIOBJ
       Dim MyWStrPtr as WSTRING PTR
       dim yPtr as long ptr
       dim xPtr as long ptr
       dim MyRect as RECT
       dim MyRectPtr as RECT PTR
       
       MyRectPtr = varptr(MyRect) 
  
  IF wMsg <> WM_CREATE THEN v = GetProp(hWnd,"GridData")
  skey=GetKeyState(VK_SHIFT) AND &H8000
   
  SELECT CASE wMsg
     CASE WM_CREATE  'Allocate storage for the GridData structure.
        v = HeapAlloc(GetProcessHeap(),  HEAP_ZERO_MEMORY, SIZEOF(GridData))
       
        IF v THEN
           SetProp hWnd, "GridData",cast(HANDLE,v)
        ELSE
           MessageBox hWnd, "Could not allocate memory for the grid" & CHR$(0), _
                              "GRID MESSAGE" & CHR$(0),  MB_OK
           FUNCTION = -1 : EXIT FUNCTION  'Return -1 to break the action
        END IF
        
        v->hParent = GetParent(hWnd)
        v->id      = GetWindowLongPtr(hWnd,  GWLP_ID)
        v->hInst   = cast(hInstance,GetWindowLongPtr(hWnd, GWLP_HINSTANCE))
        v->TotalSheets = MAXGS
        GS_.SelRow = -1 : GS_.RowCount = -1
                
        GRD_Create hWnd,v  
        i= GRD_InitializeCellsAllSheet(hWnd) 
         
        SendMessage hWnd,  GRD_SIZEHANDLER, 0, 0
       
     CASE WM_DESTROY
       
        IF v THEN
           'Destroy the Fonts
           for i = 0 to NUMGRDFONTS
             if v->CellFonts(i) <> 0 then DeleteObject v->CellFonts(i)
           next i
           'Destroy Pens
           IF v->CellLineColorPen <> 0 THEN DeleteObject v->CellLineColorPen
           IF v->BlackLineColorPen<> 0 THEN DeleteObject v->BlackLineColorPen
           'Destroy Brushes
           IF v->CellBkColorBrush <> 0 THEN DeleteObject v->CellBkColorBrush
           IF v->WindowBKColorBrush <> 0 THEN DeleteObject v->WindowBKColorBrush
           IF v->RCHiLiteBrush <> 0 THEN DeleteObject v->RCHiLiteBrush
           IF v->HeaderBrush <> 0 THEN DeleteObject v->HeaderBrush
           IF v->HeaderSelectBrush <> 0 THEN DeleteObject v->HeaderSelectBrush
           IF v->BtnFaceBrush <> 0 THEN DeleteObject v->BtnFaceBrush
           IF v->SolidLine <> 0 THEN DeleteObject v->SolidLine
           IF v->SolidLine2 <> 0 THEN DeleteObject v->SolidLine2
           IF v->SolidLine3 <> 0 THEN DeleteObject v->SolidLine3
           
           
           'Destroy Menus
           for i = 0 to GRD_NUMMENUS
               IF v->RClickMenu(i) <> 0 THEN DestroyMenu v->RClickMenu(i)
           NEXT
          
           'Destroy ListStrings
           if v->ListStrings <> 0 then
              ttt = cast(GRDArray Ptr,v->ListStrings)
              for i = 0 to GRD_NUMLISTS
                  if ttt[i].ws <> 0 then  
                     Deallocate ttt[i].ws : ttt[i].ws = 0
                  end if
              NEXT
              Deallocate v->ListStrings
           end if
           
           'Restore the Subclasses
           If v->OriginalEditProc Then SetWindowLongPtr(v->hEditControl, GWLP_WndProc, Cast(LONG_PTR, v->OriginalEditProc)) 
           If v->OriginalStaticProc Then SetWindowLongPtr(v->hSheetControl, GWLP_WndProc, Cast(LONG_PTR, v->OriginalStaticProc))
           If v->OriginalComboProc Then SetWindowLongPtr(v->hListControl, GWLP_WndProc, Cast(LONG_PTR, v->OriginalComboProc))
           
           
           'Destroy off screen bitmap   
           IF v->hbit  THEN DeleteObject SelectObject(v->memDC, v->hbit)
           IF v->memDC THEN DeleteDC cast(hDC,v->memDC)
           'and the bitmap canvas
           if v->PrintDC THEN DeleteDC cast(hDC,v->PrintDC)
           
           'Destroy bitmaps
           FOR i = 0 TO GRD_NUMBITMAPS
             IF v->Bitmaps(i) <> 0 THEN DeleteObject v->Bitmaps(i)
           NEXT i 
          
           'Destroy Cell Arrays 
           GRD_EraseCellsAllSheets(hWnd)
          
           'Destroy User Storage
           HeapFree(GetProcessHeap(), 0, BYVAL v)
           RemoveProp hWnd,"GridData"
        END IF

     CASE WM_SETFONT
        
     CASE WM_DRAWITEM
       
          DIM sheetTabDC as hDC, tabdc AS hDC, temp as hDC
          Dim sheettabmp as hGDIOBJ
          Dim oldselect as hGDIOBJ
          Dim temprgn as hRGN
          dim rc as RECT
          dim MyOffset as LONG
          dim oldfont as hGDIOBJ
          Dim counter as Long
          dim MyWStr as wstring * 50
          Dim MyWStrPtr as wstring ptr
'
          tDrawItem = cast (DRAWITEMSTRUCT PTR,lParam)
          MyWStrPtr = varptr(MyWStr)
          
          IF tDrawItem->CtlID = GRD_STATB THEN
               
                i= (tDrawItem->rcItem.Right - tDrawItem->rcItem.Left)\3
               
                FillRect(tDrawItem->hdc, @tDrawItem->rcItem, v->BtnFaceBrush)
                SetTextColor tDrawItem->hdc, v->CellColors(CELLCOLORBLACK)
                SetBkColor tDrawItem->hdc,v->BtnFaceBrushRGB
                SelectObject tDrawItem->hdc,v->CellFonts(GRDSYSFONTSTATUSBAR)
                 
                rc.Left = tDrawItem->rcItem.Left  : rc.Top = tDrawItem->rcItem.Top 
                rc.Bottom = tDrawItem->rcItem.Top + + SBSIZE  : rc.Right = tDrawItem->rcItem.Right
                rc.Right = i
                MyWStr = v->StatusBarText1 
                DrawTextW tDrawItem->hdc,BYVAL MywStrPtr, len(MyWStr),@rc, DT_NOPREFIX OR DT_LEFT or DT_SINGLELINE
                rc.Left = rc.Left + i : rc.Right = rc.Right + i 
                MyWStr = v->StatusBarText2 
                DrawTextW tDrawItem->hdc,BYVAL MywStrPtr, len(MyWStr),@rc, DT_NOPREFIX OR DT_CENTER or DT_SINGLELINE
                rc.Left = rc.Left + i : rc.Right = rc.Right + i
                MyWStr = v->StatusBarText3 
                DrawTextW tDrawItem->hdc,BYVAL MywStrPtr, len(MyWStr),@rc, DT_NOPREFIX OR DT_RIGHT or DT_SINGLELINE
                
          end if
          
         IF tDrawItem->CtlID = GRD_SCNTL THEN
                
                sheetTabDC = CreateCompatibleDC(tDrawItem->hdc)
                temp = GetDC(hWnd)
                sheettabmp  = CreateCompatibleBitmap(temp, tDrawItem->RcItem.Right,tDrawItem->RcItem.Top + SBSIZE)
                ReleaseDC hWnd,temp
                oldselect = SelectObject(sheetTabDC, sheettabmp)

'
                IF v->HeaderBrush = 0 THEN
                    FillRect(sheetTabDC, @tDrawItem->rcItem, GetStockObject(LTGRAY_BRUSH))
                  ELSE
                    FillRect(sheetTabDC, @tDrawItem->rcItem, v->HeaderBrush)
                END IF
'
                MoveToEx (sheetTabDC,tDrawItem->RcItem.Left,tDrawItem->RcItem.Top,@lp)
                LineTo (sheetTabDC,tDrawItem->RcItem.Right,tDrawItem->RcItem.Top)

                MoveToEx (sheetTabDC,tDrawItem->RcItem.Left+16,tDrawItem->RcItem.Top,@lp): LineTo (sheetTabDC,tDrawItem->RcItem.Left+16,GS_.StatusBarWidth)
                MoveToEx (sheetTabDC,tDrawItem->RcItem.Left+33,tDrawItem->RcItem.Top,@lp): LineTo (sheetTabDC,tDrawItem->RcItem.Left+33,GS_.StatusBarWidth)
                MoveToEx (sheetTabDC,tDrawItem->RcItem.Left+49,tDrawItem->RcItem.Top,@lp): LineTo (sheetTabDC,tDrawItem->RcItem.Left+49,GS_.StatusBarWidth)
                MoveToEx (sheetTabDC,tDrawItem->RcItem.Left+65,tDrawItem->RcItem.Top,@lp): LineTo (sheetTabDC,tDrawItem->RcItem.Left+65,GS_.StatusBarWidth)
'

                IF  v->TabSheetTransport <> 1 THEN
                    rc.Top=4 : rc.Left=4 : rc.Bottom=13 : rc.Right=6
                    FillRect(sheetTabDC, @rc, GetStockObject(BLACK_BRUSH))
                    temprgn=GRD_ButtonPoly(1)
                    FillRgn(sheetTabDC,temprgn,GetStockObject(BLACK_BRUSH))
                    DeleteObject temprgn
                END IF
                
                IF  v->TabSheetTransport <> 2 THEN
                    temprgn=GRD_ButtonPoly(2)
                    FillRgn(sheetTabDC,temprgn,GetStockObject(BLACK_BRUSH))
                    DeleteObject temprgn
                END IF
 
                IF  v->TabSheetTransport <> 3 THEN
                    temprgn=GRD_ButtonPoly(3)
                    FillRgn(sheetTabDC,temprgn,GetStockObject(BLACK_BRUSH))
                    DeleteObject temprgn
                END IF
 
                IF  v->TabSheetTransport <> 4 THEN
                    rc.Top=4 : rc.Left=59 : rc.Bottom=13 : rc.Right=61
                    FillRect(sheetTabDC, @rc, GetStockObject(BLACK_BRUSH))
                    temprgn=GRD_ButtonPoly(4)
                    FillRgn(sheetTabDC,temprgn,GetStockObject(BLACK_BRUSH))
                    DeleteObject temprgn
                END IF
 
                'find the left offset
                temp3 = v->SH(cast(long,v->TabSheetLeft)).SheetOffset1 - 73
                for counter = v->TabSheetLeft to v->ActiveSheetUbound
                  SetBkMode(sheetTabDC, TRANSPARENT )
                  SetTextColor sheetTabDC, v->SH(counter).SheetColor
                  rc.Top = tDrawItem->RcItem.Top : rc.Bottom = tDrawItem->RcItem.Top + SBSIZE : rc.Left = v->SH(counter).SheetOffset1-temp3 : rc.Right = v->SH(counter).SheetOffset2-temp3
                
                   if counter = v->CS THEN
                      SetBkMode(sheetTabDC,OPAQUE)
                      FillRect sheetTabDC,@rc,GetStockObject(WHITE_BRUSH)
                      MoveToEx (sheetTabDC,rc.Left-1,rc.Top,@lp): LineTo (sheetTabDC,rc.Left-1,GS_.StatusBarWidth)
                      MoveToEx (sheetTabDC,rc.Right,rc.Top,@lp): LineTo (sheetTabDC,rc.Right,GS_.StatusBarWidth)
                      SelectObject sheetTabDC,v->CellFonts(GRDSYSFONTTABBOLD)    
                    else  
                      SelectObject sheetTabDC,v->CellFonts(GRDSYSFONTTABNORM)    
                  END IF
                
                  
                  DrawTextW(sheetTabDC, v->SH(counter).SheetName, len(v->SH(counter).SheetName), @rc, DT_SINGLELINE OR DT_CENTER OR DT_VCENTER OR DT_NOPREFIX) 
                  
                  if rc.Left > tDrawItem->RcItem.Right then exit for
                NEXT
                
                MoveToEx (sheetTabDC,tDrawItem->RcItem.Left,tDrawItem->RcItem.Top + SBSIZE -1 ,@lp): LineTo (sheetTabDC,tDrawItem->RcItem.Right,tDrawItem->RcItem.Top + SBSIZE-1)
                '
                'Sizing handle 
                rc.Top = tDrawItem->RcItem.Top : rc.Bottom = tDrawItem->RcItem.Top + SBSIZE : rc.Left = tDrawItem->RcItem.Right - SHEETTABBUTTONWIDTH : rc.Right = tDrawItem->RcItem.Right
                FillRect sheetTabDC,@rc,GetStockObject(GRAY_BRUSH) 
                FrameRect sheetTabDC,@rc,GetStockObject(BLACK_BRUSH)
              
               BitBlt tDrawItem->hdc,tDrawItem->RcItem.top,tDrawItem->RcItem.left,tDrawItem->RcItem.right,tDrawItem->RcItem.bottom,sheetTabDC,0,0,SRCCOPY
'
               IF sheettabmp  THEN DeleteObject SelectObject(sheetTabDC, sheettabmp)
               IF sheetTabDC THEN DeleteDC sheetTabDC
'
          FUNCTION = FALSE
          EXIT FUNCTION
      END IF
         
   CASE WM_MOUSEWHEEL
      
           'GRD_ActivateEditControl(HWnd,0)
           'GRD_ActivateListControl(Hwnd,0)
                          
           temp3=HIWoRD(wParam)
           IF temp3 = WHEEL_DELTA THEN 'Scroll Down
                 temp4= max(GS_.TopRow - 1,1) 
             else
                 temp4=min(GS_.TopRow + 1,GS_.TotRows)
           end if
           
           SendMessage(hWnd,GRD_SETTOPROW,temp4,1) 'Do a refresh with 1
 
 
     case WM_COMMAND
        'wNotifyCode = HIWORD(wParam); // notification code 
        'wID = LOWORD(wParam); // item, control, or accelerator identifier 
        'hwndCtl = (HWND) lParam; // handle of control 
        if v->ProcessRClickMenuFlag <> 0 THEN   
         v->ProcessRClickMenuFlag = 0
           if v->RightClickWhich = GRDRM_TAB THEN
              GRD_SendNotifyMessage v->hParent,hWnd,v->id,GRDN_RCLICKMENU,cast(long,wParam),0,v->TempSheetRClickMenu,v->RightClickWhich
             else
              GRD_SendNotifyMessage v->hParent,hWnd,v->id,GRDN_RCLICKMENU,cast(long,wParam),v->RightClickRow,v->RightClickCol,v->RightClickWhich 
           END IF   
        END IF
        
'Probably won't use this.        
        IF LOWORD(wParam) = GRD_LCNTL AND HIWORD(wParam) = LBN_DBLCLK THEN
'       
       END IF

        
        
 
     CASE WM_CONTEXTMENU
           'GRD_ActivateEditControl(HWnd,0)
           'GRD_ActivateListControl(Hwnd,0)
        
           v->RightClickRow = GS_.MyRowTemp
           v->RightClickCol = GS_.MyColTemp
           v->ProcessRClickMenuFlag = 1
           if wParam = cast(wParam,v->hSheetControl) THEN
               v->RightClickWhich = GRDRM_TAB
               if v->RClickMenu(GRDRM_TAB) <> 0 THEN
                 TrackPopupMenu(v->RClickMenu(GRDRM_TAB), TPM_LEFTALIGN OR TPM_LEFTBUTTON,LOWORD(lParam), HIWORD(lParam), 0, hWnd, BYVAL Null)
               END IF
             else 'main grid
               v->RightClickWhich = GS_.TempMMResult
               if v->RClickMenu(v->RightClickWhich) <> 0 THEN
                 TrackPopupMenu(v->RClickMenu(v->RightClickWhich), TPM_LEFTALIGN OR TPM_LEFTBUTTON,LOWORD(lParam), HIWORD(lParam), 0, hWnd, BYVAL Null)
               END IF
            END IF
     CASE WM_SIZE
        IF HIWORD(lParam) THEN
           GS_.wMaxHeight = HIWORD(lParam) / GS_.cyChar
           SendMessage hWnd,  GRD_SIZEHANDLER, 0, 0
        END IF
        
        'GRD_ActivateEditControl(HWnd,0)
        'GRD_ActivateListControl(Hwnd,0)
        
         IF v->TabSheetPos > 0 and v->TabSheetPercent > 0 THEN
            GetClientRect(hWnd,@rc)
            temp = INT((v->TabSheetPercent/1000) * (rc.Right))
            IF temp > 0 THEN v->TabSheetPos = temp
        END IF   
        
        'GS_.TabSheetPos =100
        GetClientRect(hWnd,@rc)
        if rc.Right <> 0 THEN
            IF rc.Right - v->TabSheetPos < 70 THEN  v->TabSheetPos = rc.Right - 70
        END IF
        IF v->TabSheetPos > 0 AND v->TabSheetPos < 70 THEN v->TabSheetPos = 70  
        IF v->TabSheetPos < 0 THEN v->TabSheetPos = 70 
       
        
        MoveWindow( v->hSheetControl,0, rc.bottom-GS_.HScrollWidth-GS_.StatusBarWidth,v->TabSheetPos, rc.bottom -GS_.StatusBarWidth, True )
        MoveWindow( v->hLRCorner,rc.right-GS_.VScrollWidth, rc.bottom-GS_.HScrollWidth-GS_.StatusBarWidth,rc.right, rc.Bottom-GS_.StatusBarWidth, True )
        MoveWindow( v->hHScrollBar,v->TabSheetPos , rc.bottom-GS_.HScrollWidth-GS_.StatusBarWidth, rc.right-v->TabSheetPos-GS_.VScrollWidth , GS_.HScrollWidth, True ) 
        MoveWindow( v->hVScrollBar,rc.right-GS_.VScrollWidth,0 , GS_.VScrollWidth, rc.bottom-GS_.HScrollWidth-GS_.StatusBarWidth, True )
        MoveWindow( v->hStatusBar,rc.left , rc.bottom-GS_.StatusBarWidth, rc.right  , rc.Bottom, True ) 
' 
'
      #IF DEBUGMODE = 1
 '          print "Test"
      #ENDIF
                                       

     CASE GRD_SIZEHANDLER  'create a virtual window that fits current window size
        IF v->hbit  THEN DeleteObject SelectObject(v->memDC,v->hbit)
        IF v->memDC THEN DeleteDC v->memDC
        IF v->PrintDC THEN DeleteDC v->PrintDC

        hDC = GetDC(hWnd)
           GetClientRect(hWnd, @wRect)
           v->memDC = CreateCompatibleDC(hDC)
           v->PrintDC = CreateCompatibleDC(hDC)
           v->hbit  = CreateCompatibleBitmap(hDC, wRect.Right, wRect.Bottom)
           v->hbit = SelectObject(v->memDC, v->hbit)

           hbrush   = GetStockObject(LTGRAY_BRUSH)
           PatBlt v->memDC, 0, 0, wRect.Right, wRect.Bottom,  PATCOPY
           
           'Will probably delete next 6 lines
           IF hbrush THEN SelectObject v->memDC, hbrush
      
        ReleaseDC (hWnd, hdc)
        if GS_.cyChar = 0 then GS_.cyChar = 20
        
        GS_.wMaxHeight = wRect.Bottom / GS_.cyChar
        GS_.wMaxWidth = MAX(1, (wRect.Right / 2) / (GS_.charWidth - 1) ) 'Get window size in characters
        'Print "GS_.wMaxHeight=  "  & str(GS_.wMaxHeight)
        GS_.iHscrollMax = MAX (0, 1024 - GS_.wMaxWidth) '1024, max number of characters..
        GS_.iHscrollPos = MIN (GS_.iHscrollPos, GS_.iHscrollMax)
        GS_.xPos = GS_.charWidth * (- GS_.iHscrollPos)
        GRD_SetSBar hwnd, GS_.wMaxWidth, GS_.iHscrollMax, GS_.iHscrollPos,  GRD_HSBAR
     
        GS_.wMaxHeight = wRect.Bottom / GS_.cyChar
        GRD_SetSBar hwnd, GS_.wMaxHeight, MAX(0, GS_.RowCount), GS_.TopRow,  GRD_VSBAR
        'GRD_SetSBar hwnd, 5, 200, GS_.TopRow,  GRD_VSBAR
        InvalidateRect hWnd, BYVAL  NULL, 0 : UpdateWindow hwnd

     CASE WM_VSCROLL
      IF GS_.TotRows < 0 THEN EXIT SELECT
      GRD_ActivateEditControl(HWnd,0)
      GRD_ActivateListControl(Hwnd,0)

        IF lParam = v->hVScrollBar THEN
            SELECT CASE LOWORD(wParam)
               CASE SB_TOP        : iVscrollInc = -GS_.TopRow
               CASE SB_BOTTOM     : iVscrollInc =  GS_.TotRows - GS_.wMaxHeight
               CASE SB_LINEUP     : iVscrollInc = -1
               CASE SB_LINEDOWN   : IF GS_.TopRow < GS_.TotRows - GS_.wMaxHeight + 2 THEN iVscrollInc =  1
               CASE SB_PAGEUP     : iVscrollInc = MIN(-1, -GS_.wMaxHeight)
               CASE SB_PAGEDOWN   : iVscrollInc = MAX(1,   GS_.wMaxHeight)
               CASE SB_THUMBTRACK ' getScrollInfo enables 32-bit scroll positions
                  si.cbSize = SIZEOF(si)
                  si.fMask  = SIF_TRACKPOS
                  GetScrollInfo v->hVScrollBar, 2, @si    'SB_CTL = 2 in WinUser.h
                  iVscrollInc = si.nTrackPos - GS_.TopRow
               CASE ELSE : iVscrollInc = 0
            END SELECT

            iVscrollInc = MAX(-GS_.TopRow, MIN(iVscrollInc, GS_.TotRows - GS_.wMaxHeight + 1))
        END IF 
        IF iVscrollInc <> 0 THEN  
           IF lParam = v->hVScrollBar THEN
             '
              GS_.TopRow = MIN(GS_.TopRow + iVscrollInc, GS_.TotRows - GS_.wMaxHeight + 2)
              'GS_.TopRow = GS_.TopRow + 1
              IF GS_.TopRow > GS_.TotRows - 2 THEN  GS_.TopRow = GS_.TotRows - 2
              
              IF GS_.TopRow < 1  THEN  GS_.TopRow = 1
              GRD_SetSBar hwnd, GS_.wMaxHeight, MAX(1, GS_.TotRows), GS_.TopRow, GRD_VSBAR
              InvalidateRect hWnd, BYVAL NULL, 0 : UpdateWindow hwnd
            END IF

        END IF


     CASE WM_HSCROLL
           GRD_ActivateEditControl(HWnd,0)
           GRD_ActivateListControl(Hwnd,0)
          IF lParam = v->hHScrollBar THEN
            SELECT CASE LOWORD(wParam)
               CASE SB_LINELEFT      : hScrlInc = -1
               CASE SB_LINERIGHT     : hScrlInc =  1
               CASE SB_PAGELEFT      : hScrlInc = -3  '@v.wMaxWidth
               CASE SB_PAGERIGHT     : hScrlInc = 3' @v.wMaxWidth
               CASE SB_THUMBTRACK    : hScrlInc = HIWORD(wParam) - GS_.LeftCol
               CASE ELSE             : hScrlInc = 0
            END SELECT

          hScrlInc = MAX(-GS_.LeftCol, MIN(hScrlInc, GS_.TotCols - GS_.LeftCol))
            
          IF hScrlInc <> 0 THEN
           'Set page to 3 for now
           IF lParam = v->hHScrollBar THEN
               GS_.LeftCol = GS_.LeftCol + hScrlInc
               GS_.LeftCol = MAX(GS_.LeftCol,1): GS_.LeftCol = MIN(GS_.LeftCol,GS_.TotCols)
               IF GS_.LeftCol < 1  THEN  GS_.LeftCol = 1 'for the freeze cols
               GRD_SetSBar hwnd, 3, GS_.TotCols, GS_.LeftCol, GRD_HSBAR
               InvalidateRect hwnd, BYVAL NULL, 0 : UpdateWindow hwnd
           END IF

         
        END IF    

        END IF       

     CASE WM_GETDLGCODE ' Ensure that the control processes all keys by itself
        DIM iMsg AS UINT, pMsg AS TAGMSG PTR
        pMsg = cast(TAGMSG PTR,lParam)
        IF pMsg > 0 THEN iMsg = pMsg->Message
        IF iMsg = WM_KEYDOWN OR iMsg =  WM_CHAR THEN
           SELECT CASE pMsg->wParam
               'Nothing for now
              CASE VK_TAB, VK_ESCAPE   'let system handle these
                 FUNCTION = DefWindowProc(hWnd, wMsg, wParam, lParam)
             '   FUNCTION = DLGC_WANTTAB  ''Function = DLGC_WANTTAB    DLGC_WANTALLKEYS
            
              CASE ELSE
                 FUNCTION = DLGC_WANTALLKEYS
           END SELECT
        END IF

     CASE WM_CHAR  ' 8 is backspace
           GRD_ActivateListControl HWND,0
           IF (wParam > 31 AND wParam < 128) THEN  
              MyEditText=CHR$(wParam)
              SetWindowTextW v->hEditControl,varptr(MyEditText)
              SendMessage v->hEditControl,EM_SETSEL,2,2                                
              GRD_ActivateEditControl(HWnd,2)
           end if
           
           
     CASE  WM_KEYDOWN
              temp1 = GS_.SelRow
              temp2 = GS_.SelCol
              GRD_ActivateListControl HWND,0
        SELECT CASE (wParam)
           case VK_F2
              GRD_ActivateEditControl(HWnd,1)
             
           
           case VK_LEFT
                 GS_.SelCol = Max(GS_.SelCol - 1, 1)
                 if GS_.SelCol < GS_.LeftCol THEN
                     GS_.LeftCol = Max(GS_.LeftCol - 1,1)
                  END IF
                 InvalidateRect hWnd, BYVAL  NULL, 0 : UpdateWindow hwnd
                 
            case VK_RIGHT
                  GS_.SelCol = Min(GS_.SelCol + 1, GS_.TotCols)
                  if GS_.SelCol > GS_.RightCol - 1 THEN
                     GS_.LeftCol = Min(GS_.LeftCol + 1,GS_.TotCols - 3)
                  END IF
                 InvalidateRect hWnd, BYVAL  NULL, 0 : UpdateWindow hwnd     
                 
            CASE  VK_UP
                 GS_.SelRow = Max(GS_.SelRow - 1, 1)
                 if GS_.SelRow < GS_.TopRow THEN
                    GS_.TopRow = Max(GS_.TopRow - 1,1)
                 END IF
                 InvalidateRect hWnd, BYVAL  NULL, 0 : UpdateWindow hwnd

           CASE  VK_DOWN 
                 GS_.SelRow = Min(GS_.SelRow + 1, GS_.TotRows)
                 if GS_.SelRow > GS_.BottomRow - 2 THEN
                    GS_.TopRow = Min(GS_.TopRow + 1,GS_.TotRows - 3)
                 END IF
                 InvalidateRect hWnd, BYVAL  NULL, 0 : UpdateWindow hwnd

           CASE  VK_PRIOR 'PgUp
              temp = GS_.BottomRow - GS_.TopRow -2
              GS_.SelRow = Max(GS_.SelRow - temp, 1)
              if GS_.SelRow < GS_.TopRow THEN 'Scrolldown a page
                 GS_.TopRow = Max(GS_.SelRow, 1)
              END IF
              InvalidateRect hWnd, BYVAL  NULL, 0 : UpdateWindow hwnd

           CASE  VK_NEXT  'PgDn
              temp = GS_.BottomRow - GS_.TopRow -2
              GS_.SelRow = Min(GS_.SelRow + temp, GS_.TotRows)
              if GS_.SelRow > GS_.BottomRow - 2 THEN 'Scrolldown a page
                 GS_.TopRow = Min(GS_.SelRow - temp , GS_.TotRows -3)
              END IF
              
              InvalidateRect hWnd, BYVAL  NULL, 0 : UpdateWindow hwnd

           CASE  VK_SPACE  'Space bar pressed
              'GRD_SendNotifyMessage v->hParent,hWnd,v->id,GRDN_SPACE,0,0,0,0
              
               'v->ReturnKeyMode           
               'v->TabKeyMode                

           CASE  VK_RETURN 'Enter key pressed
              'GRD_SendNotifyMessage v->hParent,hWnd,v->id,GRDN_RETURN,0,0,0,0
              ' Needs work
              'GRD_ActivateEditControl(HWnd,0)
              if skey = 0 THEN
                  if v->ReturnKeyMode = 0 then
                    GS_.SelRow = Min(GS_.SelRow + 1, GS_.TotRows)
                    if GS_.SelRow > GS_.BottomRow - 2 THEN 'Scroll
                       GS_.TopRow = Min(GS_.TopRow + 1,GS_.TotRows - 3)
                    END IF
                   else
                    GS_.SelRow = 1
                    GS_.SelCol = Min(GS_.SelCol + 1, GS_.TotCols)  
                  END IF  
                 else
                    GS_.SelRow = Max(GS_.SelRow - 1, 1)
                    if GS_.SelRow < GS_.TopRow THEN 'Scroll
                       GS_.TopRow = Max(GS_.TopRow - 1,1)
                    END IF
              END IF
              InvalidateRect hWnd, BYVAL  NULL, 0 : UpdateWindow hwnd
              
              CASE  VK_TAB 'Tab key pressed
                 
              'GRD_SendNotifyMessage v->hParent,hWnd,v->id,GRDN_TAB,0,0,0,0
              'GRD_ActivateEditControl(HWnd,0)
              if skey = 0 THEN
                    GS_.SelCol = Min(GS_.SelCol + 1, GS_.TotCols)
                    if GS_.SelCol > GS_.RightCol - 1 THEN 'Scroll
                     GS_.LeftCol = Min(GS_.LeftCol + 1,GS_.TotCols - 3)
                    END IF
                 else
                    GS_.SelCol = Max(GS_.SelCol - 1, 1)
                    if GS_.SelCol < GS_.LeftCol THEN 'Scroll
                     GS_.LeftCol = Max(GS_.LeftCol - 1,1)
                    END IF   
              END IF
              InvalidateRect hWnd, BYVAL  NULL, 0 : UpdateWindow hwnd

           CASE  VK_DELETE 'Delete key pressed
              'GRD_SendNotifyMessage v->hParent,hWnd,v->id,GRDN_DELETE,0,0,0,0

           CASE  VK_HOME
              GS_.SelRow = 1 : GS_.TopRow = 1
              GS_.SelCol = 1 : GS_.LeftCol = 1
              SetScrollPos (hWnd,  GRD_VSBAR, GS_.TopRow,  TRUE)
              SetScrollPos (hWnd,  GRD_HSBAR, GS_.LeftCol,  TRUE)
              
              InvalidateRect hWnd, BYVAL  NULL, 0 : UpdateWindow hwnd

           CASE  VK_END
              'Print "End Key"
              GS_.SelRow = GS_.TotRows : GS_.SelCol = GS_.TotCols
              GS_.TopRow = Max(GS_.TotRows - 10,1)
              GS_.LeftCol = Max(GS_.TotCols - 3,1)
              SetScrollPos (hWnd,  GRD_VSBAR, GS_.TopRow,  TRUE)
              SetScrollPos (hWnd,  GRD_HSBAR, GS_.LeftCol,  TRUE)
              InvalidateRect hWnd, BYVAL  NULL, 0 : UpdateWindow hwnd
        END SELECT
        
        if temp1 <> GS_.SelRow or temp2 <> GS_.SelCol THEN
           GRD_SendNotifyMessage v->hParent,hWnd,v->id,GRDN_CELLCHANGED,GS_.SelRow,GS_.SelCol,temp1,temp2
           GS_.OldSelRow = GS_.SelRow
           GS_.OldSelCol = GS_.SelCol
        END IF
        

     CASE  WM_MOUSEMOVE
        skey=GetKeyState(VK_SHIFT) AND &H8000
        lp.x = Loword(lParam)
        lp.y = HIword(lParam)
        GetClientRect(hWnd,@rc)
        
        MyMouseMoveResult = GRD_MouseMoveUpdate(hwND,wParam,lParam)
        GS_.TempMMResult =  MyMouseMoveResult
         
        ' RowColResize
         'SetClassLongPtr hWnd, GCLP_HCURSOR, LoadCursor (GetModuleHandle("mmm.DLL"), "SSCUR")
        ' Select CASE MyMouseMoveResult
        '    case 0 'grid
        '       SetClassLongPtr hWnd, GCLP_HCURSOR, cast(LONG_PTR,LoadCursor (NULL, BYVAL IDC_ARROW))
        '    case 1 'row header
        '       SetClassLongPtr hWnd, GCLP_HCURSOR, cast(LONG_PTR,LoadCursor (NULL, BYVAL IDC_SIZENS))
        '    case 2 'col header
        '       SetClassLongPtr hWnd, GCLP_HCURSOR, cast(LONG_PTR,LoadCursor (NULL, BYVAL IDC_SIZEWE))      
        '    CASE ELSE
               SetClassLongPtr hWnd, GCLP_HCURSOR, cast(LONG_PTR,LoadCursor (NULL, BYVAL IDC_ARROW))
               v->CurrentCursor = 0
        ' END SELECT
         
         'Check for Column Width Change
          v->ColResizeFlag=0
          temp3 = GRD_FormatAttribute(hwnd,0,0,GRDFORMAT_GETCOLWIDTH)
          temp4 = GRD_FormatAttribute(hwnd,0,0,GRDFORMAT_GETROWHEIGHT) 
          
          IF (lp.y < (temp4 + GS_.GridTop)) AND (lp.y > GS_.GridTop) AND temp3 > 0 THEN
            temp=temp3 + GS_.GridLeft
            FOR temp1= GS_.LeftCol TO GS_.TotCols
               IF temp > rc.right THEN EXIT FOR
               temp=temp + GRD_FormatAttribute(hwnd,0,temp1,GRDFORMAT_GETCOLWIDTH) - 1
               IF (lp.x > temp - GRD_COLRESIZE) AND (lp.x < temp + GRD_COLRESIZE) AND v->SuppressResize=0 AND GS_.ColSelectionOn=0 THEN
                  SetClassLongPtr hWnd, GCLP_HCURSOR, cast(LONG_PTR,LoadCursor (NULL, BYVAL IDC_SIZEWE))
                  v->CurrentCursor = 2
                  IF v->ColResizeFlag = 0 THEN v->ColResizeFlag = 1
                  v->RowResizeFlag = 0
                  v->ColResizeTemp = temp1
                  EXIT FOR
                ELSE
                  IF v->ColResizeFlag = 0 THEN
                     v->ColResizeTemp =0
                     IF lp.x < rc.Right AND lp.x > temp3 + GS_.GridLeft THEN
                       '  IF GS_.ColSelection = 0 THEN
                            'temporary cross
                            ' SetClassLongPtr hWnd, GCLP_HCURSOR, cast(LONG_PTR,LoadCursor (NULL, BYVAL IDC_CROSS))
                             v->CurrentCursor = 4
                        ' END IF
                       ELSE
                          SetClassLongPtr hWnd, GCLP_HCURSOR, cast(LONG_PTR,LoadCursor (NULL, BYVAL IDC_ARROW))
                          v->CurrentCursor = 0 'off grid
                     END IF
                  END IF
               END IF
            NEXT temp1
          END IF
          
          IF (wParam AND MK_LBUTTON) AND  v->colresizeflag = 1 THEN
               IF v->ColResizeChanging =  0 THEN
                   v->ColResizeStart= temp - GRD_FormatAttribute(hwnd,0,temp1,GRDFORMAT_GETCOLWIDTH) -1
                   v->ColResizeChanging = 1
                   v->ColResizeCol=v->ColResizeTemp
               END IF
               SetCapture(hWnd)
          END IF 
          
           IF (wParam AND MK_LBUTTON) AND v->ColResizeChanging > 0 THEN
                'GetClientRect hWnd, wRect
                hDC = GetDC(hWnd)
                ExcludeClipRect(hDC,0,rc.Bottom - GS_.HScrollWidth-GS_.StatusBarWidth,rc.Right,rc.Bottom)
                ExcludeClipRect(hDC,rc.Right - GS_.VScrollWidth,0,rc.Right,rc.Bottom)
                SetROP2 hDC, R2_notxorpen
                tempgdiobj = SelectObject(hDC,v->SolidLine)

                IF v->ColResizeChanging = 2 THEN 'redraw old line to erase it
                      MovetoEx (hDC, rc.Left + v->Mousemovex , rc.Top + GS_.GridTop,@templp)
                      Lineto hDC, rc.Left + v->Mousemovex  ,rc.Bottom - GS_.HScrollWidth
                  ELSE
                       v->ColResizeChanging = 2 'This is the first time thru - no need to erase old line
                END IF
               
                MovetoEx (hDC, rc.Left + lp.x , rc.Top + GS_.GridTop,@templp)
                Lineto hDC, rc.Left + lp.x  ,rc.Bottom - GS_.HScrollWidth

                SelectObject hDC, tempgdiobj
                SetROP2 hDc, R2_copypen
                ReleaseDC hWnd, hDC

          END IF 
          
         'Row Resize 
         v->rowresizeflag=0
         IF (lp.x < temp3 + GS_.GridLeft) AND (lp.x > GS_.GridLeft)AND temp4 > 0  AND v->SuppressResize=0  THEN 
            temp=temp4 + GS_.GridTop    
            FOR temp1=GS_.TopRow TO GS_.TotRows
               IF temp > rc.bottom THEN EXIT FOR
               temp=temp+GRD_FormatAttribute(hwnd,temp1,0,GRDFORMAT_GETROWHEIGHT)
               IF temp > rc.Bottom - GS_.HScrollWidth THEN EXIT FOR
               IF lp.y > temp-GRD_COLRESIZE AND lp.y < temp + GRD_COLRESIZE AND v->SuppressResize=0 AND GS_.RowSelectionOn=0 THEN
            
                     SetClassLongPtr hWnd, GCLP_HCURSOR, cast(LONG_PTR,LoadCursor (NULL, BYVAL IDC_SIZENS))
                     
                     v->CurrentCursor = 3
                     IF v->rowresizeflag=0 THEN v->rowresizeflag=1
                     v->rowresizetemp=temp1
                     v->colresizeflag = 0
                     EXIT FOR
                 ELSE
                     IF v->rowresizerow=0 AND v->rowresizeflag=0 THEN
                         v->rowresizetemp=0
                         IF lp.y < rc.Bottom AND lp.y > temp4 + GS_.GridTop THEN
                            IF GS_.RowSelection = 1 THEN
                                'SetClassLong hWnd, %GCL_HCURSOR, @v.LTCUR
                                'SetClassLong hWnd, %GCL_HCURSOR, LoadCursor (GetModuleHandle("MLG.DLL"), "LTCUR")
                                v->CurrentCursor = 5
                            END IF
                          ELSE
                          '  SetClassLongPtr hWnd, GCLP_HCURSOR, cast(LONG_PTR,LoadCursor (NULL, BYVAL IDC_ARROW))
                          '  v->CurrentCursor = 0
                         END IF
                     END IF
               END IF
            NEXT temp1
         END IF
 
       
 
         IF (wParam AND MK_LBUTTON) AND v->rowresizeflag = 1 THEN
               IF v->rowresizechanging =  0 THEN
                   v->rowresizechanging = 1
                   v->rowresizestart=temp-GRD_FormatAttribute(hwnd,temp1,0,GRDFORMAT_GETROWHEIGHT)
                   v->rowresizerow=v->rowresizetemp
               END IF
               SetCapture(hWnd)
         END IF 
         
         IF (wParam AND MK_LBUTTON) AND v->rowresizechanging > 0 THEN
                hDc = GetDC(hWnd)
                ExcludeClipRect(hdc,0,rc.Bottom - GS_.HScrollWidth,rc.Right,rc.Bottom)
                ExcludeClipRect(hdc,rc.Right - GS_.VScrollWidth,0,rc.Right,rc.Bottom)
                SetROP2 hDc, R2_notxorpen
                tempgdiobj = SelectObject(hDc,v->SolidLine)
 
                IF v->rowresizechanging = 2 THEN 'redraw old line to erase it
                      MovetoEx (hDc, rc.Left + GS_.GridLeft , v->Mousemovey,@templp)
                      Lineto hDc, rc.Right - GS_.VScrollWidth ,v->Mousemovey
                  ELSE
                       v->rowresizechanging = 2 'First time thru @v.rowresizeflag = 1 ; no need to erase old line
                END IF
 
                MovetoEx (hDc, rc.Left + GS_.GridLeft, lp.y,@templp)
                Lineto hDc, rc.Right - GS_.VScrollWidth ,lp.y
 
                SelectObject hDc, tempgdiobj
                SetROP2 hDc, R2_copypen
                ReleaseDC hWnd, hDc
 
         END IF

        MouseMoveBail:
        v->Mousemovex=lp.x
        v->Mousemovey=lp.y
    
     CASE  WM_LBUTTONDBLCLK
        'Enact the edit control
         GRD_ActivateEditControl(HWnd,3)
         GRD_ActivateListControl(Hwnd,0) 
      
     CASE  WM_LBUTTONDOWN
           skey=GetKeyState(VK_SHIFT) AND &H8000
       
           if (GS_.MyRowTemp <> GS_.OldSelRow or GS_.MyColTemp <> GS_.OldSelCol) and GS_.TempMMResult = 0 THEN  'on the grid
                GS_.OldSelCol = GS_.MyColTemp
                GS_.OldSelRow = GS_.MyRowTemp
                if skey = 0 then GRD_SendNotifyMessage v->hParent,hWnd,v->id,GRDN_CELLCHANGED,GS_.SelRow,GS_.SelCol,GS_.OldSelRow,GS_.OldSelCol         
           end if
           
          ' if (GS_.MyRowTemp = GS_.OldSelRow and GS_.MyColTemp = GS_.OldSelCol) and GS_.TempMMResult = 0 THEN  'on the grid
          '      exit function         
          ' end if
        
          'if lbuttondown in edit control then no effect else we changed
          'cells so don't save the edit text
          
          GRD_ActivateEditControl HWND, 0  'Deactivate Edit Control
          GRD_ActivateListControl Hwnd, 0
       
     
          
          if v->CurrentCursor <> 2 and  v->CurrentCursor <> 3 THEN  '2 is EW and 3 in NS cursor
            if GS_.TempMMResult = 1 THEN 'We are in the row header
                 if skey=0 THEN
                      GS_.SelRow = GS_.MyRowTemp
                      GS_.SelCol = GS_.MyColTemp
                      GS_.SelRow2 = GS_.SelRow
                      GS_.SelCol2 = GS_.TotCols
                   else
                      'GS_.SelRow = GS_.SelRow
                      GS_.SelCol = 1 
                      GS_.SelRow2 = GS_.MyRowTemp
                      GS_.SelCol2 = GS_.TotCols 
                  END IF
                  
                  if GS_.SelCol2 = GS_.TotCols THEN
                     GRD_SendNotifyMessage v->hParent,hWnd,v->id,GRDN_ROWSSELECTED,GS_.SelRow,GS_.SelCol,GS_.SelRow2,GS_.SelCol2  
                  END IF
                 
                 SetFocus hWnd
                 InvalidateRect hWnd, BYVAL  NULL, 0 : UpdateWindow hwnd
            END IF  
            
            if GS_.TempMMResult = 2 THEN 'We are in the column header
                 if skey=0 THEN
                      GS_.SelRow = GS_.MyRowTemp
                      GS_.SelCol = GS_.MyColTemp
                      GS_.SelRow2 = GS_.TotRows
                      GS_.SelCol2 = GS_.SelCol
                   else 
                      GS_.SelRow =1
                      GS_.SelRow2 = GS_.TotRows
                      GS_.SelCol2 = GS_.MyColTemp 
                  END IF
                  
                  if GS_.SelRow2 = GS_.TotRows THEN
                     GRD_SendNotifyMessage v->hParent,hWnd,v->id,GRDN_COLSSELECTED,GS_.SelRow,GS_.SelCol,GS_.SelRow2,GS_.SelCol2  
                  END IF
                 
                 SetFocus hWnd
                 InvalidateRect hWnd, BYVAL  NULL, 0 : UpdateWindow hwnd
            END IF 
            
            if GS_.TempMMResult = 5 THEN 'We are in the corner
                 
                      GS_.SelRow = 1
                      GS_.SelCol = 1
                      GS_.SelRow2 = GS_.TotRows
                      GS_.SelCol2 = GS_.TotCols
                
                 SetFocus hWnd
                 InvalidateRect hWnd, BYVAL  NULL, 0 : UpdateWindow hwnd
            END IF  
               
            if GS_.TempMMResult = 0 THEN 'We are on the grid
            
                        if skey=0 THEN
                            GS_.SelRow = GS_.MyRowTemp
                            GS_.SelCol = GS_.MyColTemp
                            GS_.SelRow2 = 0
                            GS_.SelCol2 = 0
                         else 
                            GS_.SelRow2 = GS_.MyRowTemp
                            GS_.SelCol2 = GS_.MyColTemp 
                        END IF
              
                  GRD_CalcCellFormat(Hwnd,GS_.MyrowTemp,GS_.MycolTemp,@Mycell)

                
                  If MyCell.CellType=GRDFCT_USERBUT or MyCell.CellType=GRDFCT_CKBOX or MyCell.CellType=GRDFCT_DDCBS THEN
                      
                        Select CASE MyCell.CellType
                             case  GRDFCT_USERBUT     
                                 v->UserButtonEffectFlag = MakeLong(GS_.MyrowTemp,GS_.MycolTemp)
                                 GRD_SendNotifyMessage v->hParent,hWnd,v->id,GRDN_USERBUTTON,GS_.MyrowTemp,GS_.MycolTemp,0,0
                                 
                             case  GRDFCT_CKBOX
                                GRD_GetCell(HWND,GS_.MyrowTemp,GS_.MycolTemp,varptr(MyWStr1))
                                if Len(MyWStr1) > 0 THEN
                                       MyWStr1 = ""
                                    else
                                       MyWStr1 = "X" 
                                END IF
                                GRD_PutCell(HWND,GS_.MyrowTemp,GS_.MycolTemp,varptr(MyWStr1))
                                GRD_SendNotifyMessage v->hParent,hWnd,v->id,GRDN_CHECKBOXTOGGLE,GS_.MyrowTemp,GS_.MycolTemp,Len(MyWStr1),0
                             case  GRDFCT_DDCBS
                                  GRD_ActivateListControl Hwnd, 1
                                  v->UserButtonEffectFlag = MakeLong(GS_.MyrowTemp,GS_.MycolTemp)
                                  GRD_SendNotifyMessage v->hParent,hWnd,v->id,GRDN_COMBOBOX,GS_.MyrowTemp,GS_.MycolTemp,MyCell.CellFormat,0
                           CASE ELSE
                              'Nothing for now
                        END SELECT 
                        
                                                                                                         
                   
                      else
'                       'Nothing for now
                  END IF 'MyCell
                  
                  if GS_.SelRow2 <> GS_.SelRow and GS_.SelCol2 <> GS_.SelCol THEN
                       if skey <> 0 then   GRD_SendNotifyMessage v->hParent,hWnd,v->id,GRDN_CELLSSELECTED,GS_.SelRow,GS_.SelCol,GS_.SelRow2,GS_.SelCol2
                  end if
                  
                 SetFocus hWnd
                 InvalidateRect hWnd, BYVAL  NULL, 0 : UpdateWindow hwnd
            END IF 
         END IF 'resize
            
     CASE  WM_LBUTTONUP
      
        IF  v->ColResizeCol > 0 THEN
          GRD_FormatAttribute(HWND,0,v->ColResizeCol,GRDFORMAT_SETCOLWIDTH,lp.x - v->ColResizeStart)
          GRD_SendNotifyMessage v->hParent,hWnd,v->id,GRDN_COLWIDTHCHANGED,lp.x - v->ColResizeStart,v->ColResizeCol,0,0
         
          v->ColResizeFlag=0
          v->ColResizeCol=0
          v->ColResizeStart=0
          v->RowResizeTemp =0
          v->ShowResizeLine=0
          v->ColResizeChanging = 0
          
          SetFocus(hWnd)
         
        END IF
        
        IF  v->RowResizeRow > 0 THEN
         
          GRD_FormatAttribute(HWND,v->RowResizeRow,0,GRDFORMAT_SETROWHEIGHT,lp.y - v->RowResizeStart)
          GRD_SendNotifyMessage v->hParent,hWnd,v->id,GRDN_ROWHEIGHTCHANGED,lp.y - v->RowResizeStart,v->RowResizeRow,0,0
        
          v->RowResizeFlag=0
          v->RowResizeRow=0
          v->RowResizeStart=0
          v->RowResizeTemp =0
          v->ShowResizeLine=0
          v->RowResizeChanging = 0
          
          SetFocus(hWnd)
          
        END IF

        'All Events
          v->colResizeChanging = 0
          v->RowResizeChanging = 0
          v->ColResizeFlag = 0
          v->ColResizeCol = 0
          v->ColResizeStart = 0
          v->ColResizeTemp = 0
          v->RowresizeFlag = 0
          v->RowresizeRow = 0
          v->RowresizeStart = 0
          'ClipCursor BYVAL %NULL
          GS_.SelectionProcess=0
          GS_.RowSelectionOn=0
          GS_.ColSelectionOn=0
          GS_.RowColSelectionOn=0
          v->UserButtonEffectFlag = 0 'turn off depressed button effect
          ReleaseCapture
          MyMouseMoveResult = GRD_MouseMoveUpdate(hwND,wParam,lParam)
          InvalidateRect hWnd, BYVAL NULL, 0 : UpdateWindow hwnd
          
     CASE  WM_KILLFOCUS,  WM_SETFOCUS ' Must process these if line is selected
        'IF GS_.SelRow > - 1 THEN
        '   InvalidateRect hWnd, BYVAL  NULL, 0 : UpdateWindow hwnd
        'ELSE
        '   GetClientRect hWnd, @wRect
        '   rc.Left = 0 : rc.Top = 0 : rc.Right = wRect.Right : rc.Bottom = GS_.cyChar
        '   hdc = GetDc(hWnd)
        '     DrawFocusRect hDC, @rc
        '   ReleaseDc hWnd, hdc
        'END IF
        
        'MouseMove important for pre locating selection rects etc.
        GetCursorPos(@lp)
        ScreentoClient(HWND,@lp)
        SendMessage Hwnd, WM_MOUSEMOVE, 0, cast(lParam,MAKELONG(lp.x,lp.y))
        

        IF wMsg =  WM_KILLFOCUS THEN 'Send notifications to parent
               SendMessage v->hParent,  WM_COMMAND, MAKELONG(v->id,  LBN_KILLFOCUS), cast(lparam,hwnd)
        ELSE
               SendMessage v->hParent,  WM_COMMAND, MAKELONG(v->id, LBN_SETFOCUS), cast(lparam,hwnd)
        END IF

     CASE  WM_PAINT 'Draw only the list items that needs to be shown
    
        hdc = BeginPaint(hWnd, @ps)
           GetClientRect(hWnd, @wRect)

           GRD_SetSBarLimits(hWnd)
          
           GRD_PaintCells hWnd
          
           GRD_PaintSelection hWnd
           
           BitBlt hDC, wRect.top, wRect.left, wRect.right, wRect.bottom, v->memDC, 0, 0,  SRCCOPY
        EndPaint(hWnd, @ps)

     CASE  WM_ERASEBKGND 'Prevent erasing background
        FUNCTION = 1

     CASE  LB_SETHORIZONTALEXTENT
   

     CASE  GRD_REFRESH
        GRD_SetSBar hwnd, GS_.wMaxHeight, MAX(0, GS_.RowCount), GS_.TopRow,  SB_VERT
        InvalidateRect hWnd, BYVAL  NULL, 0 : UpdateWindow hwnd

     CASE  GRD_GETSELECTED
        IF wParam > 0 THEN POKE LONG,wParam,GS_.SelRow
        IF lParam > 0 THEN POKE LONG,lParam,GS_.SelCol

     CASE  GRD_SETSELECTED
        if wParam > 0 and wParam <= GS_.TotRows and lParam > 0 and lParam <= GS_.TotCols  THEN
            GS_.SelRow = wParam
            GS_.SelCol = lParam
        END IF

     CASE  GRD_GETTOPROW
        FUNCTION = GS_.TopRow

     CASE  GRD_SETTOPROW
        IF wParam < 0 THEN
           GS_.TopRow = 1
        ELSE
           GS_.TopRow = MIN(wParam, GS_.TotRows)
        END IF
        GRD_SetSBAR hwnd, (GS_.BottomRow - GS_.TopRow), MAX(1, GS_.TotRows), GS_.TopRow, GRD_VSBAR
        IF lParam THEN 'Refresh control
           InvalidateRect hWnd, BYVAL  NULL, 0 : UpdateWindow hwnd
        END IF
       'FUNCTION = GS_.TopRow
        
     CASE GRD_GETROWCOLDIM
         IF wParam > 0 THEN POKE LONG,wParam,GS_.DimRows
         IF lParam > 0 THEN POKE LONG,lParam,GS_.DimCols
          
     CASE GRD_GETROWCOLVIEW
         IF wParam > 0 THEN POKE LONG,wParam,GS_.TotRows
         IF lParam > 0 THEN POKE LONG,lParam,GS_.TotCols
         
      CASE GRD_SETROWCOLVIEW
         IF wParam > 0 THEn GS_.TotRows = cast(LONG,wParam)
         IF lParam > 0 THEN GS_.TotCols = cast(LONG,lParam)
         
     CASE GRD_GETTOPLEFT
         IF wParam > 0 THEN POKE LONG,wParam,GS_.TopRow
         IF lParam > 0 THEN POKE LONG,lParam,GS_.LeftCol
         
     case GRD_GETTOTALSHEETS
          Function = v->TotalSheets
 
     CASE GRD_GETCURRENTSHEET
          Function = v->CS
          
     case GRD_SETCURRENTSHEET
         'wParam is not allowed to be negative
          if wParam <= v->TotalSheets THEN
              GRD_SendNotifyMessage v->hParent,hWnd,v->id,GRDN_SHEETCHANGED,v->CS,wParam,0,0
              v->CS = wParam
              SetFocus HWND
              GetClientRect(hWnd,@MyRect)
              PostMessage(hWnd,WM_SIZE,SIZE_RESTORED,MAKELONG(MyRect.Right-MyRect.Left,MyRect.Bottom-MyRect.Top))
              'InvalidateRect v->hSheetControl, BYVAL NULL, 0 : 'UpdateWindow v->hSheetControl
              InvalidateRect HWND, BYVAL NULL, 0 : UpdateWindow HWND  
          END IF
          
     cASE GRD_SETSTATUSBAR
        i = wParam
        if wParam < 1 THEN i = 1
        if wParam > 3 THEN i = 3
        if lParam <> 0 THEN
           MyWStrPtr = cast(wstring ptr, lParam)
           if i =1 THEN v->StatusBarText1 = *MyWStrPtr
           if i =2 THEN v->StatusBarText2 = *MyWStrPtr
           if i =3 THEN v->StatusBarText3 = *MyWStrPtr
        end if
        
     CASE GRD_SETSWITCH
 
           select CASE wParam
              case GRD_COMBOSUPRESSEDIT
                   v->ComboEdit = lParam
                  if lParam < 0 then v-> ComboEdit = 0
                  if lParam > 1 then v-> ComboEdit = 1
              case GRD_ROWCOLRESIZE
                   v->SuppressResize = lParam
                  if lParam < 0 then v-> SuppressResize = 0
                  if lParam > 1 then v-> SuppressResize = 1
              case GRD_RETURNKEYMODE
                   v->ReturnKeyMode = lParam
                  if lParam < 0 then v-> ReturnKeyMode = 0
                  if lParam > 1 then v-> ReturnKeyMode = 1   
              case GRD_TABKEYMODE
                   v->TabKeyMode = lParam
                  if lParam < 0 then v-> TabKeyMode = 0
                  if lParam > 1 then v-> TabKeyMode = 1      
              case GRD_HIDEGRIDLINES
                  v->HideGridLines = lParam
                  if lParam < 0 then v-> HideGridLines = 0
                  if lParam > 1 then v-> HideGridLines = 1         
              case GRD_CYTWEAK
                  v->cyTweak = lParam
                  if lParam < 0 then v-> cyTweak = 0
                  if lParam > 8 then v-> cyTweak = 8                 
              CASE GRD_SHOWSTATUSBAR  'Works on the current sheet
                   GS_.StatusBarWidth = lParam
                  if lParam < 0 then GS_.StatusBarWidth = 0
                  if lParam > 0 then GS_.StatusBarWidth = SBSIZE
                  GetClientRect(hWnd,@MyRect)
                  PostMessage(hWnd,WM_SIZE,SIZE_RESTORED,MAKELONG(MyRect.Right-MyRect.Left,MyRect.Bottom-MyRect.Top))
           END SELECT  
  
        
     CASE GRD_GETEDITHANDLE
         FUNCTION = cast(integer,v->hEditControl)
  
     CASE GRD_GETLISTHANDLE
         FUNCTION = cast(integer,v->hListControl)
  
     case GRD_SETTOTROWCOL  'for current sheet only
        if wParam > 0 and wParam <= GS_.DimRows and lParam > 0 and lParam <= GS_.DimCols  THEN
            GS_.TotRows = wParam
            GS_.TotCols = lParam
        END IF
        
        
     case GRD_GETTOTROWCOL
        if wParam <> 0 and lParam <> 0 THEN
            xPtr = cast(long ptr,wParam)
            yPtr = cast(long ptr,lParam)
            
           *xPtr = GS_.TotRows
           *yPtr = GS_.TotCols
        END IF
        
     case GRD_SETHSCROLLWIDTH 
          if wParam <> 0 THEN
             GS_.HScrollWidth = SBSIZE
           else
             GS_.HScrollWidth = SBSIZE 
          END IF
                  
     case GRD_SETVSCROLLWIDTH
          if wParam <> 0 THEN
             GS_.VScrollWidth = SBSIZE
           else
             GS_.VScrollWidth = 0 
          END IF
         
     case GRD_SETSTATUSBARWIDTH
           if wParam <> 0 THEN
             GS_.StatusBarWidth = SBSIZE
           else
             GS_.StatusBarWidth = 0
          END IF
         
     case GRD_SETRETURNKEYMODE
          if wParam <> 0 THEN
             v->ReturnKeyMode = 1
           else
             v->ReturnKeyMode = 0 
          END IF
         
     case GRD_SETTABKEYMODE
           if wParam <> 0 THEN
             v->TabKeyMode = 1
           else
             v->TabKeyMode = 0 
          END IF
          
      case GRD_HIDEGRIDLINES
         if wParam <> 0 THEN
             v->HideGridLines = 1
           else
             v->HideGridLines = 0 
          END IF
         
     case GRD_SETNUMBERDATEFORMATSTR
          if (wParam >= 0 and wParam <= GRD_NUMDATEFORMATSTR) and lParam <> 0  THEN
            MyWStrPtr = cast(Wstring ptr,lParam) 
            v->NumberDateFormatString(wParam) = *MyWStrPtr
          END IF
         
     case GRD_SETDATEINPUTFORMAT
          if wParam > 1 and wParam < NUMDATEINPUTS THEN
              v->DefaultDateInput = wParam 
          END IF
          
          
     case GRD_SETTABPOSITION
        if wParam >= 0 and wParam < 500 THEN
            v->TabSheetPos = wParam
        END IF
          
     
     case GRD_LOADBITMAP 'wParam = 0 is reserved for DRAWYOUROWN
        if (wParam > 0 and wParam < GRD_NUMBITMAPS) and lParam <> 0 THEN
              if v->Bitmaps(wParam) <> 0 THEN DeleteObject v->Bitmaps(wParam)
              v->Bitmaps(wParam) = cast(hBitmap,lParam)
              'print  v->Bitmaps(wParam)
          END IF
          
             
     CASE GRD_BITMAPTODRAWRECT
'         'wParam is handle to bitmap except if = 0 then PrintDC has been pre-filled
'         'lParam is RECT of cell
         if lParam = 0 THEN exit function
            
         if wParam = 0 THEN 'PrintDC has been pre loaded
            MyRectPtr = cast(RECT ptr,lParam)
            BitBlt v->memDC, MyRect.Left, MyRect.Top, MyRect.Right-MyRect.Left,MyRect.Bottom-MyRect.Top,v->PrintDC, 0, 0, SRCCOPY '
           else
            SelectObject(v->PrintDC, v->Bitmaps(wParam)) 
            temp3 = cast(integer,VARPTR(myRect))
            temp4 = SIZEOF(myRect)
            CopyMemory(@temp3,cast(any ptr,lparam),temp4)
            print MyRect.Left

            BitBlt v->memDC, MyRect.Left, MyRect.Top, MyRect.Right-MyRect.Left,MyRect.Bottom-MyRect.Top,v->PrintDC, 0, 0, SRCCOPY '
         END IF

     CASE GRD_GETSETCELL
         Dim MyCell as GETSETCELL
         Dim MyWstrGetPut as WSTRING ptr
         Dim MystrGetPut as ZSTRING ptr
          
         if wParam = 0 THEN exit function
         if lParam = 0 THEN exit function
         CopyMemory(@MyCell,cast(any ptr,wParam),20)
         if MyCell.Ansi =0 THEN
             MyWstrGetPut=cast(any ptr,lParam)
           else
             MystrGetPut=cast(any ptr,lParam)  
         END IF
         
          
         if MyCell.Action=0 and MyCell.Ansi = 0 then 'get
            if MyCell.Ansi =0 THEN
                 GRD_GetCell(hWnd,MyCell.row,MyCell.col,MyWstrGetPut,MyCell.Sheet)
              else
                 GRD_GetCellA(hWnd,MyCell.row,MyCell.col,MystrGetPut,MyCell.Sheet)
            end if
          else 'put
            if MyCell.Ansi =0 THEN
                 GRD_PutCell(hwnd,MyCell.row,MyCell.col,MyWstrGetPut,MyCell.Sheet)
              else
                 GRD_PutCellA(hwnd,MyCell.row,MyCell.col,MystrGetPut,MyCell.Sheet)
            end if
         END IF
           
     CASE ELSE  ' Process all other messages
        FUNCTION = DefWindowProc(hWnd, wMsg, wParam, lParam)

  END SELECT

END FUNCTION

Sub GRD_Int()

  Dim wc AS WNDCLASSEX
  Dim szClassName AS string

  szClassName      = "JEKGRID"
 
  IF GetClassInfoEx(GetModuleHandle(BYVAL NULL), STRPTR(szClassName), @wc) <> 0 THEN
'    'It is already registered
    EXIT SUB
 END IF
  
  wc.cbSize        = SIZEOF(wc)
  wc.style         = CS_DBLCLKS
  wc.lpfnWndProc   = @GRD_GridProc
  wc.cbWndExtra    = 4  '4 extra bytes for pointer to UDT
  wc.hInstance     = GetModuleHandle(BYVAL NULL)
  wc.hCursor       = LoadCursor(NULL, BYVAL IDC_ARROW)
  wc.lpszClassName = STRPTR(szClassName)

  RegisterClassEx @wc
END SUB

